﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

using Utils;
using FlightPlan.Optimizer.TSPAlgorithms;

namespace FlightPlan.Optimizer
{
    /// <summary>
    /// Represents an utility class to generate FlightPlans according to some parameters.
    /// </summary>
    /// <author>Santiago Jiménez Serrano</author>
    public static class FPGenerator
    {
        #region - Public -

        /// <summary>
        /// Gets the optimized flight plan according to the specified waypoints and the algorithm to be used.
        /// </summary>
        /// <param name="wayPoints">The points where the fly has to pass.</param>
        /// <param name="algType">The type of TSP Algorithm to be used.</param>
        /// <returns>The whole FlightPlan object.</returns>
        public static FlightPlan GetOptimizedFlighPlan(Waypoint[] wayPoints, TSPAlgorithmType algType)
        {
            ITSPAlgorithm alg = TSPAlgorithm.GetTSPAlgorithm(algType);

            Leg[] legs = alg.GetSortedPath(wayPoints);
            return GetDefaultFlightPlan(new Fix[] { }, legs); 
        }

        /// <summary>
        /// Gets the optimized flight plan according to the specified waypoints and the algorithm to be used.
        /// </summary>
        /// <param name="wayPoints">The points where the fly has to pass.</param>
        /// <param name="alg">The Algorithm to be used.</param>
        /// <returns>The whole FlightPlan object.</returns>
        public static FlightPlan GetOptimizedFlighPlan(Waypoint[] wayPoints, ITSPAlgorithm alg)
        {
            Leg[] legs = alg.GetSortedPath(wayPoints);
            return GetDefaultFlightPlan(new Fix[] { }, legs); 
        }

        /// <summary>
        /// Creates a Default FlightPlan according to the specified fixes and legs lists.
        /// </summary>
        /// <param name="fixList">The list containing the fixes.</param>
        /// <param name="legList">The list containing the sorted legs.</param>
        /// <returns>The FlighPlan object by default.</returns>
        /// <exception cref="System.ArgumentNullException">If some list is null or empty.</exception>
        public static FlightPlan GetDefaultFlightPlan(Fix[] fixList, Leg[] legList)
        {
            //if(fixList == null || fixList.Length == 0)
            //    throw new ArgumentNullException("fixList");

            if (legList == null || legList.Length == 0)
                throw new ArgumentNullException("legList");

            return GenerateFP(fixList, legList);
        }

        /// <summary>
        /// Creates a Default FlightPlan according to the specified fixes and legs lists.
        /// </summary>
        /// <param name="fixList">The list containing the fixes.</param>
        /// <param name="legList">The list containing the sorted legs.</param>
        /// <returns>The FlighPlan object by default.</returns>
        /// <exception cref="System.ArgumentNullException">If some list is null or empty.</exception>
        public static FlightPlan GetDefaultFlightPlan(List<Fix> fixList, List<Leg> legList)
        {
            //if (fixList == null || fixList.Count == 0)
            //    throw new ArgumentNullException("fixList");

            if (legList == null || legList.Count == 0)
                throw new ArgumentNullException("legList");

            return GetDefaultFlightPlan(fixList.ToArray(), legList.ToArray());
        }

        public static FlightPlan GetTestFlightPlan()
        {
            Object[] o = ToUse2();
            Fix[] fixes = (Fix[])o[0];
            Leg[] legs  = (Leg[])o[1];

            return GetDefaultFlightPlan(fixes, legs);
        }

        #endregion
        #region - Private -

        #region - Stages -

        private static Stage CreateStage(FlightPlan fp, String id, Stage.Type type, String xmlComment)
        {
            Stage st      = new Stage(fp);
            st.ID         = id;
            st.type       = type;
            st.XmlComment = xmlComment;
            return st;
        }

        private static Stage CreateStageTaxi1(FlightPlan fp)
        {
            Stage st = CreateStage(fp, "taxi", Stage.Type.Taxi, "Taxi to runway");
            st.ManualOnly = true;
            st.Ignore = true;
            return st;
        }

        private static Stage CreateStageTakeOff(FlightPlan fp)
        {
            Stage st = CreateStage(fp, "takeoff", Stage.Type.TakeOff, "Take Off");
            st.Ignore = true;
            return st;
        }

        private static Stage CreateStageDepart(FlightPlan fp, Leg[] legList)
        {
            Stage st = CreateStage(fp, "depart", Stage.Type.Departure, "Depart stage");

            // Name & Description
            st.Name        = "Departure Procedure";
            st.Description = "With the VAS in WP Navigation, this leg actually implements a Takeoff."; // ????

            // Gets the first leg in the list
            // it must be a IFLeg
            Leg l = legList[0];
            if (!(l is IFLeg))
                throw new ArgumentException("The first leg in the sorted list must be an IFLeg");

            // add the leg
            l.stage = st;
            st.Add(l);

            // Initial & Final legs
            st.AddInitial(l.ID);
            st.AddFinal(l.ID);

            return st;
        }

        private static Stage CreateStageRoute(FlightPlan fp, Leg[] legList)
        {
            Stage st = CreateStage(fp, "CtrlLegs", Stage.Type.Route, "Traffic pattern extended with an iterative leg");

            // Name & Description
            st.Name = "Traffic pattern";

            // test for errors
            if (legList.Length < 3)
                throw new ArgumentException("The leg list must contain at less 3 legs.");

            // Add the legs
            for (int i = 1; i < legList.Length - 1; i++)
            {
                legList[i].stage = st;
                st.Add(legList[i]);
            }

            // Initial & Final legs
            st.AddInitial(legList[1].ID);              // the second one
            st.AddFinal(legList[legList.Length-2].ID); // the one before the last

            return st;
        }

        private static Stage CreateStageArrival(FlightPlan fp, Leg[] legList)
        {
            Stage st = CreateStage(fp, "arrival", Stage.Type.Arrival, "In this example, arrival actually finalizes the FP");

            // Name & Description
            st.Name = "Arrival";
            st.Description = "Set head of runway as destination together with 0kt speed.";

            // Gets the last leg in the list
            // it must be a TFLeg
            Leg l = legList[legList.Length-1];
            if (!(l is TFLeg))
                throw new ArgumentException("The last leg in the sorted list must be a TFLeg");

            // add the leg
            l.stage = st;
            st.Add(l);

            // Initial & Final legs
            st.AddInitial(l.ID);
            st.AddFinal(l.ID);

            return st;
        }

        private static Stage CreateStageApproach(FlightPlan fp)
        {
            Stage st = CreateStage(fp, "approach", Stage.Type.Approach, "Approach stage");
            st.Ignore = true;
            return st;
        }

        private static Stage CreateStageLand(FlightPlan fp)
        {
            Stage st = CreateStage(fp, "land", Stage.Type.Land, "Landing stage");
            st.Ignore = true;
            return st;
        }

        private static Stage CreateStageTaxi2(FlightPlan fp)
        {
            Stage st = CreateStage(fp, "taxi2", Stage.Type.Taxi, "Final Taxi");
            st.ManualOnly = true;
            st.Ignore = true;
            return st;
        }

        private static void AddStagesToFlightPlan(FlightPlan fp, Leg[] legList)
        {
            fp.stages.Add(CreateStageTaxi1(fp));
            fp.stages.Add(CreateStageTakeOff(fp));
            fp.stages.Add(CreateStageDepart(fp, legList));
            fp.stages.Add(CreateStageRoute(fp, legList));
            fp.stages.Add(CreateStageArrival(fp, legList));
            fp.stages.Add(CreateStageApproach(fp));
            fp.stages.Add(CreateStageLand(fp));
            fp.stages.Add(CreateStageTaxi2(fp));
        }

        #endregion


        private static void AddLocaleSettings(FlightPlan fp)
        {
            fp.LocaleInfo.speedUnits    = Units.Speed.kt;
            fp.LocaleInfo.altitudeUnits = Units.Length.m;
            fp.LocaleInfo.distanceUnits = Units.Length.m;
            fp.LocaleInfo.fprovider.NumberDecimalSeparator = ".";
            fp.LocaleInfo.fprovider.NumberGroupSeparator = "";
        }

        /// <summary>
        /// Checks the correctness of the initial legs distribution.
        /// </summary>
        private static void CheckInitialLegs(FlightPlan fp)
        {
            foreach (Leg l in fp.stages[0].initialLegs)
            {
                if (!((l is IFLeg) || (l is IRFLeg) || (l is TOLeg)))                
                {
                    throw new Exception(String.Format("Leg {0} cannot be an initial flightplan leg.", l.ID));
                }
            }
        }

        private static FlightPlan GenerateFP(Fix[] fixList, Leg[] legList)
        {
            FlightPlan fp = new FlightPlan();

            // Locale Settings
            AddLocaleSettings(fp);

            // Add the fixes
            for (int i = 0; i < fixList.Length; i++)
                fp.Fixes.Add(fixList[i]);


            // list of emergency flight plans
            // EMPTY

            // Main Flight Plans
            fp.id          = "AutoGenFP";
            fp.name        = "Auto-generated FlightPlan";
            fp.description = "Auto-generated FlightPlan description";            

            // Add all the stages
            AddStagesToFlightPlan(fp, legList);

            // Once all stages have been read, initialize references.
            for (int i = 0; i < fp.stages.Count; i++)
                fp.stages[i].UpdateReferences();

            CheckInitialLegs(fp);

            return fp;
        }

        private static Object[] ToUse()
        {
            String f1Name = "Barcelona-El Prat LEBL 02 Runway";
            String f1Desc = "Runway L12 of Barcelona airport";
            RwyFix f1 = new RwyFix("LEBL02", f1Name, f1Desc, new GPSCoords(41.287772, true), new GPSCoords(2.084840, false), 2, 190, 2598);
           // fp.Fixes.Add(f1);



            // Leg List specification

            // the first leg in the list must be a IFLeg
            // the last  leg in the list must be a TFLeg

            // Initial leg
            Waypoint dest = new Waypoint(240, 600, new GPSCoords(41.38765627165685, true), new GPSCoords(2.122877987908149, false), false, null);
            IFLeg legIni = new IFLeg(null, "depart", dest, null, null, null);

            // legs
            DFLeg l1; BasicScan l2; TFLeg l3;

            // leg1
            Waypoint w1 = new Waypoint(null, null, new GPSCoords(41.42878320802662, true), new GPSCoords(1.93906967322802, false), false, null);
            l1 = new DFLeg(null, "toscan", w1, "basicScan", null, null);

            // leg2            
            Waypoint w2 = new Waypoint(null, null, new GPSCoords(41.14917883473012, true), new GPSCoords(1.646321457416098, false), false, null);
            TFLeg tf2 = new TFLeg(null, "basicScan", w2, "torwy", null, null); 
            l2 = new BasicScan(null, "basicScan", w2, "torwy", null, null, -15000, 35000, 65, 10000, null, null, null, null);

            // leg3
            Waypoint w3 = new Waypoint(null, null, new GPSCoords(41.20205334508894, true), new GPSCoords(2.046275812440561, false), false, null);
            l3 = new TFLeg(null, "torwy", w3, null, null, null);

            // Final leg
            Waypoint destination = new Waypoint(0, null, f1, false, null);
            TFLeg legEnd = new TFLeg(null, "final-approach", destination, null, null, null);

           

            // List of legs and fixes
            Fix[] fixList = new Fix[] { f1 };
            Leg[] legList = new Leg[] { legIni, l1, tf2, l3, legEnd};

            return new Object[] { fixList, legList };
        }

        private static Object[] ToUse2()
        {
            // WayPoints
            Waypoint dep = new Waypoint(240, 600,   new GPSCoords(41.38765627165685, true), new GPSCoords(2.122877987908149, false), false, null);
            Waypoint w1  = new Waypoint(null, null, new GPSCoords(41.42878320802662, true), new GPSCoords(1.93906967322802, false), false, null);
            Waypoint w2  = new Waypoint(null, null, new GPSCoords(41.14917883473012, true), new GPSCoords(1.646321457416098, false), false, null);
            Waypoint w3  = new Waypoint(null, null, new GPSCoords(41.20205334508894, true), new GPSCoords(2.046275812440561, false), false, null);
            Waypoint end = new Waypoint(null, null, new GPSCoords(41.287772, true),         new GPSCoords(2.084840, false),          false, null);
          
            // Leg List specification

            // the first leg in the list must be a IFLeg
            // the last  leg in the list must be a TFLeg

            // Initial leg            
            IFLeg legIni = new IFLeg(null, "depart", dep, null, null, null);

            // Route legs
            TFLeg l1 = new TFLeg(null, "toscan",    w1, "basicScan", null, null);
            TFLeg l2 = new TFLeg(null, "basicScan", w2, "torwy", null, null);
            TFLeg l3 = new TFLeg(null, "torwy",     w3, null, null, null);

            // Final leg            
            TFLeg legEnd = new TFLeg(null, "final-approach", end, null, null, null);

            // List of legs and fixes
            Fix[] fixList = new Fix[] {  };
            Leg[] legList = new Leg[] { legIni, l1, l2, l3, legEnd };

            return new Object[] { fixList, legList };
        }

        #endregion
    }
}
