﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.IO;
using System.Text;
using System.Windows.Forms;

using FlightPlan.Optimizer.Airports;
using FlightPlan.Optimizer.TSPAlgorithms;
using FlightPlan.Optimizer.Wfs;
using FlightPlan.Optimizer.Wfs.FeatureTypes;

namespace FlightPlan.Optimizer.GUI
{
    /// <summary>
    /// Wfs Fires Form. 
    /// Helps to create a Optimized FlightPlan with the Wfs Web server data.
    /// </summary>
    /// <author>Santiago Jiménez Serrano [sanjiser@gmail.com]</author>
    public partial class WfsFiresForm : Form
    {
        #region - Attributes -

        private ListOfAirports    _listOfAirports;    // list of fires
        private ListOfFires       _listOfFires;       // list of nodes in the graph
        private WfsDataDownloader _wfsDataDownloader; // data downloader object.
        private String            _csvPath;           // input data file path
        private ITSPAlgorithm     _algorithm;         // algorithm been used.

        #endregion
        #region - Constructor -

        /// <summary>
        /// Creates a new instance of WfsFiresForm.
        /// </summary>
        public WfsFiresForm()
        {
            InitializeComponent();
            Init();
        }

        #endregion
        #region - Properties -

        /// <summary>
        /// Gets the list of Airports for this Form.
        /// </summary>
        public ListOfAirports Airports
        {
            get { return _listOfAirports; }
        }

        /// <summary>
        /// Gets the list of fires for this Form.
        /// </summary>
        public ListOfFires ListOfFires
        {
            get { return _listOfFires; }
        }

        /// <summary>
        /// Gets the WfsDataDownloader object for this Form.
        /// </summary>
        public WfsDataDownloader WfsDataDownloader
        {
            get { return _wfsDataDownloader; }
        }

        /// <summary>
        /// Gets or sets the path to the input csv file.
        /// </summary>
        public String CsvPath
        {
            get { return _csvPath;  }
            set { _csvPath = value; }
        }

        /// <summary>
        /// Gets or sets the status text in this Form.
        /// </summary>
        public String Status
        {
            get { return _lblStatus.Text; }
            set { _lblStatus.Text = value; }
        }

        /// <summary>
        /// Gets or Sets the FeatureType in this form GUI.
        /// </summary>
        public FeatureType FeatureType
        {
            get
            {
                if (_rbtFires7Days.Checked)  return Wfs.FeatureType.Fires7Days;
                if (_rbtFires30Days.Checked) return Wfs.FeatureType.Fires30Days;
                else                         return Wfs.FeatureType.FiresAll;
            }

            set
            {
                     if (value == Wfs.FeatureType.Fires7Days)  _rbtFires7Days.Checked  = true;
                else if (value == Wfs.FeatureType.Fires30Days) _rbtFires30Days.Checked = true;
                else                                           _rbtFiresAll.Checked    = true;
            }
        }

        /// <summary>
        /// Gets or sets the output path for the FlightPlan.
        /// </summary>
        public String PathToOutputFlightPlan
        {
            get { return _txtOutputFlightPlan.Text; }
            set 
            {
                _txtOutputFlightPlan.Text = value;
                _ttC.SetToolTip(_txtOutputFlightPlan, value);
            }
        }

        /// <summary>
        /// Gets the number of items at the list view, and set such info at the status bar.
        /// </summary>
        private int NbItems
        {
            get
            {
                if(_listOfFires != null)
                    return _listOfFires.Count;
                return 0;
            }
            set
            {
                _lblNbItems.Text = value.ToString() + " items";
            }
        }

        /// <summary>
        /// Gets the selected airport at the combobox.
        /// </summary>
        public Airport SelectedAirport
        {
            get 
            {
                int index = _cbAirport.SelectedIndex;
                return Airports[index];
            }
        }

        #endregion
        #region - Private -

        #region - Init -

        /// <summary>
        /// Initialize all the gui and attributes elements in this form.
        /// </summary>
        private void Init()
        {
            // Init the attributes
            _listOfAirports = new ListOfAirports();
            _wfsDataDownloader = new WfsDataDownloader();

            InitListViewColumns();
            InitPathToOutputFP();
            InitComboAlgorithmTypes();
            InitComboAirports();
        }

        /// <summary>
        /// Initialize the listview columns.
        /// </summary>
        private void InitListViewColumns()
        {
            String[] colNames = Fire.GetListViewColumnNames();

            _listView.Columns.Clear();

            foreach (String cn in colNames)
            {
                _listView.Columns.Add(cn, 125);
            }
        }

        /// <summary>
        /// Initialize the Output FlightPlan file to a non existing one.
        /// </summary>
        private void InitPathToOutputFP()
        {
            string s = Environment.GetFolderPath(Environment.SpecialFolder.Personal);
            s = Path.Combine(s, "EFFIS_results");
            if(!Directory.Exists(s))
                Directory.CreateDirectory(s);
            s = Utils.CreateUniqueFileName(s, "OptimizedFlightPlan", ".xml");
            PathToOutputFlightPlan = s;
        }

        /// <summary>
        /// Initialize the AlgorithmTypes in the combo box.
        /// </summary>
        private void InitComboAlgorithmTypes()
        {
            foreach (TSPAlgorithmType t in Enum.GetValues(typeof(TSPAlgorithmType)))
            {
                ITSPAlgorithm alg = TSPAlgorithm.GetTSPAlgorithm(t);
                _cbAlgType.Items.Add(alg);
            }

            if (_cbAlgType.Items.Count > 0)
            {
                _cbAlgType.SelectedIndex = 0;
            }
        }

        /// <summary>
        /// Initialize the Airports in the combo box.
        /// </summary>
        public void InitComboAirports()
        {
            foreach (Airport a in this.Airports)
            {
                _cbAirport.Items.Add(a);
            }


            if (_cbAirport.Items.Count > 0)
            {
                _cbAirport.SelectedIndex = 0;
            }
        }

        #endregion
        #region - Data Tab -

        /// <summary>
        /// Loads the fires data from the file specified by the user and the current gui FeatureType.
        /// </summary>
        private void LoadFromFile()
        {
            using (OpenFileDialog ofd = new OpenFileDialog())
            {
                ofd.Filter = "csv files (*.csv)|*.csv|All files (*.*)|*.*";
                ofd.Title = "Select csv file";
                if (ofd.ShowDialog(this) == DialogResult.OK)
                {
                    LoadFromFile(ofd.FileName, this.FeatureType);
                }
            }
        }

        /// <summary>
        /// Loads the fires data from the specified file and FeatureType
        /// </summary>
        /// <param name="filename">The csv file from which to get the data.</param>
        /// <param name="type">The kind of data inside the file.</param>
        private void LoadFromFile(String filename, FeatureType type)
        {
            try
            {
                _listOfFires = new ListOfFires(type, filename);
                CsvPath = filename;
                RefreshListView();
                Status = filename + " loaded";
                this.Text = "FlightPlan Optimizer - EFFIS:Fires - " + Path.GetFileName(CsvPath);
                this._btRefresh.Enabled = true;
                this._btFilterSelected.Enabled = true;
            }
            catch (Exception e)
            {
                MessageBox.Show(this, e.Message, "Error", MessageBoxButtons.OK, MessageBoxIcon.Exclamation);
                Status = "Error loading file: " + filename;
            }
        }

        /// <summary>
        /// Loads the fires data from the Wfs Server.
        /// </summary>
        /// <param name="type">The type of data to be downloaded. Must be a fire type.</param>
        private void LoadFromWfsWebServer(FeatureType type)
        {
            // test feature
            if (type != FeatureType.FiresAll && 
                type != FeatureType.Fires7Days && 
                type != FeatureType.Fires30Days)
            {
                MessageBox.Show(this, "This form only is able to load 'Fire' data.", "Error", MessageBoxButtons.OK, MessageBoxIcon.Exclamation);
                Status = "Error: This form only is able to load 'Fire' data.";
            }

            try
            {
                WfsUrlFactory fac = new WfsUrlFactory();
                fac.FeatureType = this.FeatureType;
                String url = fac.GetUrl();
                Status = "Downloading data from " + url + " ...";
                String path = WfsDataDownloader.DownloadData(type);
                LoadFromFile(path, type);
            }
            catch (Exception e)
            {
                MessageBox.Show(this, e.Message, "Error", MessageBoxButtons.OK, MessageBoxIcon.Exclamation);
                Status = "Error downloading data: " + e.Message;
            }
        }

        /// <summary>
        /// Loads the fires data from the Wfs Server with the gui FeatureType.
        /// </summary>
        private void LoadFromWfsWebServer()
        {
            LoadFromWfsWebServer(this.FeatureType);
        }

        /// <summary>
        /// Ask the user for the folder where to save the downloaded data.
        /// </summary>
        private void ChangeDownloadDataDirectory()
        {
            using (FolderBrowserDialog fbd = new FolderBrowserDialog())
            {
                fbd.SelectedPath = _wfsDataDownloader.DirectoryPath;
                fbd.ShowNewFolderButton = true;
                fbd.Description = "Select folder where to save the downloaded data";
                if (fbd.ShowDialog(this) == DialogResult.OK)
                {
                    _wfsDataDownloader.DirectoryPath = fbd.SelectedPath;
                }
            }
        }

        #endregion
        #region - Main Functions -

        /// <summary>
        /// Creates the FlightPlan with the current gui data.
        /// </summary>
        private void Run()
        {
            if (ListOfFires == null)
            {
                Status = "No data loaded";
                MessageBox.Show(this, "No data loaded", "Error", MessageBoxButtons.OK, MessageBoxIcon.Exclamation);
                return;
            }

            Status = "Creating Optimized FlightPlan...";

            Waypoint startPoint = SelectedAirport.Waypoint;
            Waypoint[] points = _listOfFires.GetWaypointsArray(startPoint);

            DisableControlsBeforeRun();

            // Event Listeners
            _algorithm.Ends             += new EventHandler(_algorithm_Ends);
            _algorithm.NotSolutionFound += new EventHandler(_algorithm_NotSolutionFound);

            // Run in a new thread
            _algorithm.GetSortedPathInNewThread(points);
        }

        /// <summary>
        /// Selects the Output FlightPlan xml file by the user.
        /// </summary>
        private void SelectOutputFlightPlan()
        {
            using (SaveFileDialog sfd = new SaveFileDialog())
            {
                sfd.Title = "Output Flight Plan File";
                sfd.Filter = "Xml files (*.xml)|*.xml";
                if (PathToOutputFlightPlan != "")
                    sfd.FileName = PathToOutputFlightPlan;

                if (sfd.ShowDialog() == DialogResult.OK)
                {
                    PathToOutputFlightPlan = sfd.FileName;
                }
            }
        }

        #endregion
        #region - Filter Functions -

        /// <summary>
        /// Filters the Selected Items in the list vies.
        /// </summary>
        private void FilterSelectedItems()
        {
            if (ListOfFires == null)
            {
                Status = "Don't exist items to be filtered.";
                return;
            }

            if (_listView.SelectedItems.Count == 0)
            {
                Status = "Select some items in the list to be filtered.";
                return;
            }

            foreach (ListViewItem item in _listView.SelectedItems)
            {
                AbstractFeatureType aft = (AbstractFeatureType)item.Tag;
                ListOfFires.FilterId(aft.Id);
            }

            RefreshListView();
        }

        #endregion
        #region - GUI Utils -

        /// <summary>
        /// Refresh the ListView items from the current fires list.
        /// </summary>
        private void RefreshListView()
        {
            try
            {
                SuspendLayout();
                _listView.Items.Clear();
                if (_listOfFires != null)
                {
                    foreach (AbstractFeatureType f in _listOfFires)
                    {
                        _listView.Items.Add(f.ToListViewItem());
                    }
                }
                NbItems = _listOfFires.Count;
            }
            finally
            {
                ResumeLayout();
            }
        }

        /// <summary>
        /// Disables the gui controls before a run.
        /// </summary>
        private void DisableControlsBeforeRun()
        {
            // FlightPlan tab controls
            _btRun.Enabled = false;
            _prBar.Visible = true;
            _prBar.Style   = ProgressBarStyle.Marquee;
            _pgAlgProperties.Enabled = false;

            // ComboBoxes
            _cbAirport.Enabled =
            _cbAlgType.Enabled = false;

            // Buttons
            _btFilterAfter.Enabled =
            _btFilterBefore.Enabled =
            _btFilterByArea.Enabled =
            _btFilterByBB.Enabled =
            _btFilterByClass.Enabled =
            _btFilterbyCommune.Enabled =
            _btFilterByCountry.Enabled =
            _btFilterByCountryFul.Enabled =
            _btFilterByProvince.Enabled =
            _btFilterSelected.Enabled =
            _btGetDataFromWfs.Enabled =
            _btRefresh.Enabled =
            _btSelectCsvFile.Enabled =
            _btSelectOutputFlightPlan.Enabled =
            _btSetClosestAirport.Enabled = 
            _btChangeFolder.Enabled = false;
        }

        /// <summary>
        /// Enables the gui controls before a run.
        /// </summary>
        private void EnableControlsAfterRun()
        {
            // FlightPlan tab controls
            _btRun.Enabled = true;
            _prBar.Style = ProgressBarStyle.Blocks;
            _prBar.Visible = false;
            _pgAlgProperties.Enabled = true;

            // ComboBoxes
            _cbAirport.Enabled =
            _cbAlgType.Enabled = true;

            // Buttons
            _btFilterAfter.Enabled =
            _btFilterBefore.Enabled =
            _btFilterByArea.Enabled =
            _btFilterByBB.Enabled =
            _btFilterByClass.Enabled =
            _btFilterbyCommune.Enabled =
            _btFilterByCountry.Enabled =
            _btFilterByCountryFul.Enabled =
            _btFilterByProvince.Enabled =
            _btFilterSelected.Enabled =
            _btGetDataFromWfs.Enabled =
            _btRefresh.Enabled =
            _btSelectCsvFile.Enabled =
            _btSelectOutputFlightPlan.Enabled =
            _btSetClosestAirport.Enabled =
            _btChangeFolder.Enabled      = true;
        }

        /// <summary>
        /// Set the closest Airport at the gui.
        /// </summary>
        private void SetClosestAirport()
        {
            if (ListOfFires != null)
            {
                Airport a = Airports.GetClosestAirport(ListOfFires);
                int index = Airports.IndexOf(a);
                _cbAirport.SelectedIndex = index;
            }
        }

        #endregion
        #region - Event Listeners -

        #region - Buttons -

        #region - Normal Buttons -

        private void _btSelectCsvFile_Click(object sender, EventArgs e)
        {
            LoadFromFile();
        }

        private void _btGetDataFromWfs_Click(object sender, EventArgs e)
        {
            LoadFromWfsWebServer();
        }

        private void _btRefresh_Click(object sender, EventArgs e)
        {
            LoadFromFile(CsvPath, this.FeatureType);
        }

        private void _btChangeFolder_Click(object sender, EventArgs e)
        {
            ChangeDownloadDataDirectory();
        }

        /// <summary>
        /// Handles the click event over the Run button.
        /// </summary>
        private void _btRun_Click(object sender, EventArgs e)
        {
            Run();
        }

        /// <summary>
        /// Handles the click event over the Set Closest Airport button.
        /// </summary>
        private void _btSetClosestAirport_Click(object sender, EventArgs e)
        {
            SetClosestAirport();
        }

        /// <summary>
        /// Handles the click event over the Set Closest Airport button.
        /// </summary>
        private void _btSelectOutputFlightPlan_Click(object sender, EventArgs e)
        {
            SelectOutputFlightPlan();
        }

        #endregion
        #region - Filter -

        private void _btFilterSelected_Click(object sender, EventArgs e)
        {
            FilterSelectedItems();
        }

        private void _btFilterByCountry_Click(object sender, EventArgs e)
        {
            if (_txtFilterCountry.Text != "")
            {
                if (_listOfFires != null)
                {
                    _listOfFires.FilterByCountry(_txtFilterCountry.Text);
                    RefreshListView();
                }
            }
        }

        private void _btFilterByProvince_Click(object sender, EventArgs e)
        {
            if (_txtFilterProvince.Text != "")
            {
                if (_listOfFires != null)
                {
                    _listOfFires.FilterByProvince(_txtFilterProvince.Text);
                    RefreshListView();
                }
            }
        }

        private void _btFilterbyCommune_Click(object sender, EventArgs e)
        {
            if (_txtFilterCommune.Text != "")
            {
                if (_listOfFires != null)
                {
                    _listOfFires.FilterByCommune(_txtFilterCommune.Text);
                    RefreshListView();
                }
            }
        }

        private void _btFilterByCountryFul_Click(object sender, EventArgs e)
        {
            if (_txtFilterCountryFul.Text != "")
            {
                if (_listOfFires != null)
                {
                    _listOfFires.FilterByCountryFul(_txtFilterCountryFul.Text);
                    RefreshListView();
                }
            }
        }

        private void _btFilterByArea_Click(object sender, EventArgs e)
        {
            if (_txtFilterArea.Text != "")
            {
                if (_listOfFires != null)
                {
                    bool b = false;
                    int area;
                    b = int.TryParse(_txtFilterArea.Text, out area);

                    if (b)
                    {
                        _listOfFires.FilterByArea_HA(area, int.MaxValue);
                        RefreshListView();
                    }
                    else
                    {
                        MessageBox.Show("The area is not a number.");
                        Status = "The area is not a number";
                    }
                }
            }
        }

        private void _btFilterBefore_Click(object sender, EventArgs e)
        {
            if (_dtPickerBefore.Value != null)
            {
                if (_listOfFires != null)
                {
                    _listOfFires.FilterDateBefore(_dtPickerBefore.Value);
                    RefreshListView();
                }
            }
        }

        private void _btFilterAfter_Click(object sender, EventArgs e)
        {
            if (_dtPickerAfter.Value != null)
            {
                if (_listOfFires != null)
                {
                    _listOfFires.FilterDateAfter(_dtPickerAfter.Value);
                    RefreshListView();
                }
            }
        }

        private void _btFilterByClass_Click(object sender, EventArgs e)
        {
            if (_txtFilterClass.Text != "")
            {
                if (_listOfFires != null)
                {
                    _listOfFires.FilterByClass(_txtFilterClass.Text);
                    RefreshListView();
                }
            }
        }

        private void _btFilterByBB_Click(object sender, EventArgs e)
        {
            double lat1, lon1, lat2, lon2;

            if (!ParseDouble(_txtLat1.Text, out lat1)) return;
            if (!ParseDouble(_txtLon1.Text, out lon1)) return;
            if (!ParseDouble(_txtLat2.Text, out lat2)) return;
            if (!ParseDouble(_txtLon2.Text, out lon2)) return;

            if (_listOfFires != null)
            {
                _listOfFires.FilterByBoundingBox(lat1, lon1, lat2, lon2);
                RefreshListView();
            }
        }

        private bool ParseDouble(string text, out double value)
        {
            if (text == "")
            {
                value = double.NaN;
                return false;
            }

            bool b = double.TryParse(text, out value);

            if (!b)
                MessageBox.Show(text + " is not a valid number.");

            return b;
        }

        #endregion

        #endregion
        #region - ComboBoxes -

        /// <summary>
        /// Handles the SelectedIndexChanged over the cbAlgType.
        /// </summary>
        private void _cbAlgType_SelectedIndexChanged(object sender, EventArgs e)
        {
            int index = _cbAlgType.SelectedIndex;

            if (index >= 0)
            {
                if (_cbAlgType.Items.Count > 0)
                {
                    ITSPAlgorithm alg = (ITSPAlgorithm)_cbAlgType.Items[index];

                    // create a new instance of the same type
                    _algorithm = TSPAlgorithm.GetTSPAlgorithm(alg.Type);
                    _pgAlgProperties.SelectedObject = _algorithm;
                    _txtAlgDesc.Text = _algorithm.Description;
                }
            }
        }

        #endregion
        #region - Algorithm -

        /// <summary>
        /// Necessary delegate to perform calls asyncrhonously to functions.
        /// </summary>
        private delegate void FunctionDelegate();

        /// <summary>
        /// Handles the Ends event for the algorithm.
        /// </summary>
        private void _algorithm_Ends(object sender, EventArgs e)
        {
            if (InvokeRequired)
                this.BeginInvoke(new FunctionDelegate(WhenEnds));
            else
                WhenEnds();
        }

        /// <summary>
        /// Handles the NotSolutionFound event for the algorithm.
        /// </summary>
        private void _algorithm_NotSolutionFound(object sender, EventArgs e)
        {
            if (InvokeRequired)
                this.BeginInvoke(new FunctionDelegate(WhenNotSolutionFound));
            else
                WhenNotSolutionFound();
        }

        /// <summary>
        /// Function to be called when the algorithm ends.
        /// </summary>
        private void WhenEnds()
        {
            // to set a more expressive name to the legs
            for (int i = 1; i < _algorithm.LastResultPath.Length; i++)
            {
                // the legs are zero-index based, and the path are 1-index based
                // also, discount the airport position, that is the first one, inserted before starting the run
                int index = _algorithm.LastResultPath[i] - 2;

                AbstractFeatureType t = ListOfFires[index];
                _algorithm.LastResult[i].ID += "_" + t.Id;
            }

            // Route legs //////////////////////////////////////////////////
            for (int i = 1; i < _algorithm.LastResult.Length - 2; i++)
            {
                _algorithm.LastResult[i].Nextid = _algorithm.LastResult[i + 1].ID;
            }

            // Gets the flight plan with the last results
            FlightPlan fp = FPGenerator.GetDefaultFlightPlan(new Fix[] { }, _algorithm.LastResult);

            // Writes it to a file
            XMLFPWriter.Save(fp, PathToOutputFlightPlan);

            // GUI
            Status = "Done! Fligh Plan saved at " + PathToOutputFlightPlan;
            EnableControlsAfterRun();
            _algorithm.Ends             -= _algorithm_Ends;
            _algorithm.NotSolutionFound -= _algorithm_NotSolutionFound;
        }

        /// <summary>
        /// Function to be called when the algorithm ens with no solution.
        /// </summary>
        private void WhenNotSolutionFound()
        {
            // GUI
            Status = "" + _algorithm.ErrorMessage;
            MessageBox.Show(this, _algorithm.ErrorMessage, "Error", MessageBoxButtons.OK, MessageBoxIcon.Exclamation);
            EnableControlsAfterRun();
            _algorithm.Ends             -= _algorithm_Ends;
            _algorithm.NotSolutionFound -= _algorithm_NotSolutionFound;
        }

        #endregion

        #endregion

        #endregion
    }
}
