﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading;

using FlightPlan;
using Utils;

namespace FlightPlan.Optimizer.TSPAlgorithms
{
    /// <summary>
    /// Interface that all TSPAlgorithm classes must implement.
    /// </summary>
    public interface ITSPAlgorithm
    {
        #region - Events -

        /// <summary>
        /// Occurs when the algorithm starts.
        /// </summary>
        event EventHandler Starts;

        /// <summary>
        /// Occurs when the algorithm ends successfully.
        /// </summary>
        event EventHandler Ends;

        /// <summary>
        /// Occurs when the algorithm has ended without a possible solution found.
        /// </summary>
        event EventHandler NotSolutionFound;

        #endregion
        #region - Properties -

        /// <summary>
        /// Gets the Name for this Algorithm.
        /// </summary>
        String Name { get; }

        /// <summary>
        /// Gets a brief Description for this Algorithm.
        /// </summary>
        String Description { get; }

        /// <summary>
        /// Gets the TSP Algorithm Type.
        /// </summary>
        TSPAlgorithmType Type { get; }

        /// <summary>
        /// The current running thread.
        /// </summary>
        Thread RunningThread { get; }

        /// <summary>
        /// Gets the last Leg array resulting of running the algorithm.
        /// </summary>
        Leg[] LastResult { get; }

        /// <summary>
        /// Gets the last result path for the Algorithm.
        /// </summary>
        int[] LastResultPath { get; }

        /// <summary>
        /// Gets an error message if the execution fails.
        /// </summary>
        String ErrorMessage { get; }

        #endregion
        #region - Functions -

        /// <summary>
        /// Returns the sortest possible path according to the specified points.
        /// </summary>
        /// <param name="points">The points.</param>
        /// <returns>A sorted list of legs representing the best solution found.</returns>
        Leg[] GetSortedPath(Waypoint[] points);

        /// <summary>
        /// Calls the GetSortedPath function in a new thread.
        /// </summary>
        /// <param name="points">The points.</param>
        void GetSortedPathInNewThread(Waypoint[] points);

        #endregion
    }
}
