﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading;

using FlightPlan;
using Tsp;
using System.ComponentModel;

namespace FlightPlan.Optimizer.TSPAlgorithms
{
    #region - Enum TSPAlgorithmType -

    /// <summary>
    /// Enums the distinct types of TSP Algorithms.
    /// </summary>
    public enum TSPAlgorithmType
    {
        /// <summary>
        /// The Branch and Bound Algorithm Type.
        /// </summary>
        TypeBB,

        /// <summary>
        /// The Genetic Algorithm type.
        /// </summary>
        TypeGA
    }

    #endregion

    /// <summary>
    /// Abstract class that all TSPAlgorithm classes must extend.
    /// </summary>
    public abstract class TSPAlgorithm : ITSPAlgorithm
    {
        #region - Static -

        /// <summary>
        /// Gets the TSPAlgorithm object according to the specified type.
        /// </summary>
        /// <param name="type">The type of TSP Algorithm.</param>
        /// <returns>The TSP Algorithm.</returns>
        public static ITSPAlgorithm GetTSPAlgorithm(TSPAlgorithmType type)
        {
            switch (type)
            {
                case TSPAlgorithmType.TypeBB: return new TSPAlgorithm_BB();
                case TSPAlgorithmType.TypeGA: return new TSPAlgorithm_GA(); 
                default:
                    throw new ArgumentException("The type doesn't exist or isn't implemented.");                    
            }
        }

        #endregion

        #region - Events -

        /// <summary>
        /// Occurs when the algorithm starts.
        /// </summary>
        public event EventHandler Starts;

        /// <summary>
        /// Occurs when the algorithm ends successfully.
        /// </summary>
        public event EventHandler Ends;

        /// <summary>
        /// Occurs when the algorithm has ended without a possible solution found.
        /// </summary>
        public event EventHandler NotSolutionFound;

        #endregion
        #region - Properties -

        /// <summary>
        /// Gets the last Leg array resulting of running the algorithm.
        /// </summary>
        [Browsable(false)]
        public Leg[] LastResult
        {
            get;
            protected set;
        }

        /// <summary>
        /// Gets the last result path for the Algorithm.
        /// </summary>
        [Browsable(false)]
        public int[] LastResultPath
        {
            get;
            protected set;
        }

        /// <summary>
        /// The current running thread.
        /// </summary>
        [Browsable(false)]
        public Thread RunningThread 
        {
            get;
            private set; 
        }

        /// <summary>
        /// Gets an error message if the execution fails.
        /// </summary>
        [Browsable(false)]
        public String ErrorMessage
        {
            get;
            protected set;
        }

        #endregion
        #region - Public -

        /// <summary>
        /// Calls the GetSortedPath function in a new thread.
        /// </summary>
        /// <param name="points">The points.</param>
        public void GetSortedPathInNewThread(Waypoint[] points)
        {
            this.RunningThread = new Thread(new ParameterizedThreadStart(this.ThreadStartPoint));
            this.RunningThread.Start(points);
        }

        /// <summary>
        /// Auxiliar function to run the GetSortedPath throw a new Thread.
        /// </summary>
        private void ThreadStartPoint(Object o)
        {
            try
            {
                GetSortedPath((Waypoint[])o);
            }
            catch (Exception e)
            {
                ErrorMessage = e.Message;
                OnNotSolutionFound(new EventArgs());
            }
        }

        /// <summary>
        /// Gets the string representation for this Algorithm.
        /// Really, gets the Name, to set this object in ComboBox controls.
        /// </summary>
        /// <returns>The Name.</returns>
        public override string ToString()
        {
            return this.Name;
        }

        #endregion
        #region - Abstract -

        /// <summary>
        /// Gets the Name for this Algorithm.
        /// </summary>
        [Browsable(false)]
        public abstract String Name { get; }

        /// <summary>
        /// Gets a brief Description for this Algorithm.
        /// </summary>
        [Browsable(false)]
        public abstract String Description { get; }

        /// <summary>
        /// Gets the TSP Algorithm Type.
        /// </summary>
        [Browsable(false)]
        public abstract TSPAlgorithmType Type { get; }

        /// <summary>
        /// Returns the sortest possible path according to the specified points.
        /// </summary>
        /// <param name="points">The points</param>
        /// <returns></returns>
        public abstract Leg[] GetSortedPath(Waypoint[] points);

        #endregion
        #region - Protected -

        /// <summary>
        /// Creates a Tsp.Graph with the specified waypoints.
        /// </summary>
        /// <param name="points">The waypoints representing each graph node.</param>
        /// <returns>The resulting graph.</returns>
        protected Graph CreateGraph(Waypoint[] points)
        {
            int n = points.Length;

            Graph g = new Graph();
            // Add the nodes
            for (int i = 0; i <= n; i++)
                g.AddNode(i, n);
            g.nbNodes = n;

            // Add the links            
            for (int i = 0; i < n; i++)
            {
                Waypoint wp1 = points[i];

                for (int j = i + 1; j < n; j++)
                {
                    Waypoint wp2 = points[j];

                    double l = wp1.Distance(wp2); //// mirar qué valores da aquí
                    int ll   = (int)l;

                    g.AddLink(i + 1, j + 1, ll); // The ids must start at 1, not at 0
                }
            }

            g.ResetAdyM();
            return g;
        }

        /// <summary>
        /// Create the legList array, and add to each one the corresponding waypoint.
        /// </summary>
        /// <param name="points">The waypoints. Must be sorted.</param>
        /// <returns></returns>
        protected Leg[] CreateAndAddPoints(Waypoint[] points)
        {
            Leg[] legs = CreateLegs(points.Length);
            AddToLegs(legs, points);
            return legs;
        }

        #region - OnEvents -

        /// <summary>
        /// Raises the Starts event.
        /// </summary>
        protected virtual void OnStarts(EventArgs e)
        {
            if (Starts != null)
                Starts(this, e);
        }

        /// <summary>
        /// Raises the Ends event.
        /// </summary>
        protected virtual void OnEnds(EventArgs e)
        {
            if (Ends != null)
                Ends(this, e);
        }

        /// <summary>
        /// Raises the NotSolutionFound event.
        /// </summary>
        /// <param name="e"></param>
        protected virtual void OnNotSolutionFound(EventArgs e)
        {
            if (NotSolutionFound != null)
                NotSolutionFound(this, e);
        }

        #endregion

        #endregion
        #region - Private -

        /// <summary>
        /// Creates a legs array with the specified size.
        /// Also, respects the format that this array must have in a sequencial FlightPlan.
        /// </summary>
        /// <param name="count">The number of legs.</param>
        /// <returns>The legs array.</returns>
        private Leg[] CreateLegs(int count)
        {
            // we assume that exists at least 3 legs
            if (count < 3)
                throw new ArgumentException("Can't create less than 3 legs");

            Leg[] legList = new Leg[count];

            // Initial leg /////////////////////////////////////////////////
            legList[0] = new IFLeg(null, "depart", null, null, null, null);

            // Route legs //////////////////////////////////////////////////
            for (int i = 1; i < count - 1; i++)
            {
                String strCurrId = String.Format("tf{0}", i);
                String strNextId = String.Format("tf{0}", i + 1);
                if (i == count - 2) // last route stage leg
                {
                    legList[i] = new TFLeg(null, strCurrId, null, null, null, null);
                }
                else
                {
                    legList[i] = new TFLeg(null, strCurrId, null, strNextId, null, null);
                }
            }

            // Final leg ///////////////////////////////////////////////////////////////
            legList[count - 1] = new TFLeg(null, "arrive", null, null, null, null);

            return legList;
        }


        /// <summary>
        /// Set the specified points in the legs array.
        /// </summary>
        /// <param name="legs">The legs array.</param>
        /// <param name="points">The points array.</param>
        private void AddToLegs(Leg[] legs, Waypoint[] points)
        {
            for (int i = 0; i < legs.Length; i++)
            {
                legs[i].Destination = points[i];
            }
        }

        #endregion        
    }
}
