﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.ComponentModel;

using FlightPlan;
using Tsp;
using Tsp.BB;


namespace FlightPlan.Optimizer.TSPAlgorithms
{
    /// <summary>
    /// Tsp gui wrapper for the Branch and Bound algorithm.
    /// </summary>
    public class TSPAlgorithm_BB : TSPAlgorithm
    {
        #region - Constructors -

        /// <summary>
        /// Creates a new instance of TSPAlgorithm_BB.
        /// </summary>
        public TSPAlgorithm_BB()
        {
            this.Knn = 0;
        }

        #endregion
        #region - Properties -

        /// <summary>
        /// Gets the Name for this Algorithm.
        /// </summary>
        public override string Name
        {
            get { return "Branch And Bound"; }
        }

        /// <summary>
        /// Gets a brief Description for this Algorithm.
        /// </summary>
        public override string Description
        {
            get { return "Branch and Bound.\nProvides an optimal solution but expends too many time with more than 30 nodes. In such case, use the GA algorithm."; }
        }

        /// <summary>
        /// Gets the TSP Algorithm Type.
        /// </summary>
        public override TSPAlgorithmType Type
        {
            get { return TSPAlgorithmType.TypeBB; }
        }
        
        /// <summary>
        /// Number of closest edges to take for each node.
        /// 0 means all the edges.
        /// </summary>
        [DisplayName("NbKClosestEdgesToTake")]
        [Description("Number of closest edges to take for each node.\n0 means all the edges.")]
        [DefaultValue(0)]
        [Category("Options")]
        public int Knn { get; set; }

        #endregion
        #region - Public -

        /// <summary>
        /// Returns the sortest possible path according to the specified points.
        /// </summary>
        /// <param name="points">The points.</param>
        /// <returns>A sorted list of legs representing the best solution found.</returns>
        public override Leg[] GetSortedPath(Waypoint[] points)
        {
            OnStarts(new EventArgs());

            int n = points.Length;
            Graph g = CreateGraph(points);
            BranchAndBound bab = CreateBaB(g.nbNodes);
            bab.Run(g);

            if (State.BestSolution == null)
                throw new Exception("It doesn't exist possible solution with the supplied points and parameters.");


            // Put the waypoints in the order of the path
            Waypoint[] result = new Waypoint[n + 1];
            for (int i = 0; i < State.BestSolution.path.Length; i++)
            {
                int index = State.BestSolution.path[i] - 1; // the legs are zero-index based, and the path are 1-index based
                result[i] = points[index];
            }
            result[n] = result[0]; // last point is equals to the first one

            // Create the legs
            base.LastResult = CreateAndAddPoints(result);
            base.LastResultPath = State.BestSolution.path;

            OnEnds(new EventArgs());
            return base.LastResult;
        }

        #endregion
        #region - Private -

        /// <summary>
        /// Creates a BranchAndBound instance with the correct parameters to run the algorithm.
        /// </summary>
        /// <param name="nbNodes">The number of nodes in the graph.</param>
        /// <returns>A BranchAndBound instance.</returns>
        private BranchAndBound CreateBaB(int nbNodes)
        {
            BranchAndBound bab = new BranchAndBound(false);

            bab.branchingMethod             = BranchAndBound.BranchingMethod.State1;
            bab.nbKClosestEdgesToTake       = Knn;
            bab.usePolynomialInitialization = nbNodes < 16; // for values higher than 16 is better use GA initialization

            return bab;
        }

        #endregion
    }
}
