﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Text;

using FlightPlan;

using Tsp;
using Tsp.BB;

namespace FlightPlan.Optimizer.TSPAlgorithms
{
    /// <summary>
    /// Tsp gui wrapper for the Genetic Algorithm.
    /// </summary>
    public class TSPAlgorithm_GA : TSPAlgorithm
    {
        #region - Constructor -

        /// <summary>
        /// Creates a new instance of TSPAlgorithm_GA.
        /// </summary>
        public TSPAlgorithm_GA()
        {
            MaxGenerations = 300000;
        }

        #endregion
        #region - Properties -

        /// <summary>
        /// Gets the Name for this Algorithm.
        /// </summary>
        public override string Name
        {
            get { return "Genetic Algorithm"; }
        }

        /// <summary>
        /// Gets a brief Description for this Algorithm.
        /// </summary>
        public override string Description
        {
            get { return "Genetic Algorithm.\n" +
                         "Not always gives the optimal solution, but provides one near the best. "+
                         "Usefull when exists more than 30 nodes."; }
        }

        /// <summary>
        /// Gets the TSP Algorithm Type.
        /// </summary>
        public override TSPAlgorithmType Type
        {
            get { return TSPAlgorithmType.TypeGA; }
        }

        [DisplayName("MaxGenerations")]
        [Description("Number of generations that the algorithm will run.")]
        [DefaultValue(300000)]
        [Category("Options")]
        public int MaxGenerations 
        {
            get; 
            set;
        }

        #endregion        
        #region - Public -

        /// <summary>
        /// Returns the sortest possible path according to the specified points.
        /// </summary>
        /// <param name="points">The points.</param>
        /// <returns>A sorted list of legs representing the best solution found.</returns>
        public override Leg[] GetSortedPath(Waypoint[] points)
        {
            OnStarts(new EventArgs());

            int n   = points.Length;
            Graph g = CreateGraph(points);
            State s = Run(g);


            // Put the waypoints in the order of the path
            Waypoint[] result = new Waypoint[n + 1];
            for (int i = 0; i < s.path.Length; i++)
            {
                int index = s.path[i] - 1; // the legs are zero-index based, and the path are 1-index based
                result[i] = points[index];
            }
            result[n] = result[0]; // last point is equals to the first one
            
            // Create the legs
            base.LastResult = CreateAndAddPoints(result);
            base.LastResultPath = s.path;

            OnEnds(new EventArgs());

            return base.LastResult;
        }

        #endregion
        #region - Private -

        /// <summary>
        /// Run the Genetic Algorithm with the specified graph.
        /// </summary>
        /// <param name="g">The graph representing the nodes and links.</param>
        /// <returns>The best state (with its path) found.</returns>
        private State Run(Graph g)
        {
            State.TheGraph = g;
            State.N        = g.nbNodes; 

            Tsp.GA.TspGARunner runner = new Tsp.GA.TspGARunner();
            runner.maxGenerations = MaxGenerations;

            runner.BeginTsp(g.AsTspGaCities());

            if (runner.lastResult == null)
                throw new Exception("It doesn't exist possible solution with the supplied points and parameters.");


            State s = runner.lastResult.ToState();

            return s;
        }

        #endregion
    }
}
