﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Text;

namespace FlightPlan.Optimizer
{
    /// <summary>
    /// Provides utility methods.
    /// </summary>
    /// <author>Santiago Jimenez Serrano [sanjiser@gmail.com]</author>
    /// <date>19/05/2008</date>
    public class Utils
    {
        #region - FileNames -

        /// <summary>
        /// Removes the invalid chars, replacing them by a '_', in the specified file name.
        /// </summary>
        /// <param name="fileName">The file name to remove the invalid path chars.</param>
        /// <returns>A correct file name representation, or the same if it didn't contains any invalid char.</returns>
        public static String RemoveInvalidCharsFromFileName(String fileName)
        {
            char[] invalidChars = Path.GetInvalidFileNameChars();

            foreach (char c in invalidChars)
            {
                if (fileName.Contains(new string(c, 1)))
                {
                    fileName = fileName.Replace(c, '_');
                }
            }

            return fileName;
        }


        /// <summary>
        /// Returns a file name representation with the current and unique datetime representation included.
        /// </summary>
        /// <param name="directoryPath">The directory path.</param>
        /// <param name="fileName">The file name.</param>
        /// <param name="fileExtension">The file extension.</param>
        /// <returns>A file path with the next representation [directoryPath][fileName]_[date]_[time][fileExtension]</returns>
        public static String CreateUniqueDateTimeFileName(String directoryPath, String fileName, String fileExtension)
        {
            DateTime now = DateTime.Now;

            fileName += "_";

            // Ponemos el anyo
            fileName += now.Year.ToString();
            fileName += now.Month.ToString("00");
            fileName += now.Day.ToString("00") + "_";
            fileName += now.Hour.ToString("00");
            fileName += now.Minute.ToString("00");
            fileName += now.Second.ToString("00");
            fileName += now.Millisecond.ToString("000") + fileExtension;

            return Path.Combine(directoryPath, fileName);
        }

        /// <summary>
        /// Returns a file name representation with the current and unique datetime representation included.
        /// </summary>
        /// <param name="fullPath">A full path to a file, incuding directory, filename and extension.</param>
        /// <returns>A file path with the next representation [directoryPath][fileName]_[date]_[time][fileExtension]</returns>
        public static String CreateUniqueDateTimeFileName(String fullPath)
        {
            String filename  = Path.GetFileNameWithoutExtension(fullPath);
            String directory = Path.GetDirectoryName(fullPath);
            String extension = Path.GetExtension(fullPath);

            return CreateUniqueDateTimeFileName(directory, filename, extension);
        }

        /// <summary>
        /// Creates a unique file name with the specified parameters.
        /// </summary>
        /// <param name="directoryPath">The directory path.</param>
        /// <param name="fileName">The file name.</param>
        /// <param name="fileExtension">The file extension.</param>
        /// <param name="digits">A number between [1-9] of digits to create the identifier. E.g.. digits = 3, filename with identifier betwee [000-999]</param>
        /// <returns>A file path with the next representation [directoryPath][fileName]_[identifier with x digits][fileExtension]</returns>
        public static String CreateUniqueFileName(String directoryPath, String fileName, String fileExtension, uint digits)
        {
            if (digits == 0 || digits > 9)
                throw new ArgumentException("digits must be between [1-9]", "digits");

            int max = -1; ;
            string format = "";
            switch (digits)
            {
                case 1: max = 10;         format = "0";         break;
                case 2: max = 100;        format = "00";        break;
                case 3: max = 1000;       format = "000";       break;
                case 4: max = 10000;      format = "0000";      break;
                case 5: max = 100000;     format = "00000";     break;
                case 6: max = 1000000;    format = "000000";    break;
                case 7: max = 10000000;   format = "0000000";   break;
                case 8: max = 100000000;  format = "00000000";  break;
                case 9: max = 1000000000; format = "000000000"; break;
            }

            int i = 0;
            String aux = "";

            do
            {
                aux = Path.Combine(
                            directoryPath,
                                fileName +
                                "_" +
                                i.ToString(format) +
                                fileExtension);

            } while (File.Exists(aux) && ++i < max);

            if (i == max)
                throw new Exception("Can't produce an unique file name in the specified directory with the specified digits.");

            return aux;
        }

        /// <summary>
        /// Creates a unique file name with the specified parameters.
        /// </summary>
        /// <param name="directoryPath">The directory path.</param>
        /// <param name="fileName">The file name.</param>
        /// <param name="fileExtension">The file extension.</param>        
        /// <returns>A file path with the next representation [directoryPath][fileName]_[identifier with 5 digits][fileExtension]</returns>
        public static String CreateUniqueFileName(String directoryPath, String fileName, String fileExtension)
        {
            return CreateUniqueFileName(directoryPath, fileName, fileExtension, 5);
        }

        /// <summary>
        /// Creates a unique file name with the specified parameters.
        /// </summary>
        /// <param name="fullPath">A full path to a file, incuding directory, filename and extension.</param>
        /// <returns>A file path with the next representation [directoryPath][fileName]_[identifier with 5 digits][fileExtension]</returns>
        public static String CreateUniqueFileName(String fullPath)
        {
            String filename  = Path.GetFileNameWithoutExtension(fullPath);
            String directory = Path.GetDirectoryName(fullPath);
            String extension = Path.GetExtension(fullPath);

            return CreateUniqueFileName(directory, filename, extension);
        }

        #endregion
    }
}
