﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Globalization;
using System.IO;
using System.Windows.Forms;

namespace FlightPlan.Optimizer.Wfs.FeatureTypes
{
    public abstract class AbstractFeatureType
    {
        #region - Static -

        public static String[] GetListViewColumnNames(FeatureType type)
        {
            switch (type)
            {
                case FeatureType.Hotspots1Day:
                case FeatureType.Hotspots7Days:
                case FeatureType.HotspotsAll: return Hotspot.GetListViewColumnNames();
                case FeatureType.Fires7Days:
                case FeatureType.Fires30Days:
                case FeatureType.FiresAll:    return Fire.GetListViewColumnNames();
                default:
                    throw new ArgumentException("type");
            }
        }

        public static AbstractFeatureType CreateFeature(FeatureType type)
        {
            switch (type)
            {
                case FeatureType.Hotspots1Day:
                case FeatureType.Hotspots7Days:
                case FeatureType.HotspotsAll: return new Hotspot();
                case FeatureType.Fires7Days:
                case FeatureType.Fires30Days:
                case FeatureType.FiresAll:    return new Fire();
                default: throw new ArgumentException("type");
            }
        }

        public static AbstractFeatureType CreateFeature(FeatureType type, String csvLine)
        {
            switch (type)
            {
                case FeatureType.Hotspots1Day:
                case FeatureType.Hotspots7Days:
                case FeatureType.HotspotsAll: return new Hotspot(csvLine);
                case FeatureType.Fires7Days:
                case FeatureType.Fires30Days:
                case FeatureType.FiresAll:    return new Fire(csvLine);
                default: throw new ArgumentException("type");
            }
        }

        #endregion

        #region - Attributes -

        private   String _the_geom;
        protected double _the_geom_lat;
        protected double _the_geom_lon;

        #endregion
        #region - Constructor -

        /// <summary>
        /// Creates a new instance of AbstractFeatureType.
        /// </summary>
        public AbstractFeatureType()
        {
        }

        #endregion
        #region - Properties -

        public String Id       { get; set; }
        public String The_Geom 
        {
            get { return _the_geom; }
            set 
            {
                _the_geom = value;
                ResetTheGeom();
            }
        }

        public Waypoint The_Geom_Waypoint
        {
            get;
            private set;
        }

        public String Country  { get; set; }
        public String Province { get; set; }
        public String Commune  { get; set; }
        public String Class    { get; set; }

        #endregion
        #region - Public -

        /// <summary>
        /// Gets the Waypoint representation for this object.
        /// </summary>
        /// <returns>The Waypoint representation for this object.</returns>
        public virtual Waypoint ToWaypoint()
        {
            return new Waypoint(null, null, new GPSCoords(_the_geom_lat, true), new GPSCoords(_the_geom_lon, false), false, null);
        }

        /// <summary>
        /// Loads the attribute values for this object with the specified csv line.
        /// </summary>
        /// <param name="csvLine">A csv line containing the data.</param>
        public abstract void LoadFromCsvLine(String csvLine);

        /// <summary>
        /// Creates a ListViewItem representation for this object.
        /// </summary>
        /// <returns>The ListViewItem representation for this object</returns>
        public abstract ListViewItem ToListViewItem();

        #endregion
        #region - Protected -

        /// <summary>
        /// Set the Id and The_Geom property.
        /// </summary>
        /// <param name="id"></param>
        /// <param name="the_geom"></param>
        protected void SetProperties(String id, String the_geom)
        {
            this.Id       = id;
            this.The_Geom = the_geom;
        }

        /// <summary>
        /// Splits a csv line taking into account the "" symbols.
        /// </summary>
        /// <param name="csvLine">The csv line to split.</param>
        /// <returns>The csv line splitted.</returns>
        protected String[] SplitCsvLine(String csvLine)
        {
            // Puede que estemos matando moscas a cañonazos aquí
            bool insideQuotes            = false;
            bool existsCommaInsideQuotes = false;
            char aux;
            int n = csvLine.Length;

            for (int i = 0; i < n; i++)
            {
                aux = csvLine[i];
                if (aux == '"')
                    insideQuotes = !insideQuotes;
                else if (insideQuotes && aux == ',')
                    existsCommaInsideQuotes = true;
            }

            // case 1 - save time and memory
            if (!existsCommaInsideQuotes)
                return csvLine.Split(',');

            // case 2 - exists comma inside quotes
            insideQuotes = false;
            char[] cLine = csvLine.ToCharArray();
            for (int i = 0; i < cLine.Length; i++)
            {
                aux = cLine[i];
                if (aux == '"')
                    insideQuotes = !insideQuotes;
                else if (insideQuotes && aux == ',')
                {
                    cLine[i] = '#';
                }
            }

            String s = new String(cLine);
            String [] words = s.Split(',');
            for (int i = 0; i < words.Length; i++)
            {
                if (words[i].Contains('#'))
                {
                    words[i] = words[i].Replace('#', ',');
                }
            }

            return words;
        }

        #endregion
        #region - Private -

        private void ResetTheGeom()
        {
            // Example [POINT (38.81168675172818 21.147307924824414)]
            String s = _the_geom.Replace("POINT (", "");
            s = s.Replace(")", "");
            String[] words = s.Split(' ');

            _the_geom_lat = double.Parse(words[0], CultureInfo.InvariantCulture.NumberFormat);
            _the_geom_lon = double.Parse(words[1], CultureInfo.InvariantCulture.NumberFormat);

            this.The_Geom_Waypoint = new Waypoint(null, null, new GPSCoords(_the_geom_lat, true), new GPSCoords(_the_geom_lon, false), false, null);
        }

        #endregion
    }

    public class ListOfFeaturesTypes : List<AbstractFeatureType>
    {
        #region - Static -

        public ListOfFeaturesTypes CreateListOfFeatures(FeatureType type, string filename)
        {
            switch (type)
            {
                case FeatureType.Hotspots1Day:
                case FeatureType.Hotspots7Days:
                case FeatureType.HotspotsAll: return new ListOfHotspots(type, filename);
                case FeatureType.Fires7Days:
                case FeatureType.Fires30Days:
                case FeatureType.FiresAll: return new ListOfFires(type, filename);
                default: throw new ArgumentException("type");
            }
        }

        #endregion
        #region - Constructors -

        /// <summary>
        /// Empty constructor for ListOfFeaturesTypes.
        /// </summary>
        public ListOfFeaturesTypes()
        {
        }

        /// <summary>
        /// Creates a new ListOfFeaturesTypes with the data contained at the specified csvFilename.
        /// </summary>
        /// <param name="csvFilename">The Path to the csv filename.</param>
        public ListOfFeaturesTypes(FeatureType type, String csvFilename)
        {
            Load(type, csvFilename);
        }

        #endregion
        #region - Public -

        #region - Load -

        /// <summary>
        /// Loads the data from the specified file.
        /// </summary>
        /// <param name="csvFilename">The Path to the Csv filename.</param>
        public void Load(FeatureType type, String csvFilename)
        {
            this.Clear();
            String[] lines = File.ReadAllLines(csvFilename);

            // test if the header contains the correct data
            if (type == FeatureType.Hotspots1Day &&
                !lines[0].ToLower().Equals("fid,the_geom,hs_date,hs_time,conf,country,province,commune,corine,lat,lon"))
                throw new Exception("The header doesn't contain the correct data.\nBe sure that contains Hotspot1Day information.");

            else if ((type == FeatureType.Hotspots7Days || type == FeatureType.HotspotsAll) &&
                !lines[0].ToLower().Equals("fid,the_geom,hs_date,hs_time,conf,country,province,commune,corine,class,lat,lon"))
                throw new Exception("The header doesn't contain the correct data.\nBe sure that contains Hotspot7Days information.");

            else if (type == FeatureType.Fires7Days &&
                !lines[0].ToLower().Equals("fid,the_geom,firedate,country,province,commune,area_ha,countryful,x,y,lastupdate"))
                throw new Exception("The header doesn't contain the correct data.\nBe sure that contains Fire7Days information.");

            else if ((type == FeatureType.Fires30Days || type == FeatureType.FiresAll) &&
                !lines[0].ToLower().Equals("fid,the_geom,firedate,country,province,commune,area_ha,countryful,class,x,y,lastupdate"))
                throw new Exception("The header doesn't contain the correct data.\nBe sure that contains Fire30Days information.");

            // create and add the data
            for (int i = 1; i < lines.Length; i++)
            {
                this.Add(AbstractFeatureType.CreateFeature(type, lines[i]));
            }
        }

        #endregion
        #region - Filter Methods -

        /// <summary>
        /// Removes the FeatureType with the specified Id from this list.
        /// </summary>
        /// <param name="id">The id to search.</param>
        public void FilterId(String id)
        {
            for (int i = 0; i < this.Count; i++)
            {
                if (this[i].Id.Equals(id))
                {
                    this.RemoveAt(i);
                    return;
                }
            }
        }

        public void FilterByCountry(String country)
        {
            for (int i = 0; i < this.Count; i++)
            {
                if (!this[i].Country.Equals(country))
                {
                    this.RemoveAt(i);
                    i--;
                }
            }
        }

        public void FilterByProvince(String province)
        {
            for (int i = 0; i < this.Count; i++)
            {
                if (!this[i].Province.Equals(province))
                {
                    this.RemoveAt(i);
                    i--;
                }
            }
        }

        public void FilterByCommune(String commune)
        {
            for (int i = 0; i < this.Count; i++)
            {
                if (!this[i].Commune.Equals(commune))
                {
                    this.RemoveAt(i);
                    i--;
                }
            }
        }

        public void FilterByClass(String Class)
        {
            for (int i = 0; i < this.Count; i++)
            {
                if (!this[i].Class.Equals(Class))
                {
                    this.RemoveAt(i);
                    i--;
                }
            }
        }

        public void FilterByDate(DateTime dtIni, DateTime dtEnd)
        {
            for (int i = 0; i < this.Count; i++)
            {
                DateTime dt;

                if (this[i] is Hotspot)
                    dt = ((Hotspot)this[i]).HS_DateTime;
                else
                    dt = ((Fire)this[i]).FireDate;
                if (!(dt >= dtIni && dt <= dtEnd))
                {
                    this.RemoveAt(i);
                    i--;
                }
            }
        }

        /// <summary>
        /// Filter the featureTypes with date before the specified.
        /// </summary>
        public void FilterDateBefore(DateTime dt)
        {
            FilterByDate(dt.Date, new DateTime(2050, 1, 1));
        }

        /// <summary>
        /// Filter the featureTypes with date after the specified.
        /// </summary>
        /// <param name="dt"></param>
        public void FilterDateAfter(DateTime dt)
        {
            FilterByDate(new DateTime(1900, 1, 1), dt.Date);
        }

        public void FilterByBoundingBox(BoundingBox bb)
        {
            for (int i = 0; i < this.Count; i++)
            {
                if (!bb.IsInside(this[i].The_Geom_Waypoint))
                {
                    this.RemoveAt(i);
                    i--;
                }
            }
        }

        public void FilterByBoundingBox(double lat1, double lon1, double lat2, double lon2)
        {
            FilterByBoundingBox(new BoundingBox(lat1, lon1, lat2, lon2));
        }

         

        #endregion
        #region - Get -

        /// <summary>
        /// Creates the Waypoints array corresponding to this list.
        /// </summary>
        /// <returns>The Waypoints array corresponging to this list.</returns>
        public Waypoint[] GetWaypointsArray()
        {
            Waypoint[] points = new Waypoint[this.Count];

            for (int i = 0; i < this.Count; i++)
                points[i] = this[i].The_Geom_Waypoint;

            return points;
        }

        /// <summary>
        /// Creates the Waypoints array corresponding to this list.
        /// The first element will be the specified at the parameter.
        /// </summary>
        /// <param name="startWaypoint">The first element of the array. It mustn't be at this list.</param>
        /// <returns>The waypoints array corresponding to this list, with the startWaypoint as first element.</returns>
        public Waypoint[] GetWaypointsArray(Waypoint startWaypoint)
        {

            Waypoint[] points = new Waypoint[this.Count + 1];
            points[0] = startWaypoint;

            for (int i = 0; i < this.Count; i++)
                points[i+1] = this[i].The_Geom_Waypoint;

            return points;
        }

        #endregion

        #endregion
    }
}
