﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Text;
using System.Globalization;
using System.Windows.Forms;

namespace FlightPlan.Optimizer.Wfs.FeatureTypes
{
    public class Fire: AbstractFeatureType
    {
        #region - Static -

        public static String[] GetListViewColumnNames()
        {
            return new String[]
            {
                "Id",
                "Date",
                "Country",
                "Province",
                "Commune",
                "CountryFul",
                "Area_HA",
                "Class",
                "Lat",
                "Lon"
            };
        }

        #endregion
        #region - Constructors -

        /// <summary>
        /// Empty constructor for Fires7Days.
        /// </summary>
        public Fire()
        {
        }

        /// <summary>
        /// Creates a new instance of Fires7Days with the specified parameters.
        /// </summary>
        /// <param name="csvLine">The csv line.</param>
        public Fire(String csvLine)
        {
            LoadFromCsvLine(csvLine);
        }

        #endregion
        #region - Properties -

        // FireDate,Country,Province,Commune,Area_HA,CountryFul,X,Y,LastUpdate

        public DateTime FireDate   { get; set; }
        public double   Area_HA    { get; set; }
        public String   CountryFul { get; set; }
        public double   X          { get; set; }
        public double   Y          { get; set; }
        public DateTime LastUpdate { get; set; }

        #endregion
        #region - Public -

        /// <summary>
        /// Loads the attribute values for this object with the specified csv line.
        /// </summary>
        /// <param name="csvLine">A csv line containing the data.</param>
        public override void LoadFromCsvLine(String csvLine)
        {
            String[] words = SplitCsvLine(csvLine);

            if (words.Length == 11)
                SetParameters_Fires7Days(words);
            else if (words.Length == 12)
                SetParameters_Fires30Days(words);
            else
                throw new ArgumentException("csvLine not in correct format");
        }

        /// <summary>
        /// Creates a ListViewItem representation for this object.
        /// </summary>
        /// <returns>The ListViewItem representation for this object</returns>
        public override ListViewItem ToListViewItem()
        {
            String[] items = new String[10];
            items[0] = this.Id;
            items[1] = this.FireDate.ToShortDateString();
            items[2] = this.Country;
            items[3] = this.Province;
            items[4] = this.Commune;
            items[5] = this.CountryFul;
            items[6] = this.Area_HA.ToString();
            items[7] = this.Class;
            items[8] = this._the_geom_lat.ToString();
            items[9] = this._the_geom_lon.ToString();

            ListViewItem lvi = new ListViewItem(items);
            lvi.Tag = this;
            return lvi;
        }

        #endregion
        #region - Private -

        private void SetParameters_Fires7Days(String[] words)
        {
            // FID,the_geom,FireDate,Country,Province,Commune,Area_HA,CountryFul,X,Y,LastUpdate

            base.SetProperties(words[0], words[1]);
            this.FireDate   = Create_Date(words[2]);
            this.Country    = words[3];
            this.Province   = words[4];
            this.Commune    = words[5];
            this.Area_HA    = double.Parse(words[6], CultureInfo.InvariantCulture.NumberFormat);
            this.CountryFul = words[7];
            this.X          = double.Parse(words[8], CultureInfo.InvariantCulture.NumberFormat);
            this.Y          = double.Parse(words[9], CultureInfo.InvariantCulture.NumberFormat);
            this.LastUpdate = Create_Date(words[10]);
        }

        private void SetParameters_Fires30Days(String[] words)
        {
            // FID,the_geom,FireDate,Country,Province,Commune,Area_HA,CountryFul,Class,X,Y,LastUpdate
            // same than above but adding Class

            base.SetProperties(words[0], words[1]);
            this.FireDate   = Create_Date(words[2]);
            this.Country    = words[3];
            this.Province   = words[4];
            this.Commune    = words[5];
            this.Area_HA    = double.Parse(words[6], CultureInfo.InvariantCulture.NumberFormat);
            this.CountryFul = words[7];
            this.Class      = words[8];
            this.X          = double.Parse(words[9],  CultureInfo.InvariantCulture.NumberFormat);
            this.Y          = double.Parse(words[10], CultureInfo.InvariantCulture.NumberFormat);
            this.LastUpdate = Create_Date(words[11]);
        }

        /// <summary>
        /// Creates a new DateTime with the specified parameters.
        /// </summary>
        /// <param name="date">The date in format dd/MM/yyyy</param>
        private DateTime Create_Date(String date)
        {
            String[] date_words = date.Split('-');

            int year  = int.Parse(date_words[2]);
            int month = int.Parse(date_words[1]);
            int day   = int.Parse(date_words[0]);

            return new DateTime(year, month, day);
        }

        #endregion        
    }

    public class ListOfFires : ListOfFeaturesTypes
    {
        #region - Constructors -

        /// <summary>
        /// Empty constructor for ListOfFires.
        /// </summary>
        public ListOfFires()
        {
        }

        /// <summary>
        /// Creates a new ListOfFires with the data contained at the specified csvFilename.
        /// </summary>
        /// <param name="csvFilename">The Path to the csv filename.</param>
        public ListOfFires(FeatureType type, String csvFilename)
        {
            Load(type, csvFilename);
        }

        #endregion
        #region - Public -

        public void FilterByCountryFul(String countryFul)
        {
            for (int i = 0; i < this.Count; i++)
            {
                if (!((Fire)this[i]).CountryFul.Equals(countryFul))
                {
                    this.RemoveAt(i);
                    i--;
                }
            }
        }

        public void FilterByArea_HA(int minArea, int maxArea)
        {
            for (int i = 0; i < this.Count; i++)
            {
                double area = ((Fire)this[i]).Area_HA;
                if (area < minArea || area > maxArea)
                {
                    this.RemoveAt(i);
                    i--;
                }
            }
        }

        #endregion
    }
}
