﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Text;
using System.Globalization;
using System.Windows.Forms;

namespace FlightPlan.Optimizer.Wfs.FeatureTypes
{
    public class Hotspot : AbstractFeatureType
    {
        #region - Static -

        public static String[] GetListViewColumnNames()
        {
            return new String[]
            {
                "Id",
                "Date",
                "Conf",
                "Country",
                "Province",
                "Commune",
                "Corine",
                "Class",
                "Lat",
                "Lon"
            };
        }

        #endregion
        #region - Constructor -

        /// <summary>
        /// Empty constructor for Hotspots1Day.
        /// </summary>
        public Hotspot()
        {
        }

        /// <summary>
        /// Creates a new instance of Hotspot1Day with the specified parameters.
        /// </summary>
        /// <param name="csvLine">The csv line.</param>
        public Hotspot(String csvLine)
            :this()
        {
            LoadFromCsvLine(csvLine);
        }

        #endregion
        #region - Properties -
        
        public DateTime HS_DateTime { get; set; }
        public int      Conf        { get; set; }
        public String   Corine      { get; set; }
        public double   Lat         { get; set; }
        public double   Lon         { get; set; }

        #endregion
        #region - Public -

        /// <summary>
        /// Loads the attribute values for this object with the specified csv line.
        /// </summary>
        /// <param name="csvLine">A csv line containing the data.</param>
        public override void LoadFromCsvLine(String csvLine)
        {
            String[] words = SplitCsvLine(csvLine);

            if (words.Length == 11)
                SetParameters_Hotspot1Day(words);
            else if (words.Length == 12)
                SetParameters_Hotspot7Days(words);
            else
                throw new ArgumentException("csvLine not in correct format");
        }

        /// <summary>
        /// Creates a ListViewItem representation for this object.
        /// </summary>
        /// <returns>The ListViewItem representation for this object</returns>
        public override ListViewItem ToListViewItem()
        {
            String[] items = new String[10];
            items [0] = this.Id;
            items [1] = this.HS_DateTime.ToShortDateString();
            items [2] = this.Conf.ToString();
            items [3] = this.Country;
            items [4] = this.Province;
            items [5] = this.Commune;
            items [6] = this.Corine;
            items [7] = this.Class;
            items [8] = this.Lat.ToString();
            items [9] = this.Lon.ToString();

            ListViewItem lvi = new ListViewItem(items);
            lvi.Tag = this;
            return lvi;
        }

        #endregion
        #region - Private -

        private void SetParameters_Hotspot1Day(String[] words)
        {
            // FID,the_geom,HS_Date,HS_Time,Conf,Country,Province,Commune,Corine,Lat,Lon

            base.SetProperties(words[0], words[1]);
            Set_HS_DateTime(words[2], words[3]);
            this.Conf     = int.Parse(words[4]);
            this.Country  = words[5];
            this.Province = words[6];
            this.Commune  = words[7];
            this.Corine   = words[8];
            this.Lat      = double.Parse(words[9], CultureInfo.InvariantCulture.NumberFormat);
            this.Lon      = double.Parse(words[10], CultureInfo.InvariantCulture.NumberFormat);
        }

        private void SetParameters_Hotspot7Days(String[] words)
        {
            // FID,the_geom,HS_Date,HS_Time,Conf,Country,Province,Commune,Corine,CLASS,Lat,Lon
            // the same than above, but adding CLASS

            base.SetProperties(words[0], words[1]);
            Set_HS_DateTime(words[2], words[3]);
            this.Conf     = int.Parse(words[4]);
            this.Country  = words[5];
            this.Province = words[6];
            this.Commune  = words[7];
            this.Corine   = words[8];
            this.Class    = words[9];
            this.Lat      = double.Parse(words[10], CultureInfo.InvariantCulture.NumberFormat);
            this.Lon      = double.Parse(words[11], CultureInfo.InvariantCulture.NumberFormat);
        }

        /// <summary>
        /// Set the HS_DateTime according the specified parameters.
        /// </summary>
        /// <param name="hs_date">The date in format yyyy/MM/dd</param>
        /// <param name="hs_hour">The hour en format hh:mm</param>
        private void Set_HS_DateTime(String hs_date, String hs_hour)
        {
            String[] date_words = hs_date.Split('-');
            String[] hour_words = hs_hour.Split(':');

            int year  = int.Parse(date_words[0]);
            int month = int.Parse(date_words[1]);
            int day   = int.Parse(date_words[2]);
            int hour  = int.Parse(hour_words[0]);
            int min   = int.Parse(hour_words[1]);

            HS_DateTime = new DateTime(year, month, day, hour, min, 0);
        }

        #endregion
    }

    public class ListOfHotspots : ListOfFeaturesTypes
    {
        #region - Constructors -

        /// <summary>
        /// Empty constructor for ListOfHotspots.
        /// </summary>
        public ListOfHotspots()
        {
        }

        /// <summary>
        /// Creates a new ListOfHotspots with the data contained at the specified csvFilename.
        /// </summary>
        /// <param name="csvFilename">The Path to the csv filename.</param>
        public ListOfHotspots(FeatureType type, String csvFilename)
        {
            Load(type, csvFilename);
        }

        #endregion
        #region - Public -

        public void FilterByCorine(String corine)
        {
            for (int i = 0; i < this.Count; i++)
            {
                if (!((Hotspot)this[i]).Corine.Equals(corine))
                {
                    this.RemoveAt(i);
                    i--;
                }
            }
        }

        public void FilterByConf(int minConf, int maxConf)
        {
            for (int i = 0; i < this.Count; i++)
            {
                int conf = ((Hotspot)this[i]).Conf;
                if (conf < minConf || conf > maxConf)
                {
                    this.RemoveAt(i);
                    i--;
                }
            }
        }

        #endregion
    }
}
