
using System;
using System.IO;
using System.Collections;
using System.Collections.Generic;

using Utils;
using System.Xml;

//-----------------------------------------------------------------------------
// Implements specialized leg types derived from base class.
// Steps in order to implement a new leg type:
//   (1) Add leg type to XML schema definition.
//   (2) Define leg subclass.
//   (3) Update XMLFPRdr so that it is able to read the new leg type.
//   (4) Check if Stage.UpdateGraph needs to be modified. 
//   (5) Implement subclass methods.
//
// Implemented legs:
//   * IF: Initial Fix
//   * TF: Track to a Fix
//   * DF: Direct to a Fix
//   * RF: Radius to a Fix
//   * IL: Iterative leg (partially)
//   * Intersection Leg (partially)
//   * Holding to a Fix
//   * Parametric Leg
//       - BasicScanLeg
//   * LA: Land Leg
//   * TO: Take-off Leg
// Pending:
//   * Holding to an Altitude
//   * Holding to an Condition	
//-----------------------------------------------------------------------------

namespace FlightPlan
{
    //-------------------------------------------------------------------------
    /// <summary>
    /// Initial Ramp Fix (IRF) leg class. 
    /// </summary>
    //-------------------------------------------------------------------------
    public class IRFLeg : Leg
    {
        #region - Attributes & Properties -

        /// <summary>
        /// Azimuth angle which determines the orientation of the ramp. 
        /// </summary>
        public double course;
        public double Course
        {
            get { return course; }
            set { course = value; }
        }

        /// <summary>
        /// Length which determines the North size of the ramp. 
        /// </summary>
        public double dimN;
        public double DimN
        {
            get { return dimN; }
            set { course = dimN; }
        }

        /// <summary>
        /// Length which determines the South size of the ramp. 
        /// </summary>
        public double dimS;
        public double DimS
        {
            get { return dimS; }
            set { course = dimS; }
        }

        /// <summary>
        /// Length which determines the East size of the ramp. 
        /// </summary>
        public double dimE;
        public double DimE
        {
            get { return dimE; }
            set { course = dimE; }
        }

        /// <summary>
        /// Length which determines the West size of the ramp. 
        /// </summary>
        public double dimW;
        public double DimW
        {
            get { return dimW; }
            set { course = dimW; }
        }

        #endregion
        #region - Constructors -

        //---------------------------------------------------------------------
        /// <summary>
        /// Constructor. 
        /// </summary>
        //---------------------------------------------------------------------
        public IRFLeg(Stage stg, string id, Waypoint destination, string nextid,
            string previd, ContingencyList emergency, double course, double dimN, double dimS, double dimE, double dimW)
            : base(stg, id, destination, nextid, previd, emergency) 
        {

            this.course = course;
            this.dimN = dimN;
            this.dimS = dimS;
            this.dimE = dimE;
            this.dimW = dimW;
        }

        #endregion
        #region - Public Functions -

        /// <summary>
        /// Returns deep copy of IRF leg object.
        /// </summary>
        /// <param name="s"></param>
        /// <returns></returns>
        public override Leg DeepCopy(Stage s)
        {
            Leg l = base.DeepCopy(s);
            return new IRFLeg(l.stage, l.id, l.destination, l.nextid, l.previd, l.emergency, this.course, this.dimN, this.dimS, this.dimE, this.dimW);
        }

        #endregion
        #region - Xml -

        /// <summary>
        /// Gets the XsiType for this Leg.
        /// </summary>
        protected override String XsiType
        {
            get { return "fp:IRFLeg"; }
        }

        /// <summary>
        /// Appends this Object to a xml LegsCollection representation.
        /// </summary>
        /// <param name="legsElement">The LegsCollection XmlElement representation where to append this Object.</param>
        public override XmlElement AppendXml(XmlElement legsElement)
        {
            XmlElement xmlLeg = base.AppendXml(legsElement);

            XmlHelper.AddElement(xmlLeg, "dimN", DimN.ToString());
            XmlHelper.AddElement(xmlLeg, "dimS", DimS.ToString());
            XmlHelper.AddElement(xmlLeg, "dimE", DimE.ToString());
            XmlHelper.AddElement(xmlLeg, "dimW", DimW.ToString());

            return xmlLeg;
        }

        #endregion						
    }


    //-------------------------------------------------------------------------
    /// <summary>
    /// Landing (LA) leg class. 
    /// </summary>
    //-------------------------------------------------------------------------
    public class LALeg : Leg
    {
        #region - Attributes & Properties -

        /// <summary>
        /// Runway head waypoint. 
        /// </summary>
        public Waypoint rwyhead;
        public Waypoint Rwyhead
        {
            get { return rwyhead; }
            set { rwyhead = value; }
        }

        /// <summary>
        /// Azimuth angle which determines the orientation of the runway. 
        /// </summary>
        public double course;
        public double Course
        {
            get { return course; }
            set { course = value; }
        }

        /// <summary>
        /// Integer value that determines the length of the runway. 
        /// </summary>
        public double length;
        public double Length
        {
            get { return length; }
            set { length = value; }
        }

        /// <summary>
        /// Holding waypoint 1. 
        /// </summary>
        public Waypoint hwp1;
        public Waypoint Hwp1
        {
            get { return hwp1; }
            set { hwp1 = value; }
        }

        /// <summary>
        /// Holding waypoint 2. 
        /// </summary>
        public Waypoint hwp2;
        public Waypoint Hwp2
        {
            get { return hwp2; }
            set { hwp2 = value; }
        }

        /// <summary>
        /// Holding waypoint 3. 
        /// </summary>
        public Waypoint hwp3;
        public Waypoint Hwp3
        {
            get { return hwp3; }
            set { hwp3 = value; }
        }

        /// <summary>
        /// Holding waypoint 4. 
        /// </summary>
        public Waypoint hwp4;
        public Waypoint Hwp4
        {
            get { return hwp4; }
            set { hwp4 = value; }
        }

        /// <summary>
        /// Integration waypoint. 
        /// </summary>
        public Waypoint iwp;
        public Waypoint Iwp
        {
            get { return iwp; }
            set { iwp = value; }
        }

        /// <summary>
        /// Initial downwind waypoint. 
        /// </summary>
        public Waypoint idwp;
        public Waypoint Idwp
        {
            get { return idwp; }
            set { idwp = value; }
        }

        /// <summary>
        /// Initial base waypoint. 
        /// </summary>
        public Waypoint ibwp;
        public Waypoint Ibwp
        {
            get { return ibwp; }
            set { ibwp = value; }
        }

        /// <summary>
        /// Initial final waypoint. 
        /// </summary>
        public Waypoint ifwp;
        public Waypoint Ifwp
        {
            get { return ifwp; }
            set { ifwp = value; }
        }


        /// <summary>
        /// Touch down waypoint. 
        /// </summary>
        public Waypoint tdwp;
        public Waypoint Tdwp
        {
            get { return tdwp; }
            set { tdwp = value; }
        }


        /// <summary>
        /// Break point down waypoint. 
        /// </summary>
        public Waypoint bpwp;
        public Waypoint Bpwp
        {
            get { return bpwp; }
            set { bpwp = value; }
        }

        /// <summary>
        /// Landing abort waypoint. 
        /// </summary>
        public Waypoint lawp;
        public Waypoint Lawp
        {
            get { return lawp; }
            set { lawp = value; }
        }

        /// <summary>
        /// End of abort waypoint. 
        /// </summary>
        public Waypoint eawp;
        public Waypoint Eawp
        {
            get { return eawp; }
            set { eawp = value; }
        }


        /// <summary>
        /// Condition to be applied in order to decide if FP should be followed after land abort.
        /// </summary>
        public Condition dvrtFrust;
        public Condition DvrtFrust
        {
            get { return dvrtFrust; }
            set { dvrtFrust = value; }
        }

        /// <summary>
        /// Next abort leg.
        /// </summary>
        public Leg nextAbort;
        public Leg NextAbort
        {
            get { return nextAbort; }
            set { nextAbort = value; }
        }

        /// <summary>
        /// Next abort leg id.
        /// </summary>
        public String nextAbortId;
        public String NextAbortId
        {
            get { return nextAbortId; }
            set { nextAbortId = value; }
        }

        /// <summary>
        /// Next non-abort leg id.
        /// </summary>
        public String nextNonAbortId;
        public String NextNonAbortId
        {
            get { return nextNonAbortId; }
            set { nextNonAbortId = value; }
        }

        /// <summary>
        /// Condition to be applied in order to decide the initial leg of the landing procedure.
        /// </summary>
        public Condition startOP;
        public Condition StartOP
        {
            get { return startOP; }
            set { startOP = value; }
        }

        #endregion
        #region - Constructors -

        /// <summary>
        /// Creates a LALeg object.
        /// </summary>
        /// <param name="stg"><see cref="Stage"/> the leg belongs to.</param>
        /// <param name="id">A <see cref="System.String"/> with the leg id.</param>
        public LALeg(Stage stg, string id, Waypoint destination, string nextid,
                      string previd, ContingencyList emergency, Waypoint rwyhead, double course,
                      double length, Waypoint hwp1, Waypoint hwp2, Waypoint hwp3,
                      Waypoint hwp4, Waypoint iwp, Waypoint idwp, Waypoint ibwp, Waypoint ifwp, Waypoint tdwp,
                      Waypoint bpwp, Waypoint lawp, Waypoint eawp, String nextAbortId, String dvrtFrust, String startOP)
            : base(stg, id, destination, nextid, previd, emergency)
        {
            this.rwyhead = rwyhead;
            this.course = course;
            this.length = length;
            this.hwp1 = hwp1;
            this.hwp2 = hwp2;
            this.hwp3 = hwp3;
            this.hwp4 = hwp4;
            this.iwp = iwp;
            this.idwp = idwp;
            this.ibwp = ibwp;
            this.ifwp = ifwp;
            this.tdwp = tdwp;
            this.bpwp = bpwp;
            this.lawp = lawp;
            this.eawp = eawp;
            this.nextAbortId = nextAbortId;
            this.nextNonAbortId = this.nextid;

            if (dvrtFrust != null)
            {
                this.dvrtFrust = new Condition(id+"dvtFR");

                if (dvrtFrust == "TRUE")
                {
                    this.destination = this.eawp;
                    this.dvrtFrust.SetResult(1);
                    this.nextid = this.nextAbortId;
                }
                else
                if (dvrtFrust == "FALSE")
                {
                    this.destination = this.bpwp;
                    this.dvrtFrust.SetResult(0);
                    this.nextid = this.nextNonAbortId;
                }
            } 
            else
                throw new Exception("Unknown divert to frustrated condition.");
            
            if (startOP != null){
                this.startOP = new Condition(id+"startOP");

                if (startOP == "HOLD")
                    this.startOP.SetResult(0);
                else
                if (startOP == "DOWNWIND")
                    this.startOP.SetResult(1);
                else
                if (startOP == "BASE")
                    this.startOP.SetResult(2);
                else
                if (startOP == "FINAL")
                    this.startOP.SetResult(3);
                else
                    throw new Exception("Invalid start of landing operation condition: "+ startOP);
            }
            else 
            {
                this.startOP = new Condition(id + "startOP");
                this.startOP.SetResult(0);
            }
        }

        #endregion
        #region - Public Functions -

        /// <summary>
        /// Returns deep copy of LA leg
        /// </summary>
        /// <param name="s"></param>
        /// <returns></returns>
        public override Leg DeepCopy(Stage s)
        {
            Leg l = base.DeepCopy(s);
            String startOP, dvrtFrust;

                 if (this.startOP.GetResult() == 0) startOP = "HOLD";
            else if (this.startOP.GetResult() == 1) startOP = "DOWNWIND";
            else if (this.startOP.GetResult() == 2) startOP = "BASE";
            else if (this.startOP.GetResult() == 3) startOP = "FINAL";
            else throw new Exception("Internal error on the startOP condition.");

                 if (this.dvrtFrust.GetResult() == 0) dvrtFrust = "FALSE";
            else if (this.dvrtFrust.GetResult() == 1) dvrtFrust = "TRUE";
            else throw new Exception("Internal error on the dvrtFrust condition.");

            LALeg lal = new LALeg(l.stage, l.id, l.destination, l.nextid, l.previd,
                l.emergency, this.rwyhead, this.course, this.length, this.hwp1, this.hwp2, this.hwp3,
                this.hwp4, this.iwp, this.idwp, this.ibwp, this.ifwp, this.tdwp, this.bpwp, this.lawp, this.eawp,
                this.nextAbortId, dvrtFrust, startOP);
            return lal;
        }

        #endregion
        #region - Xml -

        /// <summary>
        /// Gets the XsiType for this Leg.
        /// </summary>
        protected override String XsiType
        {
            get { return "fp:LALeg"; }
        }

        /// <summary>
        /// Appends this Object to a xml LegsCollection representation.
        /// </summary>
        /// <param name="legsElement">The LegsCollection XmlElement representation where to append this Object.</param>
        public override XmlElement AppendXml(XmlElement legsElement)
        {
            XmlElement xmlLeg = base.AppendXml(legsElement);

            XmlHelper.AddElement(xmlLeg, "length", this.Length.ToString());
            XmlHelper.AddElement(xmlLeg, "course", this.Course.ToString());

            XmlElement aux;

            if (Rwyhead != null)
            {
                aux = XmlHelper.AddElement(xmlLeg, "rwyhead", null);
                Rwyhead.AppendXml(aux);
            }

            if (Hwp1 != null)
            {
                aux = XmlHelper.AddElement(xmlLeg, "hwp1", null);
                Hwp1.AppendXml(aux);
            }

            if (Hwp2 != null)
            {
                aux = XmlHelper.AddElement(xmlLeg, "hwp2", null);
                Hwp2.AppendXml(aux);
            }

            if (Hwp3 != null)
            {
                aux = XmlHelper.AddElement(xmlLeg, "hwp3", null);
                Hwp3.AppendXml(aux);
            }

            if (Hwp4 != null)
            {
                aux = XmlHelper.AddElement(xmlLeg, "hwp4", null);
                Hwp4.AppendXml(aux);
            }


            if (Iwp != null)
            {
                aux = XmlHelper.AddElement(xmlLeg, "iwp", null);
                Iwp.AppendXml(aux);
            }
            if (Idwp != null)
            {
                aux = XmlHelper.AddElement(xmlLeg, "idwp", null);
                Idwp.AppendXml(aux);
            }
            if (Ibwp != null)
            {
                aux = XmlHelper.AddElement(xmlLeg, "ibwp", null);
                Ibwp.AppendXml(aux);
            }
            if (Ifwp != null)
            {
                aux = XmlHelper.AddElement(xmlLeg, "ifwp", null);
                Ifwp.AppendXml(aux);
            }


            if (Tdwp != null)
            {
                aux = XmlHelper.AddElement(xmlLeg, "tdwp", null);
                Tdwp.AppendXml(aux);
            }
            if (Bpwp != null)
            {
                aux = XmlHelper.AddElement(xmlLeg, "bpwp", null);
                Bpwp.AppendXml(aux);
            }
            if (Lawp != null)
            {
                aux = XmlHelper.AddElement(xmlLeg, "lawp", null);
                Lawp.AppendXml(aux);
            }
            if (Eawp != null)
            {
                aux = XmlHelper.AddElement(xmlLeg, "eawp", null);
                Eawp.AppendXml(aux);
            }

            XmlHelper.AddElement(xmlLeg, "nextFrst", this.NextAbortId);
            if(DvrtFrust != null) XmlHelper.AddElement(xmlLeg, "dvrt-frust", this.DvrtFrust.CondId);
            if(StartOP   != null) XmlHelper.AddElement(xmlLeg, "start-op", this.startOP.CondId);

            return xmlLeg;
        }

        #endregion						
    }


    //-------------------------------------------------------------------------
    /// <summary>
    /// Take-off (TO) leg class. 
    /// </summary>
    //-------------------------------------------------------------------------
    public class TOLeg : Leg
    {
        #region - Attributes & Properties -

        /// <summary>
        /// Runway head waypoint. 
        /// </summary>
        public Waypoint rwyhead;
        public Waypoint Rwyhead
        {
            get { return rwyhead; }
            set { rwyhead = value; }
        }

        /// <summary>
        /// Azimuth angle which determines the orientation of the runway. 
        /// </summary>
        public double course;
        public double Course
        {
            get { return course; }
            set { course = value; }
        }

        /// <summary>
        /// Integer value that determines the length of the runway. 
        /// </summary>
        public double length;
        public double Length
        {
            get { return length; }
            set { length = value; }
        }

        /// <summary>
        /// Departure End of Runway. 
        /// </summary>
        public Waypoint der;
        public Waypoint Der
        {
            get { return der; }
            set { der = value; }
        }

        /// <summary>
        /// End of Departure waypoint A. 
        /// </summary>
        public Waypoint edwp_a;
        public Waypoint Edwp_a
        {
            get { return edwp_a; }
            set { edwp_a = value; }
        }

        /// <summary>
        /// End of Departure waypoint A-min. 
        /// </summary>
        public Waypoint edwp_a_min;
        public Waypoint Edwp_a_min
        {
            get { return edwp_a_min; }
            set { edwp_a_min = value; }
        }

        /// <summary>
        /// End of Departure waypoint BL.
        /// </summary>
        public Waypoint edwp_bl;
        public Waypoint Edwp_bl
        {
            get { return edwp_bl; }
            set { edwp_bl = value; }
        }

        /// <summary>
        /// End of Departure waypoint BL-min. 
        /// </summary>
        public Waypoint edwp_bl_min;
        public Waypoint Edwp_bl_min
        {
            get { return edwp_bl_min; }
            set { edwp_bl_min = value; }
        }

        /// <summary>
        /// End of Departure waypoint BR. 
        /// </summary>
        public Waypoint edwp_br;
        public Waypoint Edwp_br
        {
            get { return edwp_br; }
            set { edwp_br = value; }
        }

        /// <summary>
        /// End of Departure waypoint BR-min. 
        /// </summary>
        public Waypoint edwp_br_min;
        public Waypoint Edwp_br_min
        {
            get { return edwp_br_min; }
            set { edwp_br_min = value; }
        }

        /// <summary>
        /// End of Departure waypoint CL. 
        /// </summary>
        public Waypoint edwp_cl;
        public Waypoint Edwp_cl
        {
            get { return edwp_cl; }
            set { edwp_cl = value; }
        }


        /// <summary>
        /// End of Departure waypoint CR. 
        /// </summary>
        public Waypoint edwp_cr;
        public Waypoint Edwp_cr
        {
            get { return edwp_cr; }
            set { edwp_cr = value; }
        }

        /// <summary>
        /// Runway departure area width. 
        /// </summary>
        public double rda_w;
        public double Rda_w
        {
            get { return rda_w; }
            set { rda_w = value; }
        }

        /// <summary>
        /// Runway departure area length. 
        /// </summary>
        public double rda_l;
        public double Rda_l
        {
            get { return rda_l; }
            set { rda_l = value; }
        }

        /// <summary>
        /// Collection of possible next legs when going forward.
        /// </summary>
        public LegsCollection nextList;
        public LegsCollection NextList
        {
            get { return nextList; }
            set { nextList = value; }
        }

        /// <summary>
        /// Next legs as a list of references.
        /// </summary>
        public ArrayList nextRefs = null;
        public ArrayList NextRefs
        {
            get { return nextRefs; }
            set { nextRefs = value; }
        }

        /// <summary>
        /// Condition to be applied in order to decide the final leg of the landing procedure.
        /// </summary>
        public Condition endOP;
        public Condition EndOP
        {
            get { return endOP; }
            set { endOP = value; }
        }

        #endregion
        #region - Constructors -

        /// <summary>
        /// Creates a TOLeg object.
        /// </summary>
        /// <param name="stg"><see cref="Stage"/> the leg belongs to.</param>
        /// <param name="id">A <see cref="System.String"/> with the leg id.</param>
        public TOLeg(Stage stg, string id, Waypoint destination, string nextid,
                      string previd, ContingencyList emergency, Waypoint rwyhead, double course,
                      double length, Waypoint der, Waypoint edwp_a, Waypoint edwp_a_min,
                      Waypoint edwp_bl, Waypoint edwp_bl_min, Waypoint edwp_br, Waypoint edwp_br_min, 
                      Waypoint edwp_cl, Waypoint edwp_cr, double rda_w, double rda_l,
                      ArrayList nextRefs,  String endOP)
            : base(stg, id, destination, nextid, previd, emergency)
        {
            this.rwyhead = rwyhead;
            this.course = course;
            this.length = length;
            this.der = der;
            this.edwp_a = edwp_a;
            this.edwp_a_min = edwp_a_min;
            this.edwp_bl = edwp_bl;
            this.edwp_bl_min = edwp_bl_min;
            this.edwp_br = edwp_br;
            this.edwp_br_min = edwp_br_min;
            this.edwp_cl = edwp_cl;
            this.edwp_cr = edwp_cr;
            this.rda_w = rda_w;
            this.rda_l = rda_l;

            if (endOP == null)
                throw new Exception("It is mandatory to specify a Take-off condition.");

            this.endOP = new Condition(id+"endOP");
            if (endOP == "EDWP-A")
            {
                this.destination = this.edwp_a;
                this.endOP.SetResult(2);
            } 
            else if (endOP == "EDWP-CL")
            {
                this.destination = this.edwp_cl;
                this.endOP.SetResult(0);
            } 
            else if (endOP == "EDWP-BL")
            {
                this.destination = this.edwp_bl;
                this.endOP.SetResult(1);
            } 
            else if (endOP == "EDWP-BR"){
                this.destination = this.edwp_br;
                this.endOP.SetResult(3);
            } 
            else if (endOP == "EDWP-CR")
            {
                this.destination = this.edwp_cr;
                this.endOP.SetResult(4);
            } 
            else
                throw new Exception("Incorrect end of departure code "+ endOP + ", must be EDWP-CL/BL/A/BR/CR");

            this.nextList = new LegsCollection();

            if ((nextRefs == null)||(nextRefs.Count != 5))
                    throw new Exception("Take-off next list should contain five end off departure next legs.");
            else 
            {
                // Creates an array with possible null values in case successor is not specified
                //int i = 0;
                this.nextRefs = new ArrayList();
                foreach (string s in nextRefs)
                       // this.nextRefs[i++] = s;
                    this.nextRefs.Add(s);
            }
        }

        #endregion
        #region - Public Functions -

        /// <summary>
        /// Returns deep copy of TO leg
        /// </summary>
        /// <param name="s"></param>
        /// <returns></returns>
        public override Leg DeepCopy(Stage s)
        {
            Leg l = base.DeepCopy(s);
            ArrayList nr = null;
            String endOP;

            if (this.nextRefs != null)
            {
                nr = new ArrayList();
                foreach (string r in this.nextRefs)
                    nr.Add(String.Copy(r));
            }

                 if (this.endOP.GetResult() == 0) endOP = "EDWP-CL";
            else if (this.endOP.GetResult() == 1) endOP = "EDWP-BL";
            else if (this.endOP.GetResult() == 2) endOP = "EDWP-A";
            else if (this.endOP.GetResult() == 3) endOP = "EDWP-BR";
            else if (this.endOP.GetResult() == 4) endOP = "EDWP-CR";
            else throw new Exception("Internal error on the endOP condition.");

            TOLeg tol = new TOLeg(l.stage, l.id, l.destination, l.nextid, l.previd,
                l.emergency, this.rwyhead, this.course, this.length, this.der, this.edwp_a, this.edwp_a_min,
                this.edwp_bl, this.edwp_bl_min, this.edwp_br, this.edwp_br_min, this.edwp_cl, this.edwp_cr, this.rda_w, this.rda_l, 
                nr, endOP);
            return tol;
        }

        //--------------------------------------------------------------------- 
        /// <summary>
        /// Sets field indicated by field to the given value. 
        /// </summary>
        //---------------------------------------------------------------------
        public override void SetField(string field, object value)
        {
            switch (field)
            {
                case "next":
                    string legid = value as string;
                    if (nextRefs.IndexOf(legid) >= 0)
                    {
                        log.InfoFormat("takeoff leg: next set to {0}", legid);
                        next = nextList[legid];
                    }
                    break;
                default:
                    base.SetField(field, value);
                    break;
            }
        }

        //---------------------------------------------------------------------
        /// <summary>
        /// Checks whether the leg depends on a given condition.
        /// </summary>
        /// <remarks>Overrides definition provided at base leg class.</remarks>
        /// <param name="id">A <see cref="System.String"/> with the condition identifier.</param>
        /// <returns>True if the leg depends on the condition, false otherwise.</returns>
        //---------------------------------------------------------------------
        public override bool DependsOnCondition(string id)
        {
            log.Debug(string.Format("Checking dependence of leg {0} on condition {1}", this.id, id));
            log.Debug(string.Format("nextCond field of leg {0} is {1} null", this.id, (endOP == null ? "" : "not")));

            bool depends = false;

            if (endOP != null)
                depends = (endOP.CondId == id);

            return depends;
        }

        #endregion
        #region - Xml -

        /// <summary>
        /// Gets the XsiType for this Leg.
        /// </summary>
        protected override String XsiType
        {
            get { return "fp:TOLeg"; }
        }

        /// <summary>
        /// Appends this Object to a xml LegsCollection representation.
        /// </summary>
        /// <param name="legsElement">The LegsCollection XmlElement representation where to append this Object.</param>
        public override XmlElement AppendXml(XmlElement legsElement)
        {
            XmlElement xmlLeg = base.AppendXml(legsElement);

            XmlHelper.AddElement(xmlLeg, "length", this.Length.ToString());
            XmlHelper.AddElement(xmlLeg, "course", this.Course.ToString());

            XmlElement aux;

            if (Rwyhead != null)
            {
                aux = XmlHelper.AddElement(xmlLeg, "rwyhead", null);
                Rwyhead.AppendXml(aux);
            }

            if (Der != null)
            {
                aux = XmlHelper.AddElement(xmlLeg, "der", null);
                Der.AppendXml(aux);
            }

            if (Edwp_a != null)
            {
                aux = XmlHelper.AddElement(xmlLeg, "edwp-a", null);
                Edwp_a.AppendXml(aux);
            }

            if (Edwp_a_min != null)
            {
                aux = XmlHelper.AddElement(xmlLeg, "edwp-a-min", null);
                Edwp_a_min.AppendXml(aux);
            }

            if (Edwp_bl != null)
            {
                aux = XmlHelper.AddElement(xmlLeg, "edwp-bl", null);
                Edwp_bl.AppendXml(aux);
            }

            if (Edwp_bl_min != null)
            {
                aux = XmlHelper.AddElement(xmlLeg, "edwp-bl-min", null);
                Edwp_bl_min.AppendXml(aux);
            }
            if (Edwp_br != null)
            {
                aux = XmlHelper.AddElement(xmlLeg, "edwp-br", null);
                Edwp_br.AppendXml(aux);
            }

            if (Edwp_br_min != null)
            {
                aux = XmlHelper.AddElement(xmlLeg, "edwp-br-min", null);
                Edwp_br_min.AppendXml(aux);
            }

            if (Edwp_cl != null)
            {
                aux = XmlHelper.AddElement(xmlLeg, "edwp-cl", null);
                Edwp_cl.AppendXml(aux);
            }

            if (Edwp_cr != null)
            {
                aux = XmlHelper.AddElement(xmlLeg, "edwp-cr", null);
                Edwp_cr.AppendXml(aux);
            }

            XmlHelper.AddElement(xmlLeg, "rda-w", this.Rda_w.ToString());
            XmlHelper.AddElement(xmlLeg, "rda-l", this.Rda_l.ToString());
            if (EndOP != null) XmlHelper.AddElement(xmlLeg, "end-op", this.EndOP.CondId);

            return xmlLeg;
        }

        #endregion
    }
}
