using System;
using System.IO;
using System.Collections;
using System.Collections.Generic;

using Utils;
using System.Xml;

//-----------------------------------------------------------------------------
// Implements specialized leg types derived from base class.
// Steps in order to implement a new leg type:
//   (1) Add leg type to XML schema definition.
//   (2) Define leg subclass.
//   (3) Update XMLFPRdr so that it is able to read the new leg type.
//   (4) Check if Stage.UpdateGraph needs to be modified. 
//   (5) Implement subclass methods.
//
// Implemented legs:
//   * IF: Initial Fix
//   * TF: Track to a Fix
//   * DF: Direct to a Fix
//   * RF: Radius to a Fix
//   * IL: Iterative leg (partially)
//   * Intersection Leg (partially)
//   * Holding to a Fix
//   * Parametric Leg
//       - BasicScanLeg
// Pending:
//   * Holding to an Altitude
//   * Holding to an Condition	
//-----------------------------------------------------------------------------

namespace FlightPlan
{

	//-------------------------------------------------------------------------
	/// <summary>
	/// Initial Fix (IF) leg class. 
	/// </summary>
	//-------------------------------------------------------------------------
	public class IFLeg: Leg
    {
        #region - Constructor -
        /// <summary>
		/// Creates a new instance of IFLeg. 
		/// </summary>		
		public IFLeg(Stage stg, string id, Waypoint destination, string nextid, string previd, ContingencyList emergency) 
			: base (stg, id, destination, nextid, previd, emergency) {}

        #endregion
        #region - Public -

        /// <summary>
        /// Returns deep copy of IF leg object.
        /// </summary>
        /// <param name="s"></param>
        /// <returns></returns>
        public override Leg DeepCopy(Stage s)
        {
            Leg l = base.DeepCopy(s);
            return new IFLeg(l.stage, l.id, l.destination, l.nextid, l.previd, l.emergency);
        }

        #endregion
        #region - Xml -

        /// <summary>
        /// Gets the XsiType for this Leg.
        /// </summary>
        protected override String XsiType
        {
            get { return "fp:IFLeg"; }
        }

        #endregion
	}

	//-------------------------------------------------------------------------
	/// <summary>
	/// Track to a Fix (TF) leg class. 
	/// </summary>
	//-------------------------------------------------------------------------
	public class TFLeg: Leg
    {
        #region - Constructor -

        /// <summary>
		/// Creates a new instance of TFLeg.
		/// </summary>
		//---------------------------------------------------------------------				
		public TFLeg(Stage stg, string id, Waypoint destination, string nextid, string previd, ContingencyList emergency) 
			: base (stg, id, destination, nextid, previd, emergency) {}

        #endregion
        #region - Public -

        /// <summary>
        /// Return deep copy of TF leg.
        /// </summary>
        /// <param name="s"></param>
        /// <returns></returns>
        public override Leg DeepCopy(Stage s)
        {
            Leg l = base.DeepCopy(s);
            return new TFLeg(l.stage, l.id, l.destination, l.nextid, l.previd, l.emergency);
        }

        #endregion
        #region - Xml -

        /// <summary>
        /// Gets the XsiType for this Leg.
        /// </summary>
        protected override String XsiType
        {
            get { return "fp:TFLeg"; }
        }

        #endregion
	}

	//-------------------------------------------------------------------------
	/// <summary>
	/// Direct to a Fix (DF) leg class. 
	/// </summary>
	//-------------------------------------------------------------------------
	public class DFLeg: Leg
    {
        #region - Constructor -

        /// <summary>
		/// Creates a new instance of DFLeg. 
		/// </summary>		
		public DFLeg(Stage stg, string id, Waypoint destination, string nextid, string previd, ContingencyList emergency) 
			: base (stg, id, destination, nextid, previd, emergency) {}

        #endregion
        #region - Public -

        /// <summary>
        /// Returns deep copy of DF leg object.
        /// </summary>
        /// <param name="s"></param>
        /// <returns></returns>
        public override Leg DeepCopy(Stage s)
        {
            Leg l = base.DeepCopy(s);
            return new DFLeg(l.stage, l.id, l.destination, l.nextid, l.previd, l.emergency);
        }

        #endregion
        #region - Xml -

        /// <summary>
        /// Gets the XsiType for this Leg.
        /// </summary>
        protected override String XsiType
        {
            get { return "fp:DFLeg"; }
        }

        #endregion
	}

	//-------------------------------------------------------------------------
	/// <summary>
	/// Radius to Fix (RF) leg class. 
	/// </summary>
	//-------------------------------------------------------------------------
	public class RFLeg: Leg
    {
        #region - Attributes & Properties -

        // TODO: Generated waypoints should be automatically generated having
		//   into account aircraft performance
		
		// Generate one wp every S m of arc length
		// Rascal 110 R/C
		const double S = 100;
		// Cessna 172 R
		// const double S = 500;

		/// <summary>
		/// Latitude of arc center. 
		/// </summary>
		public GPSCoords clatitude;
        public GPSCoords Clatitude
        {
            get { return clatitude; }
            set { clatitude = value; }
        }

		/// <summary>
		/// Longitude of arc center.
		/// </summary>
		public GPSCoords clongitude;
        public GPSCoords Clongitude
        {
            get { return clongitude; }
            set { clongitude = value; }
        }

		/// <summary>
		/// Turn direction.
		/// </summary>
		public Leg.TurnDir direction;
        public Leg.TurnDir Direction
        {
            get { return direction; }
            set { direction = value; }
        }
		
		private double _radius;
		/// <summary>
		/// Turn radius.
		/// </summary>
		public double radius 
        {
			get { return _radius; }
            set { _radius = value; }
		}

        #endregion
        #region - Constructor -

        //---------------------------------------------------------------------
		/// <summary>
		/// Creates an RFLeg object.
		/// </summary>
		/// <param name="stg"><see cref="Stage"/> the leg belongs to.</param>
		/// <param name="id">A <see cref="System.String"/> with the leg id.</param>
		/// <param name="destination">Destination <see cref="Waypoint"/>.</param>
		/// <param name="nextid">A <see cref="System.String"/> with the id of the next leg.</param>
		/// <param name="previd">A <see cref="System.String"/> with the id of the previous leg.</param>
		/// <param name="emergency">A <see cref="FlightPlan"/> to be executed in case of emergency.</param>
		/// <param name="clat">A <see cref="GPSCoords"/> with the latitude of the turn center.</param>
		/// <param name="clon">A <see cref="GPSCoords"/> with the longitude of the turn center.</param>
		/// <param name="direction">A <see cref="Leg.TurnDir"/> with the turn direction.</param>
		//---------------------------------------------------------------------				
		public RFLeg(Stage stg, string id, Waypoint destination, string nextid, 
			string previd, ContingencyList emergency, GPSCoords clat, GPSCoords clon,
			Leg.TurnDir direction)
			: base (stg, id, destination, nextid, previd, emergency)
		{
			this.clatitude = clat;
			this.clongitude = clon;
			this.direction = direction;
			
			// Compute center to destination distance and set _radius
			double az1, az2;
			Geodesy.geo_inverse_wgs_84(0, clatitude.AsDouble(), clongitude.AsDouble(),
			                           destination.latitude.AsDouble(),
			                           destination.longitude.AsDouble(),
			                           out az1, out az2, out _radius);			
		}

        #endregion
        #region - Public Functions -

        /// <summary>
        /// Returns a deep copy of a RF leg.
        /// </summary>
        /// <returns></returns>
        public override Leg DeepCopy(Stage s)
        {
            Leg l = base.DeepCopy(s);

            GPSCoords clat = this.clatitude.DeepCopy();
            GPSCoords clon = this.clongitude.DeepCopy();

            RFLeg nl = new RFLeg(l.stage, l.id, l.destination, l.nextid, l.previd,
                l.emergency, clat, clon, this.direction);
            
            nl._radius = this._radius;

            return nl;
        }

        #endregion
        #region - Xml -

        /// <summary>
        /// Gets the XsiType for this Leg.
        /// </summary>
        protected override String XsiType
        {
            get { return "fp:RFLeg"; }
        }

        /// <summary>
        /// Creates a String representation for the Arc Center coordinates (latitude longitude). 
        /// </summary>
        /// <returns>A String representation for this Arc Center coordinates (latitude longitude).</returns>
        private String CenterCoordinatesToStr()
        {
            if (clatitude == null) throw new Exception("The latitude can't be null");
            if (clongitude == null) throw new Exception("The longitude can't be null");

            return String.Format("{0} {1}", clatitude.AsDouble(), clongitude.AsDouble());
        }

        /// <summary>
        /// Appends this Object to a xml LegsCollection representation.
        /// </summary>
        /// <param name="legsElement">The LegsCollection XmlElement representation where to append this Object.</param>
        public override XmlElement AppendXml(XmlElement legsElement)
        {
            XmlElement xmlLeg = base.AppendXml(legsElement);

            XmlHelper.AddElement(xmlLeg, "direction", this.Direction.ToString());
            XmlHelper.AddElement(xmlLeg, "center",    CenterCoordinatesToStr());

            return xmlLeg;
        }

        #endregion
	}
	
	//-------------------------------------------------------------------------
	/// <summary>
	/// Holding to a Fix leg class. 
	/// </summary>
	//-------------------------------------------------------------------------
	public class HFLeg: Leg
    {
        #region - Attributes & Properties -

        /// <summary>
		/// Azimuth angle which determines the orientation of the holding pattern. 
		/// </summary>
		public double course;
        public double Course
        {
            get { return course; }
            set { course = value; }
        }

		/// <summary>
		/// Indicates whether to turn left or right when executing the pattern.
		/// </summary>
		public TurnDir direction;
        public TurnDir Direction
        {
            get { return direction; }
            set { direction = value; }
        }

		/// <summary>
		/// Distance between turn centers of the holding pattern.
		/// </summary>
		public double d1;
        public double D1
        {
            get { return d1; }
            set { d1 = value; }
        }

		/// <summary>
		/// Diameter of the turn segments.
		/// </summary>
		public double d2;
        public double D2
        {
            get { return d2; }
            set { d2 = value; }
        }

		/// <summary>
		/// Number of allowed iterations.
		/// </summary>
		public int upperBound;
        public int UpperBound
        {
            get { return upperBound; }
            set { upperBound = value; }
        }

		/// <summary>
		/// Condition to be applied in order to decide if another iteration
		/// should be executed.
		/// </summary>
		public Condition cond;
        public Condition Cond
        {
            get { return cond; }
            set { cond = value; }
        }
		
		/// <summary>
		/// Returns d1+d2;
		/// </summary>
		public double length 
        {
			get { return d1 + d2; }
		}

		/// <summary>
		/// Returns d2/2
		/// </summary>
		public double radius 
        {
			get { return d2/2.0; }
		}

        #endregion
        #region - Constructors -

        /// <summary>
		/// Creates a HFLeg object.
		/// </summary>
		/// <param name="stg"><see cref="Stage"/> the leg belongs to.</param>
		/// <param name="id">A <see cref="System.String"/> with the leg id.</param>
		/// <param name="destination">A <see cref="Waypoint"/> with the initial fix of the holding pattern.</param>
		/// <param name="nextid">A <see cref="System.String"/> with the id of the next leg.</param>
		/// <param name="previd">A <see cref="System.String"/> with the id of the previous leg.</param>
		/// <param name="emergency">A <see cref="FlightPlan"/> to be executed in case of emergency.</param>
		/// <param name="cour">A <see cref="System.Double"/> with the azimuth angle indicating orientation of the holding pattern.</param>
		/// <param name="dir">Turn direction.</param>
		/// <param name="d1">A <see cref="System.Double"/> with the distance between turn centers.</param>
		/// <param name="d2">A <see cref="System.Double"/> with the turn diameter.</param>
        /// <param name="upperBound"></param>
        /// <param name="condid"></param>
        public HFLeg(Stage stg, string id, Waypoint destination, string nextid,		                       
		             string previd, ContingencyList emergency, double cour, TurnDir dir, 
		             double d1, double d2, int upperBound, String condid)
			: base (stg, id, destination, nextid, previd, emergency)
		{
			course = cour;
			direction = dir;
			this.d1 = d1;
			this.d2 = d2;
			this.upperBound = upperBound;
			if (condid != null) 
				this.cond = new Condition(condid);			
		}

        #endregion
        #region - Public Functions -

        /// <summary>
        /// Returns deep copy of HF leg
        /// </summary>
        /// <param name="s"></param>
        /// <returns></returns>
        public override Leg DeepCopy(Stage s)
        {
            Leg l = base.DeepCopy(s);
            HFLeg hfl = new HFLeg(l.stage, l.id, l.destination, l.nextid, l.previd,
                l.emergency, this.course, this.direction, this.d1, this.d2, this.upperBound,
                (cond != null ? this.cond.CondId : null));
            return hfl;
        }

		//---------------------------------------------------------------------
		/// <summary>
		/// Checks whether the leg depends on a given condition.
		/// </summary>
		/// <remarks>
		/// Overrides definition provided at base leg class. 
		/// </remarks>
		/// <param name="id">A <see cref="System.String"/> with the condition identifier.</param>
		/// <returns>True if the leg depends on the condition, false otherwise.</returns>
		//---------------------------------------------------------------------
        public override bool DependsOnCondition(string id)
        {
            log.Debug(string.Format("Checking dependence of leg {0} on condition {1}", this.id, id));
            log.Debug(string.Format("cond field of leg {0} is {1} null", this.id, (cond == null ? "" : "not")));

            bool depends = false;

            if (cond != null)
                depends = (cond.CondId == id);

            return depends;
        }

        #endregion
        #region - Xml -

        /// <summary>
        /// Gets the XsiType for this Leg.
        /// </summary>
        protected override String XsiType
        {
            get { return "fp:HFLeg"; }
        }

        /// <summary>
        /// Appends this Object to a xml LegsCollection representation.
        /// </summary>
        /// <param name="legsElement">The LegsCollection XmlElement representation where to append this Object.</param>
        public override XmlElement AppendXml(XmlElement legsElement)
        {
            XmlElement xmlLeg = base.AppendXml(legsElement);

            XmlHelper.AddElement(xmlLeg, "direction", this.Direction.ToString());
            XmlHelper.AddElement(xmlLeg, "course",    course.ToString());
            XmlHelper.AddElement(xmlLeg, "d1",        d1.ToString());
            XmlHelper.AddElement(xmlLeg, "d2",        d2.ToString());
            if (Cond != null)
                XmlHelper.AddElement(xmlLeg, "cond",   Cond.CondId);
            XmlHelper.AddElement(xmlLeg, "upperBound", UpperBound.ToString());

            return xmlLeg;
        }

        #endregion		
	}

}
