// /home/esantama/DevProjs/mono/FPManager/ParamLegs.cs created with MonoDevelop
// User: esantama at 5:37 PM 12/13/2007

using System;
using System.IO;
using System.Collections;
using System.Collections.Generic;
using System.Reflection;
using System.Xml;

namespace FlightPlan
{
	/// <summary>
	/// Implements a basic scan leg that covers a rectangular area. 
	/// </summary>
	public class BasicScan: Leg
    {
        #region - Attributes & Properties -

        /// <summary>
		/// A <see cref="System.Double"/> expressing the length, in meters, of
		/// the sides parallel to the line defined by destination and angle. 
		/// </summary>
		public double dim1;
        public double Dim1
        {
            get { return dim1; }
            set { dim1 = value; }
        }

		/// <summary>
		/// A <see cref="System.Double"/> expressing the length, in meters, of
		/// the sides orthogonal to the line defined by destination and angle.
		/// </summary>
		public double dim2;
        public double Dim2
        {
            get { return dim2; }
            set { dim2 = value; }
        }

		/// <summary>
		/// A <see cref="System.Double"/> expressing an azimuth angle in degrees.
		/// Together with the initial waypoint determines one of the area of interest
		/// sides.
		/// </summary>
		public double angle;
        public double Angle
        {
            get { return angle; }
            set { angle = value; }
        }

		/// <summary>
		/// Position, belonging to the generated pattern, where the scan should start. 
		/// </summary>
		public GPSCoords[] startAt;
        public GPSCoords[] StartAt
        {
            get { return startAt; }
            set { startAt = value; }
        }

		/// <summary>
		/// Distance, in meters, between each pass of the scan pattern.
		/// </summary>
		public double separation;
        public double Separation
        {
            get { return separation; }
            set { separation = value; }
        }

		/// <summary>
		/// A <see cref="Leg.TurnDir"/> indicating a turn direction value in case a
		/// holding pattern is used for turning when reaching the end of a pass. 
		/// </summary>
		public Leg.TurnDir turndirection;
        public Leg.TurnDir Turndirection
        {
            get { return turndirection; }
            set { turndirection = value; }
        }

		/// <summary>
		/// Distance between turn centers of the holding pattern.
		/// </summary>
		public double d1;
        public double D1
        {
            get { return d1; }
            set { d1 = value; }
        }

		/// <summary>
		/// Diameter of the turn segments.
		/// </summary>
		public double d2;
        public double D2
        {
            get { return d2; }
            set { d2 = value; }
        }

		private const double ANGLE_EPSILON = 0.005;

        #endregion
        #region - Constructors -

        /// <summary>
		/// Creates a BasicScan leg object.
		/// </summary>
		/// <param name="stg"><see cref="Stage"/> the leg belongs to.</param>
		/// <param name="id"><see cref="System.String"/> containing the id of the leg.</param>
        /// <param name="destination">TBD</param>
		/// <param name="nextid"><see cref="System.String"/> with the id of the next leg or null if this is the last leg of the stage.</param>
		/// <param name="previd">A <see cref="System.String"/> with the id of the previous leg or null if this is the first leg of the stage.</param>
		/// <param name="emergency">A <see cref="FlightPlan"/> that should be followed if an emergency occurs during execution of this leg.</param>
		/// <param name="dim1">A <see cref="System.Double"/> expressing the length, in meters, of the sides parallel to the line defined by orig and angle.</param>
		/// <param name="dim2">A <see cref="System.Double"/> expressing the length, in meters, of the sides orthogonal to the line defined by orig and angle.</param>
		/// <param name="angle">A <see cref="System.Double"/> expressing an azimuth angle in degrees. Together with the initial waypoint determines one of the area of interest sides.</param>
		/// <param name="sep">Distance, in meters, between each pass of the scan pattern.</param>
		/// <param name="td">A <see cref="Leg.TurnDir"/> indicating a turn direction value in case a holding pattern is used for turning when reaching the end of a pass.</param>
		/// <param name="d1">A <see cref="System.Double"/> with the distance between turn centers of the holding pattern.</param>
		/// <param name="d2">A <see cref="System.Double"/> with the diameter of the turn segments.</param>
        /// <param name="startAt">A <see cref="GPSCoords"/> with latitude and longitude of the position, belonging to the generated pattern, where the scan should start.</param>
        public BasicScan
			(Stage stg, string id, Waypoint destination, string nextid, 
			 string previd, ContingencyList emergency,
			 double dim1, double dim2, double angle, double sep,
			 Leg.TurnDir? td, double? d1, double? d2, GPSCoords[] startAt)
				: base (stg, id, destination, nextid, previd, emergency)
		{
			this.dim1 = dim1;
			this.dim2 = dim2;
			this.angle = angle;
			this.startAt = startAt;
			this.separation = sep;
            // td, d1 and d2 may not have a value. In this case compute some
            // default params
            if (td.HasValue) this.turndirection = td.Value;
            else             this.turndirection = TurnDir.Right;

            if (d1.HasValue) this.d1 = d1.Value;
            else             this.d1 = 1.5 * separation;

            if (d2.HasValue) this.d2 = d2.Value;
            else             this.d2 = separation;
		}

        #endregion
        #region - Public Functions -

        /// <summary>
        /// Returs deep copy of basic scan leg.
        /// </summary>
        /// <param name="s"></param>
        /// <returns></returns>
        public override Leg  DeepCopy(Stage s)
        {
            Leg l = base.DeepCopy(s);
            GPSCoords[] start = null;
            if (this.startAt != null)
            {
                start = new GPSCoords[2];
                start[0] = this.startAt[0].DeepCopy();
                start[1] = this.startAt[1].DeepCopy();
            }
            BasicScan bs = new BasicScan(l.stage, l.id, l.destination, l.nextid,
                l.previd, l.emergency, this.dim1, this.dim2, this.angle,
                this.separation, this.turndirection, this.d1, this.d2,
                start);
            return bs;
        }

        #endregion
        #region - Xml -

        /// <summary>
        /// Gets the XsiType for this Leg.
        /// </summary>
        protected override String XsiType
        {
            get { return "fp:BasicScanLeg"; }
        }

        /// <summary>
        /// Appends this Object to a xml LegsCollection representation.
        /// </summary>
        /// <param name="legsElement">The LegsCollection XmlElement representation where to append this Object.</param>
        public override XmlElement AppendXml(XmlElement legsElement)
        {
            XmlElement xmlLeg = base.AppendXml(legsElement);
            
            XmlHelper.AddElement(xmlLeg, "dim1",          dim1.ToString());
            XmlHelper.AddElement(xmlLeg, "dim2",          dim2.ToString());
            XmlHelper.AddElement(xmlLeg, "angle",         angle.ToString());
            XmlHelper.AddElement(xmlLeg, "separation",    separation.ToString());
            //XmlHelper.AddElement(xmlLeg, "turndirection", this.turndirection.ToString());
            //XmlHelper.AddElement(xmlLeg, "d1",            d1.ToString());
            //XmlHelper.AddElement(xmlLeg, "d2",            d2.ToString());

            if(StartAt != null && StartAt[0] != null && StartAt[1] != null)
            {
                String s = String.Format("{0} {1}", StartAt[0].AsDouble(), StartAt[1].AsDouble());
                XmlHelper.AddElement(xmlLeg, "startAt", s);
            }

            return xmlLeg;
        }

        #endregion    
	}
}
