// /home/esantama/DevProjs/mono/FPManager/ControlLegs.cs created with MonoDevelop
// User: esantama at 3:03 PM 11/29/2007

using System;
using System.IO;
using System.Collections;
using System.Collections.Generic;
using System.Xml;

namespace FlightPlan
{
	
	// TODO: Implement reverse direction behavior.

	//-------------------------------------------------------------------------
	/// <summary>
	/// Intersection leg class. 
	/// </summary>
	//-------------------------------------------------------------------------
	public class IntersectionLeg: Leg
    {
        #region - Attributes -

        /// <summary>
		/// Condition to apply to decide next leg when going forward. 
		/// </summary>
		public Condition nextCond;
        public Condition NextCond
        {
            get { return nextCond; }
            set { nextCond = value; }
        }

		/// <summary>
		/// Collection of possible next legs when going forward.
		/// </summary>
		public LegsCollection nextList;
        public LegsCollection NextList
        {
            get { return nextList; }
            set { nextList = value; }
        }

		/// <summary>
		/// Condition to apply to decide next leg when going backwards.
		/// </summary>
		public Condition prevCond;
        public Condition PrevCond
        {
            get { return prevCond; }
            set { prevCond = value; }
        }

		/// <summary>
		/// Collection of possible next legs when going backwards.
		/// </summary>
		public LegsCollection prevList;
        public LegsCollection PrevList
        {
            get { return prevList; }
            set { prevList = value; }
        }

		/// <summary>
		/// Next legs as a list of references.
		/// </summary>
		public ArrayList nextRefs = null;
        public ArrayList NextRefs
        {
            get { return nextRefs; }
            set { nextRefs = value; }
        }

		/// <summary>
		/// Previous legs as a list of references.
		/// </summary>
		public ArrayList prevRefs = null;
        public ArrayList PrevRefs
        {
            get { return prevRefs; }
            set { prevRefs = value; }
        }

        #endregion
        #region - Constructors -

        //---------------------------------------------------------------------
		/// <summary>
		/// Constructor.
		/// </summary>
		//---------------------------------------------------------------------
		public IntersectionLeg(Stage stg, string id, Waypoint destination,
		                       string nextid, string previd, ContingencyList emergency,
		                       string nextCondId, ArrayList nextRefs, 
		                       string prevCondId, ArrayList prevRefs)
			: base (stg, id, destination, nextid, previd, emergency)
		{
			this.nextList = new LegsCollection();
			this.prevList = new LegsCollection();
			
			if (nextRefs != null)
            {
				this.nextRefs = new ArrayList();
				foreach (string s in nextRefs)
					this.nextRefs.Add(s);

				// if more than one possible path, check there is a condition
				// for deciding.
				if (nextRefs.Count > 1) 
                {
					if (nextCondId == null)
						throw new Exception("It is mandatory to specify a condition.");
					else
						nextCond = new Condition(nextCondId);
				}			
			}
			
			
			if (prevRefs != null)
            {
				this.prevRefs = new ArrayList();
				foreach (string s in prevRefs)
					this.prevRefs.Add(s);
				
				if (prevRefs.Count > 1) 
                {
					if (prevCondId == null)
						throw new Exception("It is mandatory to specify a condition.");
					else
						prevCond = new Condition(prevCondId);
				}
			}
		}

        #endregion
        #region - Public Functions -

        /// <summary>
        /// Returns a deep copy of an intersection leg object.
        /// </summary>
        /// <param name="s"></param>
        /// <returns></returns>
        public override Leg DeepCopy(Stage s)
        {
            Leg l = base.DeepCopy(s);
            ArrayList nr = null;
            if (this.nextRefs != null)
            {
                nr = new ArrayList();
                foreach (string r in this.nextRefs)
                    nr.Add(String.Copy(r));
            }
            ArrayList pr = null;
            if (this.prevRefs != null)
            {
                pr = new ArrayList();
                foreach (string r in this.prevRefs)
                    pr.Add(String.Copy(r));
            }
            IntersectionLeg il = new IntersectionLeg(l.stage, l.id, l.destination,
                l.nextid, l.previd, l.emergency, 
                (this.nextCond != null? String.Copy(this.nextCond.CondId) : null), nr, 
                (this.prevCond != null? String.Copy(this.prevCond.CondId) : null), pr);
            return il;
        }
		
		//--------------------------------------------------------------------- 
		/// <summary>
		/// Sets field indicated by field to the given value. 
		/// </summary>
		//---------------------------------------------------------------------
		public override void SetField(string field, object value)
		{
			switch (field) 
            {
			    case "next":
				    string legid = value as string; 
				    if (nextRefs.IndexOf(legid) >= 0) 
                    {					
					    log.InfoFormat("intersectionleg: next set to {0}", legid);
					    next = nextList[legid];
				    }
				break;
			    default:
				    base.SetField(field, value);
				break;
			}
		}

		//---------------------------------------------------------------------
		/// <summary>
		/// Checks whether the leg depends on a given condition.
		/// </summary>
		/// <remarks>Overrides definition provided at base leg class.</remarks>
		/// <param name="id">A <see cref="System.String"/> with the condition identifier.</param>
		/// <returns>True if the leg depends on the condition, false otherwise.</returns>
		//---------------------------------------------------------------------
		public override bool DependsOnCondition(string id)
		{
			log.Debug(string.Format("Checking dependence of leg {0} on condition {1}", this.id, id));
			log.Debug(string.Format("nextCond field of leg {0} is {1} null", this.id, (nextCond == null? "" : "not")));

			bool depends = false;
			
			if (nextCond != null)
				depends = (nextCond.CondId == id);
					
			return depends; 
		}

		//---------------------------------------------------------------------
		/// <summary>
		/// Returns true if the leg depends on a condition that has not been
		/// initialized and no default leg has been specified. If there is a
		/// default leg it returns false.
		/// </summary>
		//---------------------------------------------------------------------
		public override bool AwaitsConditionResult()
		{
			// If there is a default leg we are sure we don't have to wait.
			if (next != null)
				return false;
			
			// If there is no condition, we probably are at a join.
			if (nextCond == null)
			{
				if (nextList.Count == 1)
					// Assume default leg is the only one available
					next = nextList[0];
				else if (nextList.Count > 1)
				{
					// TODO: This should be validated before
					string msg =
						String.Format("More than one possible next with no condition at leg {0}", id);
					throw new Exception(msg);
				}
				
				return false;
			}
			
			// Otherwise check if a result has been provided
			return !nextCond.HasResult();
		}

        #endregion
        #region - Xml -

        /// <summary>
        /// Gets the XsiType for this Leg.
        /// </summary>
        protected override String XsiType
        {
            get { return "fp:IntersectionLeg"; }
        }

        /// <summary>
        /// Appends this Object to a xml LegsCollection representation.
        /// </summary>
        /// <param name="legsElement">The LegsCollection XmlElement representation where to append this Object.</param>
        public override XmlElement AppendXml(XmlElement legsElement)
        {
            XmlElement xmlLeg = base.AppendXml(legsElement);

            // Creates the strNextList
            if (this.nextRefs != null && this.nextRefs.Count>0)
            {
                String strNextList = "";
                for (int i = 0; i < this.nextRefs.Count; i++)
                {
                    strNextList += this.nextRefs[i].ToString();
                    if (i < this.nextRefs.Count - 1)
                        strNextList += " ";
                }

                XmlHelper.AddElement(xmlLeg, "nextList", strNextList);
                XmlHelper.AddElement(xmlLeg, "nextCond", this.NextCond.CondId);

            }

            // Creates the strPrevList
            if (this.prevRefs != null && this.prevRefs.Count > 0)
            {
                String strPrevList = "";
                for (int i = 0; i < this.prevRefs.Count; i++)
                {
                    strPrevList += this.prevRefs[i].ToString();
                    if (i < this.prevRefs.Count - 1)
                        strPrevList += " ";
                }
                
                XmlHelper.AddElement(xmlLeg, "prevList", strPrevList);
                XmlHelper.AddElement(xmlLeg, "prevCond", this.PrevCond.CondId);
            }

            return xmlLeg;
        }

        #endregion		
	}
		
	//-------------------------------------------------------------------------
	/// <summary>
	/// Iterative leg class. 
	/// </summary>
	//-------------------------------------------------------------------------
	public class IterativeLeg: Leg
    {
        #region - Attributes -

        /// <summary>
		/// First leg of the iterative leg body. 
		/// </summary>
		public Leg first;
		/// <summary>
		/// Last leg of the iterative leg body.
		/// </summary>
		public Leg last;
		/// <summary>
		/// Number of allowed iterations.
		/// </summary>
		public int upperBound;
		/// <summary>
		/// Legs that conform the body of the iterative leg.
		/// </summary>
		public LegsCollection body;
		/// <summary>
		/// Condition to be applied in order to decide if another iteration
		/// should be executed.
		/// </summary>
		public Condition cond;

		/// <summary>
		/// id of the first leg of the iterative leg body.
		/// </summary>
		public string firstId;
		/// <summary>
		/// id of the last leg of the iterative leg body.
		/// </summary>
		public string lastId;
		/// <summary>
		/// Legs of the iterative leg body as a list of references. 
		/// </summary>
		public ArrayList bodyRefs;

        #endregion
        #region - Constructors -

        /// <summary>
        /// Creates an iterative leg
		/// </summary>
		/// <param name="stg"><see cref="Stage"/> the leg belongs to.</param>
        /// <param name="id"><see cref="System.String"/> ID of this leg.</param>
		/// <param name="nextid"><see cref="System.String"/> ID of the next leg.</param>
		/// <param name="previd"><see cref="System.String"/> ID of the previous leg.</param>
		/// <param name="emergency">A <see cref="FlightPlan"/>. Emergency plan to be executed if a problem occurs when executing this leg.</param>
		/// <param name="body">A <see cref="ArrayList"/> with all legs that belong to the iterative leg body.</param>
		/// <param name="upperBound">A <see cref="System.Int32"/> with the number of allowed iterations.</param>
		/// <param name="condid"></param>
        /// <param name="first">A <see cref="System.String"/> with the id of the first leg.</param>
		/// <param name="last">A <see cref="System.String"/> with the id of the last leg.</param>
        public IterativeLeg(Stage stg, string id, string nextid, string previd, 
		                    ContingencyList emergency, ArrayList body, int upperBound,
		                    String condid, string first, string last)
			: base (stg, id, null, nextid, previd, emergency)
		{
			this.first = null;
			this.last = null;
			this.firstId = first;
			this.lastId = last;
			this.upperBound = upperBound;
			if (condid != null) 
				this.cond = new Condition(condid);
			this.body = new LegsCollection();
			this.bodyRefs = new ArrayList();
			foreach (string s in body)
				this.bodyRefs.Add(s);
		}

        #endregion
        #region - Public Functions -

        /// <summary>
        /// Returns a deep copy of an iterative leg object.
        /// </summary>
        /// <param name="s"></param>
        /// <returns></returns>
        public override Leg DeepCopy(Stage s)
        {
            Leg l = base.DeepCopy(s);
            ArrayList body_legs = new ArrayList();
            foreach (string lid in this.bodyRefs)
                body_legs.Add(String.Copy(lid));
            IterativeLeg il = new IterativeLeg(l.stage, l.id, l.nextid, l.previd,
                l.emergency, body_legs, this.upperBound, 
                (this.cond != null ? String.Copy(this.cond.CondId) : null),
                String.Copy(this.firstId), String.Copy(this.lastId));
            return il;
        }
		
		//---------------------------------------------------------------------
		/// <summary>
		/// Checks whether the leg depends on a given condition.
		/// </summary>
		/// <remarks>Overrides definition provided at base leg class.</remarks>
		/// <param name="id">A <see cref="System.String"/> with the condition identifier.</param>
		/// <returns>True if the leg depends on the condition, false otherwise. </returns>
		//---------------------------------------------------------------------
		public override bool DependsOnCondition(string id)
		{
			log.Debug(string.Format("Checking dependence of leg {0} on condition {1}", this.id, id));
			log.Debug(string.Format("cond field of leg {0} is {1} null", this.id, (cond == null? "" : "not")));
			
			bool depends = false;

			if (cond != null)
				depends = (cond.CondId == id);

			return depends;
		}

        #endregion
        #region - Xml -

        /// <summary>
        /// Gets the XsiType for this Leg.
        /// </summary>
        protected override String XsiType
        {
            get { return "fp:IterativeLeg"; }
        }

        /// <summary>
        /// Appends this Object to a xml LegsCollection representation.
        /// </summary>
        /// <param name="legsElement">The LegsCollection XmlElement representation where to append this Object.</param>
        public override XmlElement AppendXml(XmlElement legsElement)
        {
            XmlElement xmlLeg = base.AppendXml(legsElement);

            // Creates the strBody
            String strBody = "";
            for (int i = 0; i < this.bodyRefs.Count; i++)
            {
                strBody += this.bodyRefs[i].ToString();
                if (i < this.bodyRefs.Count - 1)
                    strBody += " ";
            }

            XmlHelper.AddElement(xmlLeg, "body",       strBody);
            XmlHelper.AddElement(xmlLeg, "upperBound", this.upperBound.ToString());
            XmlHelper.AddElement(xmlLeg, "first",      this.firstId);
            XmlHelper.AddElement(xmlLeg, "last",       this.lastId);            
            XmlHelper.AddElement(xmlLeg, "cond",       this.cond.CondId);

            return xmlLeg;
        }

        #endregion
	}
	
}
