
using System;
using System.IO;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.Xml;

namespace FlightPlan
{	
	/// <summary>
	/// A collection of fixes.
	/// </summary>
	/// <remarks>
	/// Implements a keyed collection of fixes.
	/// </remarks>	
	public class EmergencyCollection : KeyedCollection<string, Emergency>
    {
        #region - Constructor -

        /// <summary>
		/// Creates EmergencyCollection object.
		/// </summary>
		public EmergencyCollection() : base() {}

        #endregion
        #region - Public Functions -

        /// <summary>
		/// Implements GetKeyForItem method.
		/// </summary>
		/// <param name="item">A <see cref="Emeregency"/></param>
		/// <returns>A <see cref="System.String"/></returns>
	    protected override string GetKeyForItem(Emergency item)
	    {
	        return item.Id;
	    }

        /// <summary>
        /// Returns deep copy of a EmergencyCollection object.
        /// </summary>
        /// <returns></returns>
        public EmergencyCollection DeepCopy()
        {
            EmergencyCollection newec = new EmergencyCollection();
            foreach (Emergency e in this)
            {
                newec.Add(e.DeepCopy());
            }

            return newec;
        }

        #endregion
        #region - Xml -

        /// <summary>
        /// Appends this Object to a xml FlightPlan representation.
        /// </summary>
        /// <param name="fpElement">The FlighPlan XmlElement representation where to append this Object.</param>
        public void AppendXml(XmlElement fpElement)
        {
            // Test for errors
            if (fpElement == null) throw new ArgumentNullException("fpElement");
            
            // Create the elements
            XmlHelper.AddComment(fpElement, "List of emergency flight plans");
            XmlElement xmlEmergencies = XmlHelper.AddElement(fpElement, "EmergencyPlans", null);

            // Append all the fixes
            foreach (Emergency e in this)
            {
                // TODO: Implement next commented method
                //e.AppendXml(xmlEmergencies);
            }
        }

        #endregion
	}
		
	/// <summary>
	/// A contingency flight plan.
	/// </summary>	
	public class Emergency
    {
        #region - Attributes -

        protected string _id;
		protected string _name;
		protected string _desc;
        protected string _filename;
        protected FlightPlan _fp;

        #endregion
        #region - Properties -

        /// <summary>
        /// Emergency id.
        /// </summary>
        public string Id
        {
            get { return _id; }
        }

        /// <summary>
        /// Emergency name.
        /// </summary>
        public string Name
        {
            get { return _name; }
            set { _name = value; }
        }

        /// <summary>
        /// Emergency description.
        /// </summary>
        public string Description
        {
            get { return _desc; }
            set { _desc = value; }
        }

        /// <summary>
        /// Emergency filename.
        /// </summary>
        public string Filename
        {
            get { return _filename; }
            set { _filename = value; }
        }

        /// <summary>
        /// Emergency flight plan.
        /// </summary>
        public FlightPlan Fp
        {
            get { return _fp; }
            set { _fp = value; }
        }

        #endregion
        #region - Constructor -

        /// <summary>
		/// Creates an Emergency object.
		/// </summary>
		/// <param name="id">A <see cref="System.String"/>. Emergency identifier.</param>
		/// <param name="name">A <see cref="System.String"/>. Name of the Emergency.</param>
		/// <param name="desc">A <see cref="System.String"/>. Description of the Emergency.</param>
		/// <param name="filename">A <see cref="System.String"/>. Filename of the Emergency.</param>
        /// <param name="fp">A <see cref="FlightPlan"/>. Fligh plan of the Emergency.</param>        
        public Emergency(string id, string name, string desc, string filename, FlightPlan fp)
		{
			_id = id;
			_name = name;
			_desc = desc;
			_filename = filename;
            _fp = fp;
		}

        #endregion
        #region - Public Functions -

        /// <summary>
        /// Returns deep copy of a emergency.
        /// </summary>
        /// <returns></returns>
        public virtual Emergency DeepCopy()
        {
            Emergency e = this.MemberwiseClone() as Emergency;
            e._desc     = String.Copy(this._desc);
            e._id       = String.Copy(this._id);
            e._name     = String.Copy(this._name);
            e._filename = String.Copy(this._filename);
            e._fp       = this._fp;
            
            return e;
        }

        #endregion
    }


    /// <summary>
    /// A contingency action.
    /// </summary>    
    public class Contingency
    {
        #region - Attributes -

        protected string _fpid;
        protected string _legid;
        protected string _type;
        protected string _entryType;
        protected string _filename;
        protected FlightPlan _fp;

        #endregion
        #region - Properties -

        /// <summary>
        /// Flight plan id.
        /// </summary>
        public string fpId
        {
            get { return _fpid; }
            set { _fpid = value; }
        }

        /// <summary>
        /// Initial leg id (if any).
        /// </summary>
        public string legId
        {
            get { return _legid; }
            set { _legid = value; }
        }

        /// <summary>
        /// Type of contingency reaction.
        /// </summary>
        public string type
        {
            get { return _type; }
            set { _type = value; }
        }

        /// <summary>
        /// Desired Entry Type of contingency reaction.
        /// </summary>
        public string entryType
        {
            get { return _entryType; }
            set { _entryType = value; }
        }

        /// <summary>
        /// Emergency filename.
        /// </summary>
        public string Filename
        {
            get { return _filename; }
            set { _filename = value; }
        }

        /// <summary>
        /// Emergency flight plan.
        /// </summary>
        public FlightPlan Fp
        {
            get { return _fp; }
            set { _fp = value; }
        }

        #endregion
        #region - Constructors -
        
        /// <summary>
        /// Creates a Contingency object.
        /// </summary>
        /// <param name="fpId">A <see cref="System.String"/>. Emergency flight plan identifier.</param>
        /// <param name="legId">A <see cref="System.String"/>. Name of the preferred entry leg.</param>
        /// <param name="type">A <see cref="System.String"/>. Type of contingency reaction associated to this flight plan.</param>
        /// <param name="filename">A <see cref="System.String"/>. Filename of the emergency flight plan.</param>
        /// <param name="fp">A <see cref="FlightPlan"/>. Fligh plan of the Emergency.</param>
        public Contingency(string fpId, string legId, string type, string entryType )
        {
            if (fpId == null)
                throw new Exception("Contingency declaration does not contain fligh plan Id.");
            else
                _fpid = fpId;

            _legid = legId;
            if (type == null)
                throw new Exception("Contingency declaration does not contain its type definition.");
            else if ((type != "HOME") && (type != "A-RUNWAY") && (type != "C-RUNWAY") && (type != "C-RUNWAY"))
                throw new Exception("Contingency declaration does not contain its type definition.");
            else 
                _type = type;

            _entryType = entryType;
            _filename  = null;
            _fp        = null;
        }

        #endregion
        #region - Public Functions -

        /// <summary>
        /// Returns deep copy of a emergency.
        /// </summary>
        /// <returns></returns>
        public virtual Contingency DeepCopy()
        {
            Contingency c = this.MemberwiseClone() as Contingency;
            c._fpid = String.Copy(this._fpid);
            c._legid = null;
            if (this._legid != null)
                c._legid = String.Copy(this._legid);
            c._type = null;
            if (this._type != null)
                c._type = String.Copy(this._type);
            c._filename = null; 
            if (this._filename != null)
                c._filename = String.Copy(this._filename);
            c._entryType = null;
            if (this._entryType != null)
                c._type = String.Copy(this._entryType);
            c._fp = this._fp;

            return c;
        }

        #endregion
    }

    
    /// <summary>
    /// Randomly accessed list of contingencies. 
    /// </summary>    
    public class ContingencyList : List<Contingency>
    {
        #region - Constructor -

        /// <summary>
        /// Creates a ContingencyList object.
        /// </summary>
        public ContingencyList() : base() { }

        #endregion
        #region - Public Functions -

        /// <summary>
        /// Return a deep copy of the list of contingencies
        /// </summary>
        /// <returns></returns>
        public ContingencyList DeepCopy()
        {
            ContingencyList l = new ContingencyList();
            foreach (Contingency c in this)
                l.Add(c.DeepCopy());
            return l;
        }

        #endregion
    }

}
