using System;
using System.IO;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.Xml;

namespace FlightPlan
{
	//-------------------------------------------------------------------------
	/// <summary>
	/// A collection of fixes.
	/// </summary>
	/// <remarks>
	/// Implements a keyed collection of fixes.
	/// </remarks>
	//-------------------------------------------------------------------------
	public class FixesCollection : KeyedCollection<string, Fix>
    {
        #region - Constructor -

        /// <summary>
		/// Creates FixesCollection object.
		/// </summary>
		public FixesCollection() : base() {}

        #endregion
        #region - Public Functions -

        /// <summary>
		/// Implements GetKeyForItem method.
		/// </summary>
		/// <param name="item">A <see cref="Fix"/></param>
		/// <returns>A <see cref="System.String"/></returns>
	    protected override string GetKeyForItem(Fix item)
	    {
	        return item.Id;
	    }

        /// <summary>
        /// Returns deep copy of a FixesCollection object.
        /// </summary>
        /// <returns></returns>
        public FixesCollection DeepCopy()
        {
            FixesCollection newfc = new FixesCollection();
            foreach (Fix f in this)
            {
                newfc.Add(f.DeepCopy());
            }

            return newfc;
        }

        #endregion
        #region - Xml -

        /// <summary>
        /// Appends this Object to a xml FlightPlan representation.
        /// </summary>
        /// <param name="fpElement">The FlighPlan XmlElement representation where to append this Object.</param>
        public void AppendXml(XmlElement fpElement)
        {
            // Test for errors            
            if (fpElement == null) throw new ArgumentNullException("fpElement");

            // Create the elements
            XmlHelper.AddComment(fpElement, "List of fixes");
            XmlElement xmlFixes = XmlHelper.AddElement(fpElement, "Fixes", null);

            // Append all the fixes
            foreach (Fix f in this)
            {
                f.AppendXml(xmlFixes);
            }
        }

        #endregion
    }
	
	//-------------------------------------------------------------------------
	/// <summary>
	/// A flight plan Fix.
	/// </summary>
	//-------------------------------------------------------------------------
	public class Fix
    {
        #region - Attributes -

        protected string    _id;
		protected string    _name;
		protected string    _desc;
		protected GPSCoords _lat;
		protected GPSCoords _lon;

        #endregion
        #region - Properties -

        /// <summary>
        /// Fix id.
        /// </summary>
        public string Id
        {
            get { return _id; }
        }

        /// <summary>
        /// Fix name.
        /// </summary>
        public string Name
        {
            get { return _name; }
        }

        /// <summary>
        /// Fix description.
        /// </summary>
        public string Description
        {
            get { return _desc; }
        }

        /// <summary>
		/// Fix latitude.
		/// </summary>
		public GPSCoords Lat
		{
			get {return _lat;}
		}

		/// <summary>
		/// Fix longitude.
		/// </summary>
		public GPSCoords Lon
		{
			get {return _lon;}
		}

        #endregion
        #region - Constructor -
        
		/// <summary>
		/// Creates a Fix object.
		/// </summary>
		/// <param name="id">A <see cref="System.String"/>. Fix identifier.</param>
		/// <param name="name">A <see cref="System.String"/>. Name of the fix.</param>
		/// <param name="desc">A <see cref="System.String"/>. Description of the fix.</param>
		/// <param name="lat">A <see cref="GPSCoords"/>. Fix latitude.</param>
		/// <param name="lon">A <see cref="GPSCoords"/>. Fix longitude.</param>		
		public Fix(string id, string name, string desc, GPSCoords lat, GPSCoords lon)
		{
			_id   = id;
			_name = name;
			_desc = desc;
			_lat  = lat;
			_lon  = lon;
		}

        #endregion
        #region - Public Functions -

        /// <summary>
        /// Returns deep copy of a fix.
        /// </summary>
        /// <returns></returns>
        public virtual Fix DeepCopy()
        {
            Fix f   = this.MemberwiseClone() as Fix;
            f._desc = String.Copy(this._desc);
            f._id   = String.Copy(this._id);
            f._lat  = this._lat.DeepCopy();
            f._lon  = this._lon.DeepCopy();
            f._name = String.Copy(this._name);
            
            return f;
        }

        /// <summary>
        /// Creates a String representation for this fix coordinates (latitude longitude). 
        /// </summary>
        /// <returns>A String representation for this fix coordinates (latitude longitude).</returns>
        public String CoordinatesToStr()
        {
            if (_lat == null) throw new Exception("The latitude can't be null");
            if (_lon == null) throw new Exception("The longitude can't be null");

            return String.Format("{0} {1}", _lat.AsDouble(), _lon.AsDouble());
        }

        #endregion
        #region - Xml -

        /// <summary>
        /// Appends this Object to a xml FixesCollection representation.
        /// </summary>
        /// <param name="fixesElement">The FixesCollection XmlElement representation where to append this Object.</param>
        public virtual void AppendXml(XmlElement fixesElement)
        {
            // Test for errors
            if (fixesElement == null) throw new ArgumentNullException("fixesElement");

            // Creates the element            
            XmlElement xmlFix = fixesElement.OwnerDocument.CreateElement("Fix");

            // Set the attributes
            xmlFix.SetAttribute("id", this.Id);

            // Attributes
            XmlHelper.AddElement(xmlFix, "name",        this.Name);
            XmlHelper.AddElement(xmlFix, "description", this.Description);
            XmlHelper.AddElement(xmlFix, "coordinates", this.CoordinatesToStr());

            // Append the elements
            fixesElement.AppendChild(xmlFix);
        }

        #endregion
	}
	
	//-------------------------------------------------------------------------
	/// <summary>
	/// Class derived from Fix used to store runway information.
	/// </summary>
	/// <remarks>
	/// Specialized class for storing runway data.
	/// </remarks>
	//-------------------------------------------------------------------------
	public class RwyFix: Fix
    {
        #region - Attributes -

        private double _alt;
        private double _head;
        private double _len;

        #endregion
        #region - Properties -

        /// <summary>
		/// Altitude of the runway.
		/// </summary>
		public double Altitude 
        {
			get {return _alt;}
		}

        /// <summary>
        /// Heading of the runway.
        /// </summary>
        public double Heading
        {
            get { return _head; }
        }

        /// <summary>
        /// Length of the runway.
        /// </summary>
        public double Length
        {
            get { return _len; }
        }

        #endregion        
        #region - Constructor -

        /// <summary>
		/// Creates a RwyFix object.
		/// </summary>
		/// <param name="id">A <see cref="System.String"/>. Fix identifier.</param>
		/// <param name="name">A <see cref="System.String"/>. Name of the fix.</param>
		/// <param name="desc">A <see cref="System.String"/>. Description of the fix.</param>
		/// <param name="lat">A <see cref="GPSCoords"/>. Fix latitude.</param>
		/// <param name="lon">A <see cref="GPSCoords"/>. Fix longitude.</param>
		/// <param name="alt">A <see cref="System.Double"/> with the runway altitude.</param>
		/// <param name="head">A <see cref="System.Double"/> with the runway heading in degrees.</param>
		/// <param name="len">A <see cref="System.Double"/> with the length of the runway.</param>		
		public RwyFix(string id, string name, string desc, GPSCoords lat, GPSCoords lon, double alt, double head, double len)
			: base (id, name, desc, lat, lon)
		{
			_alt = alt;
			_head = head;
			_len = len;
		}

        #endregion
        #region - Public Functions -

        /// <summary>
        /// Returns deep copy of a runway fix.
        /// </summary>
        /// <returns></returns>
        public override Fix DeepCopy()
        {
            RwyFix f = this.MemberwiseClone() as RwyFix;
            f._desc = String.Copy(this._desc);
            f._id = String.Copy(this._id);
            f._lat = this._lat.DeepCopy();
            f._lon = this._lon.DeepCopy();
            f._name = String.Copy(this._name);
            return f;
        }

        #endregion
        #region - Xml -

        /// <summary>
        /// Appends this Object to a xml FixesCollection representation.
        /// </summary>
        /// <param name="fixesElement">The FixesCollection XmlElement representation where to append this Object.</param>
        public override void AppendXml(XmlElement fixesElement)
        {
            // Test for errors
            if (fixesElement == null) throw new ArgumentNullException("fixesElement");

            // Creates the element            
            XmlElement xmlFix = fixesElement.OwnerDocument.CreateElement("Fix");

            // Set the attributes
            xmlFix.SetAttribute("id",   this.Id);
            xmlFix.SetAttribute("type", XMLFPParser.xs_instance_ns, "fp:RwyFixType");

            // Attributes
            XmlHelper.AddElement(xmlFix, "name",        this.Name);
            XmlHelper.AddElement(xmlFix, "description", this.Description);
            XmlHelper.AddElement(xmlFix, "coordinates", this.CoordinatesToStr());
            XmlHelper.AddElement(xmlFix, "altitude",    this.Altitude.ToString());
            XmlHelper.AddElement(xmlFix, "heading",     this.Heading.ToString());
            XmlHelper.AddElement(xmlFix, "length",      this.Length.ToString());

            // Append the elements
            fixesElement.AppendChild(xmlFix);
        }

        #endregion
    }
	
}
