using System;
using System.IO;
using System.Globalization;
using System.Collections;
using System.Collections.Generic;
using System.Threading;
using System.Xml;

using log4net;
using Utils;

// The FlightPlan namespace contains the classes that conform the flight plan
// object model. These classes provide a representation of the flight plan as
// well as a parser to load flight plans. The functionality required for the 
// flight plan execution is also implemented within the object model. 

namespace FlightPlan
{
	
	//-------------------------------------------------------------------------
	/// <summary>
	/// Locale information used to make approppriate conversions when reading
	/// the XML flight plan.
	/// </summary>
	/// <remarks>
	/// Internally the Flight Plan manager uses the following units of measure:
	/// - speed: meters/second (ms)
	/// - distance: meters (m)
	/// - altitude: meters (m)
	/// Data read from the XML file will be converted to the approppriate units.
	/// Formatting information is required in order to interpret values in string
	/// format. This formatting info is read from the XML file and store in
	/// fprovider.
	///
	/// This information is stored in a Locale class so that it can be reused 
	/// when processing flight plan updates, also in XML format.
	/// </remarks>
	//-------------------------------------------------------------------------
	public class Locale
    {
        #region - Attributes -

        /// <summary>
		/// Speed units used in flight plan document.
		/// </summary>
		public Units.Speed speedUnits;

		/// <summary>
		/// Altitude units in flight plan document. 
		/// </summary>
		public Units.Length altitudeUnits;

		/// <summary>
		/// Distance units used in flight plan document.
		/// </summary>
		public Units.Length distanceUnits;

		/// <summary>
		/// Formatting information (decimal and group separators) 
		/// </summary>
		public NumberFormatInfo fprovider;

        #endregion
        #region - Constructor -

        /// <summary>
        /// Constructor. Inits the attributes to some default values.
        /// </summary>
        public Locale()
        {
            this.speedUnits    = Units.Speed.kt;
            this.altitudeUnits = Units.Length.m;
            this.distanceUnits = Units.Length.m;
            this.fprovider     = new NumberFormatInfo();
            this.fprovider.NumberDecimalSeparator = ".";
            this.fprovider.NumberGroupSeparator   = "";            
        }

        #endregion
        #region - Xml -

        /// <summary>
        /// Appends this Object to a xml FlightPlan representation.
        /// </summary>        
        /// <param name="fpElement">The FlighPlan XmlElement representation where to append this Object.</param>
        public void AppendXml(XmlElement fpElement)
        {
            // Test for errors
            if (fpElement == null) throw new ArgumentNullException("fpElement");

            // Creates the elements
            XmlElement xmlLocale = XmlHelper.AddElement(fpElement, "Locale", null);

            // Attributes
            XmlHelper.AddElement(xmlLocale, "speedUnits",       Units.SpeedToStr(this.speedUnits));
            XmlHelper.AddElement(xmlLocale, "altitudeUnits",    Units.LengthToStr(this.altitudeUnits));
            XmlHelper.AddElement(xmlLocale, "distanceUnits",    Units.LengthToStr(this.distanceUnits));
            XmlHelper.AddElement(xmlLocale, "decimalSeparator", this.fprovider.NumberDecimalSeparator);
            XmlHelper.AddElement(xmlLocale, "groupSeparator",   this.fprovider.NumberGroupSeparator);
        }

        #endregion
    }	

	//-------------------------------------------------------------------------
	/// <summary>
	/// Flight plan root element.
	/// Stores and manages all data related to the flight plan. 
	/// </summary>
	/// <remarks>Root of the flight plan object model.</remarks>
	//-------------------------------------------------------------------------	
	public class FlightPlan
    {
        #region - Attributes & Properties -

        /// <summary>
		/// Flight plan id.
		/// </summary>
		public string id;
		/// <summary>
		/// Name of the flight plan.
		/// </summary>
		public string name;
		/// <summary>
		/// Description of the flight plan.
		/// </summary>
		public string description;

		private Locale locale_info;
		/// <summary>
		/// Locale info (used to interpret values in the XML document) 
		/// </summary>
		public Locale LocaleInfo
		{
			get { return locale_info; }
		}

        // Type of flight plan
        public String type;
        /// <summary>
        /// Type of flight plan: STD, CONT, REACT
        /// </summary>
        public String Type
        {
            get { return type; }
            set { type = value; }
        }

		// Collection of fixes
		private FixesCollection fixes;
		/// <summary>
		/// Collection of flight plan fixes
		/// </summary>
		public FixesCollection Fixes
		{
			get { return fixes; }
		}

        // Collection of emergency
        private EmergencyCollection emergencies;
        /// <summary>
        /// Collection of emergency flight plans
        /// </summary>
        public EmergencyCollection Emergencies
        {
            get { return emergencies; }
        }

		/// <summary>
		/// List of flight plan stages.
		/// </summary>
        private StageList _stages;
        public StageList stages
        {
			get { return _stages; }
		}

        #endregion
        #region - Constructor -

        /// <summary>
		/// Flight Plan constructor.
		/// </summary>
		/// <remarks>
		/// Creates list of stages.
		/// </remarks>		
		public FlightPlan()
		{
			locale_info = new Locale();
			fixes = new FixesCollection();
            type = "STD";
				
			_stages = new StageList();
		}

        #endregion
        #region - Public Functions -

        /// <summary>
		/// Returns the stage with id stageId. null if not found
		/// </summary>
		/// <param name="stageId">A <see cref="System.String"/> with the id.</param>
		/// <returns>A <see cref="Stage"/> instance or null if not found.</returns>		
		public Stage FindStage(string stageId)
		{
			foreach (Stage s in stages)
			{
				if (s.id == stageId)
					return s;
			}
			return null;
		}		
		
		/// <summary>
		/// Returns the leg with id legId. null if not found. 
		/// </summary>
		/// <param name="legId">A <see cref="System.String"/> with the id.</param>
		/// <returns>A <see cref="Leg"/> instance or null if not found.</returns>		
		public Leg FindLeg(string legId)
		{
			foreach (Stage s in stages)
			{
				if (s.legsCol.Contains(legId))
					return s.legsCol[legId];
			}
			return null;
		}		
		
		/// <summary>
		/// Returns the list of legs in the flight plan which depend on the given condition. 
		/// </summary>
		/// <param name="condId">A <see cref="System.String"/> with the id of the condition.</param>
		/// <returns>A <see cref="ArrayList"/> with the found legs.</returns>		
		public ArrayList FindCondition(string condId)
		{
			ArrayList list = new ArrayList();
			
			foreach (Stage s in stages)
				list.AddRange(s.FindCondition(condId));
			
			return list;
		}
					
		// TODO: Emergency FP handling		
		
		/// <summary>
		/// Checks the flight plan. To be called once the whole flight plan has been read. 
		/// </summary>		
		public void Check()
		{
			foreach (Stage s in stages)				
				s.Check();
		}
		
		/// <summary>
		/// Returns the initial leg of the flight plan. 
		/// </summary>
		/// <param name="il">A <see cref="System.Int32"/> indicating the position of the leg within the initial stage.</param>
		/// <returns>The initial <see cref="Leg"/>.</returns>		
		public Leg GetInitialLeg(int il)
		{
			return stages[0].GetInitialLeg(il);
		}

        
		/// <summary>
		/// Returns the next leg of a final leg of a given stage. 
		/// </summary>
		/// <param name="s"> The stage l belongs to. </param>
		/// <param name="l"> The leg whose next we what to find out. </param>
		/// <returns> Next leg of l. </returns>		
		public Leg GetNextToFinal(Stage s, Leg l)
		{
			// Find stage s
			bool found = false;
			int i=0;
			while (i<stages.Count && !found)
			{
				found = (stages[i] == s);
				if (!found)
					i++;
			}
			
			if (!found)
				throw new Exception("Stage not found.");

            // scape the ignored stages ////////////////////////////////////
            if (i < stages.Count-1)
            {
                while (i < stages.Count-1 && stages[i+1].Ignore)
                    i++;
            }
            ////////////////////////////////////////////////////////////////
			
			if (i < stages.Count-1)
				return stages[i+1].GetInitialLeg(stages[i].GetIndexOfFinal(l));
			else
                return null; // last stage
		}

        /// <summary>
        /// Returns the previous leg of an initial leg of a given stage.
        /// </summary>
        /// <param name="s"> The stage l belongs to. </param>
        /// <param name="l"> The leg whose previous we what to find out. </param>
        /// <returns> Previoius leg of l. </returns>
        public Leg GetPrevToInitial(Stage s, Leg l)
        {
            // Find stage s
            bool found = false;
            int i = 0;
            while (i < stages.Count && !found)
            {
                found = (stages[i] == s);
                if (!found)
                    i++;
            }

            if (!found)
                throw new Exception("Stage not found.");

            // scape the ignored stages ////////////////////////////////////
            if(i > 0)
            {
                while (i > 0 && stages[i - 1].Ignore)
                    i--;
            }
            ////////////////////////////////////////////////////////////////


            if (i > 0)
            {
                return stages[i - 1].GetFinalLeg(stages[i].GetIndexOfInitial(l));
            }
            else
                // last stage
                return null;
        }
		
        /// <summary>
        /// A string that represents the current object.
        /// </summary>
        /// <returns></returns>
		public override string ToString()
		{
			string result = "";
			
			result = result + "fp id: " + id + "\n";
			result = result + "name: " + name + "\n";
			result = result + "description: " + description + "\n";

			int i=0;
			while (i<stages.Count-1)
			{
				result = result + stages[i].ToString() + "\n";
				i++;
			}
			if (stages.Count > 0)
				result = result + stages[i].ToString();
			
			return result;
		}

        /// <summary>
        /// Enables flight plan cloning
        /// </summary>
        /// <returns></returns>
        public FlightPlan DeepCopy()
        {
            FlightPlan deep_copy = this.MemberwiseClone() as FlightPlan;
            
            // Deep copy references
            deep_copy.description = String.Copy(this.description);
            deep_copy.fixes = this.fixes.DeepCopy();
            deep_copy.emergencies = null;
            deep_copy.id = String.Copy(this.id);
            deep_copy.name = String.Copy(this.name);
            deep_copy._stages = this._stages.DeepCopy(deep_copy);

            return deep_copy;
        }

        #endregion
        #region - XML -

        /// <summary>
        /// Gets the xml representation for this FlightPlan.
        /// </summary>
        /// <param name="xmlDoc">The xml document necessary to create the element.</param>
        /// <returns>The XmlElement representation for this FlightPlan.</returns>
        /// <exception cref="System.NullReferenceException">If xmlDoc is null.</exception>
        public XmlElement GetXmlElement(XmlDocument xmlDoc)
        {
            // Get the root Node <scheme>
            XmlElement fpElement = xmlDoc.CreateElement("FlightPlan");

            // Set some attributes
            fpElement.SetAttribute("xmlns:fp",  XMLFPParser.xs_flightplan_ns);
            fpElement.SetAttribute("xmlns:xsi", XMLFPParser.xs_instance_ns);
            fpElement.SetAttribute("xsi:schemaLocation", "http://icarus.upc.es/schema/FlightPlan/1.1 file:/home/esantama/Development/imp/src/fpspec/1.1/FlightPlan-1.1.xsd");

            // 1 - Append the locale Settings
            this.locale_info.AppendXml(fpElement);
            /////////////////////////////////////////////////////////////////////////////

            // 2 - Append the fixes list
            if (this.fixes == null)
                this.fixes = new FixesCollection();
            this.Fixes.AppendXml(fpElement);
            /////////////////////////////////////////////////////////////////////////////

            // 3 - Append the Emergency fight plans
            if (this.emergencies == null)
                this.emergencies = new EmergencyCollection();
            this.Emergencies.AppendXml(fpElement);
            /////////////////////////////////////////////////////////////////////////////

            // 4 - Main Flight Plan + Stages
            XmlHelper.AddComment(fpElement, "Main Flight Plan");
            XmlElement mainFPElement = XmlHelper.AddElement(fpElement, "MainFP", null);

            mainFPElement.SetAttribute("id", this.id);
            XmlHelper.AddElement(mainFPElement, "name",        this.name);
            XmlHelper.AddElement(mainFPElement, "description", this.description);

            if (_stages == null)
                _stages = new StageList();
            this.stages.AppendXml(mainFPElement);

            /////////////////////////////////////////////////////////////////////////////

            // return the xml element
            return fpElement;
        }

        public Waypoint[] GetAllWayPoints()
        {
            List<Waypoint> points = new List<Waypoint>();

            foreach (Fix f in this.fixes)
            {
                Waypoint w = new Waypoint(null, null, f, false, null);
                points.Add(w);
            }

            foreach (Stage s in this.stages)
            {
                foreach (Leg l in s.legsCol)
                {
                    if (l.Destination != null)
                    {
                        points.Add(l.Destination);
                    }
                }
            }

            return points.ToArray();
        }

        #endregion
    }

}
