
using System;
using System.IO;

namespace FlightPlan
{
	/// <summary>
	/// Stores latitude and longitude information. 
	/// </summary>
	public class GPSCoords
    {
        #region - Enums (Direction) -

        /// <summary>
		/// Directions for determining the quadrants the coordinates belong to.
		/// </summary>
		public enum Direction 
        {
			/// <summary>North</summary>
			North,

			/// <summary>East</summary>
			East,

			/// <summary>South</summary>
			South,

			/// <summary>West</summary>
			West
        };

        #endregion

        #region - Static -

        /// <summary>
        /// Converts a direction to a char value.
        /// </summary>
        /// <param name="d">A <see cref="Direction"/> with the direction.</param>
        /// <returns>A <see cref="System.Char"/> with possible values: 'N', 'E', 'S' and 'W'.</returns>		
        public static char DirectionToChar(Direction d)
        {
            switch (d)
            {
                case Direction.North: return 'N';
                case Direction.East: return 'E';
                case Direction.South: return 'S';
                case Direction.West: return 'W';
                default: throw new System.Exception("Executing unreachable case!");
            }
        }

        /// <summary>
        /// Converts a char to a direction value. 
        /// </summary>
        /// <param name="c">A <see cref="System.Char"/> with possible values: 'N', 'E', 'S' and 'W' and their corresponding lowercase characters.</param>
        /// <returns>A <see cref="Direction"/></returns>		
        public static Direction CharToDirection(char c)
        {
            switch (c)
            {
                case 'n':
                case 'N': return Direction.North;
                case 'e':
                case 'E': return Direction.East;
                case 's':
                case 'S': return Direction.South;
                case 'w':
                case 'W': return Direction.West;
                default:
                    string msg = c + " does not indicate valid direction (N/E/S/W)";
                    throw new System.Exception(msg);
            }
        }

        /// <summary>
        /// Converts a char to a direction value. Checks that the char is consistent
        /// with the given isLatitude indication. 
        /// </summary>
        /// <param name="c">A <see cref="System.Char"/> with possible values: 'N', 'E', 'S' and 'W' and their corresponding lowercase characters.</param>
        /// <param name="isLatitude">A <see cref="System.Boolean"/> equaling true if the char corresponds to a latitude value, i.e. north or south.</param>
        /// <returns>A <see cref="Direction"/></returns>
        public static Direction CharToDirection(char c, bool isLatitude)
        {
            Direction d = CharToDirection(c);
            if (isLatitude && d != Direction.North && d != Direction.South)
            {
                string msg = c + " does not indicate a latitude direction (N/S)";
                throw new System.Exception(msg);
            }

            if (!isLatitude && d != Direction.East && d != Direction.West)
            {
                string msg = c + " does not indicate a longitude direction (E/W)";
                throw new System.Exception(msg);
            }

            return d;
        }

        #endregion

        #region - Attributes -

        private int       _degrees;
		private int       _mins;
		private double    _secs;
		private Direction _dir;

        #endregion
        #region - Constructor -

        /// <summary>
		/// Creates a GPSCoords object.
		/// </summary>
		/// <param name="deg">A <see cref="System.Int32"/> with the degrees of the coordinate.</param>
		/// <param name="mins">A <see cref="System.Int32"/> with the minutes of the coordinate.</param>
		/// <param name="secs">A <see cref="System.Double"/> with the seconds of the coordinate.</param>
		/// <param name="dir">A <see cref="Direction"/> with the direction of the coordinate.</param>
		public GPSCoords( int deg, int mins, double secs, Direction dir )
		{
			// Check values
			if (deg  < 0 || deg > 360) throw new System.Exception("Degrees must be between 0 and 360. Found " + deg  + ".");
			if (mins < 0 || mins > 60) throw new System.Exception("Minutes must be between 0 and  60. Found " + mins + ".");
			if (secs < 0 || secs > 60) throw new System.Exception("Seconds must be between 0 and  60. Found " + mins + ".");
			
			_degrees = deg;
			_mins    = mins;
			_secs    = secs;
			_dir     = dir;
		}        
        
		/// <summary>
		/// Creates a GPSCoords object given its value as a single signed double.
		/// </summary>
		/// <param name="val">A <see cref="System.Double"/> with the value of the coordinate.</param>
		/// <param name="IsLatitude">A <see cref="System.Boolean"/> true if this coordinate refers to a latitude. False otherwise.</param>		
		public GPSCoords( double val, bool IsLatitude )
		{
			double m, s;
			
			if (IsLatitude)	_dir = (val > 0 ? Direction.North : Direction.South);
			else			_dir = (val > 0 ? Direction.East  : Direction.West);
			
			s = Math.Abs(val*3600.0);
			_secs = s % 60;
			s -= _secs;
			
			m = s/60.0;
			_mins = (int) m % 60;
			m -= _mins;
			
			_degrees = (int) m / 60;
		}

        #endregion
        #region - Properties -

        /// <summary>
        /// Coordinate degrees.
        /// </summary>
        public int degrees
        {
            get { return _degrees; }
        }

        /// <summary>
        /// Coordinate minutes.
        /// </summary>
        public int mins
        {
            get { return _mins; }
        }

        /// <summary>
        /// Coordinate seconds.
        /// </summary>
        public double secs
        {
            get { return _secs; }
        }

        /// <summary>
        /// Coordinate direction.
        /// </summary>
        public Direction dir
        {
            get { return _dir; }
        }

        #endregion
        #region - Public Functions -

        /// <summary>
        /// Returns a deep copy of a GPSCoords object
        /// </summary>
        /// <returns></returns>
        public GPSCoords DeepCopy()
        {
            GPSCoords res = this.MemberwiseClone() as GPSCoords;
            return res;
        }		
		
		/// <summary>
		/// Returns coordinate as a double.
		/// </summary>
		/// <returns>A <see cref="System.Double"/></returns>		
		public double AsDouble()
		{
			int sign = (dir == Direction.North || dir == Direction.East? 1 : -1);
									
			return sign * (degrees + ((double)mins)/60.0 + secs/3600.0);
		}
		
		/// <summary>
		/// Returns coordinate as a string value.
		/// </summary>
		/// <returns>A <see cref="System.String"/></returns>		
		public string AsString()
		{	
			return String.Format("{0}°{1}'{2:#.##}\"{3}", degrees, mins, secs, GPSCoords.DirectionToChar(dir));
        }

        #endregion
    }
}
