using System;
using System.IO;
using System.Collections;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.Reflection;
using log4net;
using System.Xml;

namespace FlightPlan
{
	//-------------------------------------------------------------------------
	/// <summary>
	/// A collection of legs. 
	/// </summary>
	//-------------------------------------------------------------------------
	public class LegsCollection : KeyedCollection<string, Leg>
    {
        #region - Constructor -

        /// <summary>
		/// Creates a LegsCollection object.
		/// </summary>
		public LegsCollection() : base() {}

        #endregion
        #region - Public Functions -

        /// <summary>
		/// Implements GetKeyForItem abstract method.
		/// </summary>
		/// <param name="item">A <see cref="Leg"/></param>
		/// <returns>A <see cref="System.String"/></returns>
	    protected override string GetKeyForItem(Leg item)
	    {
	        return item.id;
	    }

        #endregion
        #region - Xml -

        /// <summary>
        /// Appends this Object to a xml Stage representation.
        /// </summary>
        /// <param name="stageElement">The Stage XmlElement representation where to append this Object.</param>
        public void AppendXml(XmlElement stageElement)
        {
            // Test for errors            
            if (stageElement == null) throw new ArgumentNullException("stageElement");

            // Create the element
            XmlElement xmlStages = XmlHelper.AddElement(stageElement, "legs", null);

            // Append all the legs
            foreach (Leg l in this)
                l.AppendXml(xmlStages);
        }

        #endregion
	
	}	

	//-------------------------------------------------------------------------
	/// <summary>
	/// Base class for all leg types. 
	/// </summary>
	//-------------------------------------------------------------------------
	public class Leg
	{
        #region - Attributes & Properties -

        /// <summary>
        /// log4j logger
        /// </summary>
        protected static readonly ILog log = LogManager.GetLogger(MethodBase.GetCurrentMethod().DeclaringType);

        /// <summary>
		/// Leg id.
		/// </summary>
		public string id;        
        public string ID
        {
            get { return id;  }
            set { id = value; }
        }

		/// <summary>
		/// Leg destination waypoint. 
		/// </summary>
		public Waypoint destination;
        public Waypoint Destination
        {
            get { return destination;  }
            set { destination = value; }
        }

		/// <summary>
		/// Leg parent.
		/// </summary>
		public Leg parent;
        public Leg Parent
        {
            get { return parent; }
            set { parent = value; }
        }

		/// <summary>
		/// Next leg.
		/// </summary>
		public Leg next;        
        public Leg Next
        {
            get
            {
                if (stage.GetIndexOfFinal(this) >= 0)
                    return stage.GetNextToFinal(this);
                else
                    return next;
            }
            set { next = value; }
        }

		/// <summary>
		/// Previous leg.
		/// </summary>
		public Leg prev;
        public Leg Prev
        {
            get { return stage.GetPrev(this); }
            set { prev = value; }
        }

		/// <summary>
		/// Contingency list to be followed if an emergency occurs when executing this leg. 
		/// </summary>
		public ContingencyList emergency;
		
		/// <summary>
		/// Stage the leg belons to.
		/// </summary>
		public Stage stage;	
		
		/// <summary>
		/// Flight plan the leg belongs to.
		/// </summary>
		public FlightPlan FlightPlan 
        {
			get { return stage.flightPlan; }
		}

		/// <summary>
		/// Returns depth of this leg in the tree structure.
		/// First leg below a stage is at depth 0, a leg contained under a leg
		/// which is directly below stage level is at depth 1, and so on.
		/// </summary>	
		public int Depth 
        {
			get 
            {
				if (parent != null)
					return parent.Depth + 1;
				else
					return 0;
			}
		}

		/// <summary>
		/// Next leg id.
		/// </summary>
		public string nextid;
        public string Nextid
        {
            get { return nextid;  }
            set { nextid = value; }
        }

		/// <summary>
		/// Previous leg id.
		/// </summary>
		public string previd;
        public string Previd
        {
            get { return previd; }
            set { previd = value; }
        }

		/// <summary>
		/// Possible values for turn direction.
		/// </summary>
		public enum TurnDir 
        {
			/// <summary>
			/// Turn left. 
			/// </summary>
			Left, 
			/// <summary>
			/// Turn right.
			/// </summary>
			Right
        };

        #endregion
        #region - Constructors -

        //---------------------------------------------------------------------
		/// <summary>
		/// Creates a leg object. 
		/// </summary>
		//---------------------------------------------------------------------		
		public Leg(Stage s, string id, Waypoint destination, string nextid, string previd, ContingencyList emergency)
		{
			stage = s;
		
			this.id = id;
			this.destination = destination;
			this.nextid = nextid;
			this.previd = previd;
			this.emergency = emergency;
			
			// Set this leg as destination waypoint parent
			if (destination != null)
				destination.SetParent(this);
			
			// This attributes will be updated once all legs have been read
			parent = null;
			next   = null;
			prev   = null;
		}

        #endregion
        #region - Public Functions -

        //--------------------------------------------------------------------- 
		/// <summary>
		/// Compare that both legs are the same based on their ids. 
		/// </summary>
		/// <param name="l">A <see cref="Leg"/>.</param>
		/// <returns>True if they are the same, false otherwise.</returns>
		//---------------------------------------------------------------------
		public bool CompareIds(Leg l)
		{
			//log.DebugFormat("Comparing legs {0} and {1}.", l.id, id);
			
			bool same_leg   = (l.id == id);
			bool same_stage = (l.stage.id == stage.id);
			bool same_plan  = (l.stage.flightPlan.id == stage.flightPlan.id);
			
			return same_leg && same_stage && same_plan;
		}
	
		//---------------------------------------------------------------------
		/// <summary>
		/// Sets field indicated by field to the value in value. 
		/// </summary>
		/// <remarks>
		/// Used during updates.
		/// </remarks>
		/// <param name="name">A <see cref="System.String"/> with the name of the field we want to update. </param>
		/// <param name="value">A <see cref="System.Object"/> with the new value for the field.</param>
		//---------------------------------------------------------------------
		public virtual void SetField(string name, object value)
		{
			Type typeObject = this.GetType();
			FieldInfo field = typeObject.GetField(name);
			if (field == null)
				log.ErrorFormat("Field {0} of leg {1} not public or not found.",
				                name, id);
			else
			{
				string fieldTypeName = field.FieldType.FullName;
				log.InfoFormat("Setting field {0} ({1}) of leg {2}.", 
				               name, fieldTypeName, id);
				
				// if new value for waypoint set its parent first
				if (fieldTypeName.Equals("FlightPlan.Waypoint"))
					(value as Waypoint).SetParent(this);
				
				field.SetValue(this, value);
			}
		}
		
		//---------------------------------------------------------------------
		/// <summary>
		/// Converts turn direction to char. 
		/// </summary>
		/// <param name="td">A <see cref="TurnDir"/> value.</param>
		/// <returns>A <see cref="System.Char"/> ('L' or 'R') indicating a turn direction.</returns>
		//---------------------------------------------------------------------
		public static char TurnDirToChar(TurnDir td)
		{
			     if (td == TurnDir.Left)  return 'L';
			else if (td == TurnDir.Right) return 'R';
			else throw new System.Exception("Invalid turn direction.");
		}

		//---------------------------------------------------------------------
		/// <summary>
		/// Converts char to turn direction. 
		/// </summary>
		/// <param name="cd">A <see cref="System.Char"/> indicating a turn direction ('l', 'L', 'r' or 'R').</param>
		/// <returns>A <see cref="TurnDir"/> value.</returns>
		//---------------------------------------------------------------------
		public static TurnDir CharToTurnDir(char cd)
		{
			switch (cd) 
            {
				case 'l':
				case 'L': return TurnDir.Left;
				case 'r':
				case 'R': return TurnDir.Right;
				default : throw new System.Exception("Invalid turn direction.");
			}
		}

		//--------------------------------------------------------------------- 
		/// <summary>
		/// Returns the inverse of the given turn direction. 
		/// </summary>
		//---------------------------------------------------------------------
		public static TurnDir TurnDirInverse(TurnDir td)
		{
			if (td == TurnDir.Right) return TurnDir.Left;
            else                     return TurnDir.Right;
		}

		//---------------------------------------------------------------------
		/// <summary>
		/// Checks whether the leg depends on a given condition.
		/// </summary>
		/// <remarks>
		/// Must be overriden in the leg classes that support conditions. 
		/// </remarks>
		/// <param name="id">A <see cref="System.String"/> with the condition identifier.</param>
		/// <returns>True if the leg depends on the condition, false otherwise.</returns>
		//---------------------------------------------------------------------
		public virtual bool DependsOnCondition(string id)
		{
			return false;
		}
		
		//---------------------------------------------------------------------
		/// <summary>
		/// Returns true if the leg depends on a condition that has not been
		/// initialized.
		/// </summary>
		/// <remarks>
		/// Must be overriden in intersection legs. In the future maybe in other
		/// legs too.
		/// </remarks>
		//---------------------------------------------------------------------
		public virtual bool AwaitsConditionResult()
		{
			return false;
		}
		
		// Used to determine if two waypoints correspond to the exact same
		// position
		private readonly static double epsilon = Math.Pow(1, -10);

		//---------------------------------------------------------------------
		/// <summary>
		/// Returns true if wp is the last waypoint of the current leg.
		/// The comparison is made considering coordinates and altitude.
		/// </summary>
		/// <param name="wp">A <see cref="Waypoint"/></param>
		/// <returns>A <see cref="System.Boolean"/></returns>
		//---------------------------------------------------------------------
		public bool IsLastWaypoint(Waypoint wp)
		{
			// if the waypoint belongs to another leg it cannot be the las wp
			// of this one.
			if (wp.leg != this)
				return false;
			
			// Check if this is the last waypoint of the current leg
			double dist;
			double auxc;
			wp.Distance(wp.leg.destination, out auxc, out dist);
			double alt_dist;
			if (wp.leg.destination.altitude.HasValue)
				alt_dist = (wp.leg.destination.altitude.Value - wp.altitude.Value);
			else
				alt_dist = 0;
			
			return (Math.Abs(dist) < epsilon &&  Math.Abs(alt_dist) < epsilon);			
		}
		
        /// <summary>
        /// A string that represents the current object.
        /// </summary>
        /// <returns></returns>
		public override string ToString()
		{
			string result = "";
			
			result = result + "leg id: " + id + "\n";
			result = result + "type: " + this.GetType().FullName + "\n";
			if (next != null)
				result = result + "next: " + next.id + "\n";		
			
			return result;
		}

        /// <summary>
        /// Returns a deep copy of a leg object. All references to other objects
        /// are set to null, except for stage and destination waypoint. The latter
        /// is actually deeply copied.
        /// </summary>
        /// <returns></returns>
        public virtual Leg DeepCopy(Stage s)
        {
            Waypoint dest = null;
            String nid        = null;
            String pid        = null;
            ContingencyList emer = null;

            if (this.destination != null) dest = this.destination.DeepCopy(s.flightPlan.Fixes);            
            if (this.nextid      != null) nid  = String.Copy(this.nextid);            
            if (this.previd      != null) pid  = String.Copy(this.previd);
            if (this.emergency   != null) emer = this.emergency.DeepCopy();

            Leg l = new Leg(s, String.Copy(this.id), dest, nid, pid, emer);

            l.stage = s;

            return l;
        }

        #endregion
        #region - Xml -

        /// <summary>
        /// Gets the XsiType for this Leg.
        /// </summary>
        protected virtual String XsiType
        {
            get { throw new Exception("This property must be implemented in the extended class."); }
        }

        /// <summary>
        /// Appends this Object to a xml LegsCollection representation.
        /// </summary>
        /// <param name="legsElement">The LegsCollection XmlElement representation where to append this Object.</param>
        /// <returns>The xml child appended.</returns>
        public virtual XmlElement AppendXml(XmlElement legsElement)
        {
            // Test for errors
            if (legsElement == null) throw new ArgumentNullException("legsElement");

            // Creates the element            
            XmlElement xmlLeg = legsElement.OwnerDocument.CreateElement("leg");

            // id + type ///////////////////////////////////////////////////////////////
            xmlLeg.SetAttribute("id", this.ID);
            xmlLeg.SetAttribute("type", XMLFPParser.xs_instance_ns, XsiType);
            ////////////////////////////////////////////////////////////////////////////

            // Destination /////////////////////////////////////////////////////////////
            if (this.Destination != null)
            {
                XmlElement dest = XmlHelper.AddElement(xmlLeg, "dest", null);
                this.destination.AppendXml(dest);
            }
            ////////////////////////////////////////////////////////////////////////////

            // Prev ////////////////////////////////////////////////////////////////////            
            if (this.previd != null && this.previd != "")
            {
                XmlHelper.AddElement(xmlLeg, "prev", this.previd);
            }
            ////////////////////////////////////////////////////////////////////////////

            // Next ////////////////////////////////////////////////////////////////////            
            if (this.nextid != null && this.nextid != "")
            {
                XmlHelper.AddElement(xmlLeg, "next", this.nextid);
            }
            ////////////////////////////////////////////////////////////////////////////

            // Append the elements
            legsElement.AppendChild(xmlLeg);

            return xmlLeg;
        }

        #endregion
    }
}
