// LegHelper.cs created with MonoDevelop
// User: esantama at 3:22 PM 10/9/2008

using System;
using System.Reflection;
using System.Collections;
using log4net;

namespace FlightPlan
{
	/// <summary>
	/// This class provides a temporary repository to store leg related 
	/// data. It also helps to validate it before creating and initializing
	/// a new leg instance.
	/// </summary>
	/// <remarks>
	/// To add a new leg type, add additional attributes specific to the new
	/// type to the LegHelper class. Update constructor, IsValidType and
	/// CreateLeg.
	/// </remarks>
	public class LegHelper
    {
        #region - Attributes -

        private static readonly ILog log = LogManager.GetLogger(MethodBase.GetCurrentMethod().DeclaringType);
		
		// Attributes commmon to all leg types
		/// <summary>
		/// Leg id 
		/// </summary>
		public string id = null;
		/// <summary>
		/// Leg destination
		/// </summary>
		public Waypoint destination = null;
		/// <summary>
		/// Next leg
		/// </summary>
		public string next = null;
		/// <summary>
		/// Previous leg
		/// </summary>
		public string prev = null;
		/// <summary>
		/// Reference to emergency flight plan
		/// </summary>
        public ContingencyList emergency = null;
        public Contingency contingency = null;

		// Attributes for RF legs and holding patterns
		/// <summary>
		/// Latitude of turn center
		/// </summary>
		public GPSCoords centerLat = null;
		/// <summary>
		/// Longitude of turn center
		/// </summary>
		public GPSCoords centerLon = null;
		/// <summary>
		/// Turn direction
		/// </summary>
		public Leg.TurnDir? direction = null;

        public Waypoint rwyhead = null;
        public double? length = null;
        public Waypoint hwp1 = null;
        public Waypoint hwp2 = null;
        public Waypoint hwp3 = null;
        public Waypoint hwp4 = null;
        public Waypoint iwp = null;
        public Waypoint idwp = null;
        public Waypoint ibwp = null;
        public Waypoint ifwp = null;
        public Waypoint tdwp = null;
        public Waypoint bpwp = null;
        public Waypoint lawp = null;
        public Waypoint eawp = null;
        public string nextAbortId = null;
        public string dvrtFrust = null;
        public string startOP = null;

        public Waypoint der = null;
        public Waypoint edwp_a = null;
        public Waypoint edwp_a_min = null;
        public Waypoint edwp_bl = null;
        public Waypoint edwp_bl_min = null;
        public Waypoint edwp_br = null;
        public Waypoint edwp_br_min = null;
        public Waypoint edwp_cl = null;
        public Waypoint edwp_cr = null;
        public double? rda_w = null;
        public double? rda_l = null;
        public string endOP = null;

        public double? dimN = null;
        public double? dimS = null;
        public double? dimE = null;
        public double? dimW = null;

		// TODO: update according to flight plan schema (.xsd)
		// TODO: when doing so check all places where holding patterns are being used,
		//       e.g. basic scan
		
		// Attributes specific to holding patterns
		
		/// <summary>
		/// Course of holding pattern
		/// </summary>
		public double? course = null;		
		/// <summary>
		/// Distance between turn centers
		/// </summary>
		public double? d1 = null; // d1
		/// <summary>
		/// Diameter of turn
		/// </summary>
		public double? d2 = null; // d2
		
		// Attributes specific to Iterative legs
		/// <summary>
		/// First leg of iterative leg body.
		/// </summary>
		public string first = null;
		/// <summary>
		/// Last leg of iterative leg body.
		/// </summary>
		public string last = null;
		/// <summary>
		/// Number of allowed iterations.
		/// </summary>
		public int? upperBound = null;
		/// <summary>
		/// Legs that form the body of the iterative leg.
		/// </summary>
		public ArrayList body = null;
		
		// Attributes specific to Intersection legs
		/// <summary>
		/// Condition to choose next leg.
		/// </summary>
		public string nextCond = null;
		/// <summary>
		/// Possible next legs.
		/// </summary>
		public ArrayList nextList = null;
		/// <summary>
		/// Condition to choose previous leg.
		/// </summary>
		public string prevCond = null;
		/// <summary>
		/// Possible previous legs.
		/// </summary>
		public ArrayList prevList = null;

		// Attributes specific to BasicScan legs
		/// <summary>
		/// Length of scan area.
		/// </summary>
		public double? dim1 = null;
		/// <summary>
		/// Width of scan area.
		/// </summary>
		public double? dim2 = null;
		/// <summary>
		/// Angle that determines scan area orientation.
		/// </summary>
		public double? angle = null;
		/// <summary>
		/// Separation between passes
		/// </summary>
		public double? separation = null;
		// d1, d2 and turndirection as in holding pattern
		/// <summary>
		/// Position where to start executing the scan.
		/// </summary>
		public GPSCoords[] startAt;
		
		// Mapping between LegHelper fields and leg class fields.
		// Entries follow this form: {LegHelper field name, Leg class, leg class field name}
		// Fields not listed in this table cannot be updated (an exception will be raised).
		private string[,] update_table = new string[,] 
        {
			// DFLeg
			{"destination", "DFLeg",    "destination"},
			// BasicScan
			{"dim1",       "BasicScan", "dim1"    },
			{"dim2",       "BasicScan", "dim2"    },			
			{"angle",      "BasicScan", "angle"   },
			{"separation", "BasicScan", "separation" },
			{"startAt",    "BasicScan", "startAt" },
			// ScanPoint
			{"destination", "ScanPoint", "destination"},
			{"course",      "ScanPoint", "course" },
			{"d1",          "ScanPoint", "d1"     },
			{"d2",          "ScanPoint", "d2"     }
		};

        #endregion
        #region - Constructor -

        /// <summary>
		/// Constructor. 
		/// </summary>
		/// <param name="id">A <see cref="System.String"/>With the id of the leg we are using LegHelper to collect data for.</param>		
		public LegHelper(string id)
		{
			this.id = id;
		}

        #endregion
        #region - Public -

        /// <summary>
		/// Validates that all necessary data is available for creating a leg
		/// of the given type. Creates leg and returns it.
		/// </summary>
		/// <remarks>
		/// To add a new leg type create new branch specifically for the new
		/// type. Ensure that all parameters required for the new type are
		/// available. Create instance and return.
		/// </remarks>
		/// <param name="stg"><see cref="Stage"/> the new leg will belong to.</param>
		/// <param name="type">A <see cref="System.String"/> indicating the type of the leg to be created.</param>
		/// <returns>A new <see cref="Leg"/> of the given type initialized with the data held by the LegHelper.</returns>		
		public Leg CreateLeg(Stage stg, string type)
		{
			// Remove prefix
			type = type.Substring(type.IndexOf(':')+1); 
			
			// Initialize message just in case.
			string msg = "Insufficient or inconsistent data for creating " + type + " leg named " + id + ".";

			// Initial Fix
			if (type == "IFLeg") 
            {
				bool ok = (id != null) && (destination != null);
				if (!ok) throw new XMLFPException(msg);

				return new IFLeg(stg, id, destination, next, prev, emergency);
			}

			// Track to Fix, Direct to Fix
			if (type == "TFLeg" || type == "DFLeg")             
            {
				bool ok = (id != null) && (destination != null);
				if (!ok) throw new XMLFPException(msg);

				if (type == "TFLeg")
					return new TFLeg(stg, id, destination, next, prev, emergency);
				else
					return new DFLeg(stg, id, destination, next, prev, emergency);
			}

			// Radius to Fix
			if (type == "RFLeg") 
            {
				bool ok = (id != null) && (destination != null) && 
// TODO: uncomment?				          ((next != null) || (prev!=null)) &&
						  (centerLat != null) && (centerLon != null) &&
				          (direction != null);
				if (!ok) throw new XMLFPException(msg);

				return new RFLeg(stg, id, destination, next, prev, emergency,
					centerLat, centerLon, direction.Value);
			}
			// Hold to Fix
			if (type == "HFLeg") 
            {
				bool ok = (id != null) && (destination != null) && 
// TODO: uncomment?				          ((next != null) || (prev!=null)) &&
						  (d1 !=  null) && (d2 !=  null) &&					
				          (direction != null) && (course != null);
				if (!ok) throw new XMLFPException(msg);

				return new HFLeg(stg, id, destination, next, prev, emergency,				                 
				    course.Value, direction.Value, d1.Value, d2.Value,
				    (upperBound == null? 1:upperBound.Value), nextCond);
			}
            // Landing Leg
            if (type == "LALeg")
            {
                bool ok = (id != null) && 
                    // TODO: uncomment?				          ((next != null) || (prev!=null)) &&
                            (rwyhead != null) && (course != null) && (length != null) &&
                            (hwp1 != null) && (hwp2 != null) && (hwp3 != null) && (hwp4 != null) && (iwp != null) &&
                            (idwp != null) && (ibwp != null) && (ifwp != null) && (tdwp != null) && (bpwp != null) &&
                            (lawp != null) && (eawp != null) && (dvrtFrust != null) && (startOP != null);
                if (!ok) throw new XMLFPException(msg);

                return new LALeg(stg, id, destination, next, prev,
                emergency, rwyhead, course.Value, length.Value, hwp1, hwp2, hwp3,
                hwp4, iwp, idwp, ibwp, ifwp, tdwp, bpwp, lawp, eawp, nextAbortId, dvrtFrust, startOP);
            }
            // Takeoff Leg
            if (type == "TOLeg")
            {
                bool ok = (id != null) &&
                    // TODO: uncomment?				          ((next != null) || (prev!=null)) &&
                    //		(endOP == null || nextList.Count>0) &&
                            (rwyhead != null) && (course != null) && (length != null) &&
                            (der != null) && (edwp_a != null) && (edwp_a_min != null) && (edwp_bl != null) && (edwp_bl_min != null) &&
                            (edwp_br != null) && (edwp_br_min != null) && (edwp_cl != null) && (edwp_cr != null) &&
                            (rda_w != null) && (rda_l != null) && (endOP != null);
                if (!ok) throw new XMLFPException(msg);

                return new TOLeg(stg, id, destination, next, prev,
                emergency, rwyhead, course.Value, length.Value, der, edwp_a, edwp_a_min,
                edwp_bl, edwp_bl_min, edwp_br, edwp_br_min, edwp_cl, edwp_cr, rda_w.Value, rda_l.Value, 
                nextList, endOP);
            }
            // Initial Ramp Fix
            if (type == "IRFLeg")
            {
                bool ok = (id != null) && (destination != null) &&
                            (dimN != null) && (dimS != null) && (dimE != null) && (dimW != null) && (course != null);
                if (!ok) throw new XMLFPException(msg);

                return new IRFLeg(stg, id, destination, next, prev, emergency, course.Value, dimN.Value, dimS.Value, dimE.Value, dimW.Value);
            }
            // Iterative Leg
			if (type == "IterativeLeg") 
            {
				bool ok = (id != null) && //(destination != null) && 
					    (first != null) && (last != null) &&
						(upperBound != null) && (upperBound.Value >= 0) && 
						(body.Count > 0);
				if (!ok) throw new XMLFPException(msg);

				return new IterativeLeg(stg, id, next, prev, emergency,
					body, upperBound.Value, nextCond, first, last);
			}
			// Intersection Leg
			if (type == "IntersectionLeg")
            {
				bool ok = (id != null);
//					      (nextCondId == null || nextList.Count>0) &&
//						  (prevCondId == null || prevList.Count>0);
				if (!ok) throw new XMLFPException(msg);

				return new IntersectionLeg(stg, id, destination, next, prev, emergency,
				                           nextCond, nextList, 
				                           prevCond, prevList);
			}
			// BasicScanLeg
			if (type == "BasicScanLeg") 
            {
                bool ok = (id != null) && (destination != null) &&
                          (dim1 != null) && (dim2 != null) && (angle != null) && (separation != null);
                          // && (d1 !=  null) && (d2 !=  null) && (direction != null);
				if (!ok) throw new XMLFPException(msg);

				return new BasicScan(stg, id, destination, next, prev, emergency,
				                     dim1.Value, dim2.Value, angle.Value,
				                     separation.Value, direction, d1, d2, startAt);
			}
			// ScanPointLeg
			if (type == "ScanPointLeg")
            {				
				bool ok = (id != null) && (destination != null) && 
					(d1 !=  null) && (d2 !=  null) && (course != null);
				if (!ok) throw new XMLFPException(msg);

				return new ScanPoint(stg, id, destination, next, prev, emergency,				                 
				                     course.Value, d1.Value, d2.Value); 
			}
			// No valid leg type
			else 
            {
				throw new XMLFPException(type + " is not a valid leg type.");
			}
		}

        #endregion
        #region - Private -

        /// <summary>
        /// Searchs for the given emergency plan and returns its reference.
        /// </summary>
		private FlightPlan getEmergencyPlan(string emergency)
		{
			if (emergency != null) 
            {
				// TODO: search for emergency flight plan and initialize reference
				return null;
			}
			else
				return null;
		}

        /// <summary>
        //// Given a LegHelper field name and a leg class name, looks up in a
		/// mappings table and returns the name of the corresponding field in
		/// the given leg class.
        /// </summary>
		private string FindLegField(string helperField, string legClass)
		{
			for (int i=0; i<update_table.GetLength(0); i++)
			{
				if ((update_table[i,0] == helperField) && (update_table[i,1] == legClass))
						return update_table[i,2];
			}
			return null;
		}

		/// <summary>
		/// Once a LegHelper object has been initialized with data coming from
		/// a flight plan update message, this method can be used to perform
		/// the update of the flight plan leg.
		/// </summary>
		/// <remarks>
		/// Similar to CreateLeg but with less restrictions on what data
		/// should be available. 
		/// </remarks>
		/// <param name="l"><see cref="Leg"/> to be updated.</param>
		public void UpdateLeg(Leg l)
		{
			// LegHelper fields			
			FieldInfo[] fields = this.GetType().GetFields();
			
			// Name of actual type of l
			string legTypeName = l.GetType().Name;
			
			// Look for non-null fields, i.e. fields that have been given a value in
			// the update message.
			for (int i=0; i<fields.Length; i++)
			{
				if (fields[i].GetValue(this) != null) {
					// Look for field name of leg
					string legField = FindLegField(fields[i].Name, legTypeName);
					if (legField == null) {
						string msg = String.Format("field {0} of {1} leg {2} cannot be updated.",
						                           fields[i].Name, legTypeName, (l as Leg).id);
						log.Error(msg);
						throw new XMLFPException(msg);
					}
					else
					{
						log.Info(String.Format("updating field {0} of {1} leg {2}.", 
						                       legField, legTypeName, (l as Leg).id));
						l.SetField(legField, fields[i].GetValue(this));							
					}
				}
			}
        }

        #endregion
    }
}
