using System;
using System.IO;
using System.Collections;
using System.Collections.Generic;
using System.Reflection;

using log4net;
using System.Xml;

namespace FlightPlan
{
    /// <summary>
	/// Randomly accessed list of stages. 
	/// </summary>
	public class StageList : List <Stage>
    {
        #region - Constructor -

        /// <summary>
		/// Creates and StageList object.
		/// </summary>
        public StageList() : base() { }

        #endregion
        #region - Public Functions -

        /// <summary>
        /// Return a deep copy of the list of stages
        /// </summary>
        /// <returns></returns>
        public StageList DeepCopy(FlightPlan fp)
        {
            StageList l = new StageList();
            foreach (Stage s in this)
                l.Add(s.DeepCopy(fp));
            return l;
        }

        #endregion
        #region - Xml -

        /// <summary>
        /// Appends this Object to a xml Main FlightPlan representation.
        /// </summary>
        /// <param name="mainFpElement">The Main FlighPlan XmlElement representation where to append this Object.</param>
        public void AppendXml(XmlElement mainFpElement)
        {
            // Test for errors            
            if (mainFpElement == null) throw new ArgumentNullException("fpElement");

            // Create the elements
            XmlElement xmlStages = XmlHelper.AddElement(mainFpElement, "stages", null);

            // Append all the fixes
            foreach (Stage s in this)
                s.AppendXml(xmlStages);
        }

        #endregion
	}
	
	/// <summary>
	/// Represents a flight plan stage.
	/// </summary>
	/// <remarks>A stage groups flight plan legs that belong to it.</remarks>
	public class Stage		
	{
		private static readonly ILog log = LogManager.GetLogger(MethodBase.GetCurrentMethod().DeclaringType);

        #region - Enums (Type) -

        /// <summary>
		/// Stage types.
		/// </summary>
		public enum Type 
        {
			/// <summary>
			/// Taxi stage. 
			/// </summary>
			Taxi, 
			/// <summary>
			/// Take-off stage.
			/// </summary>
			TakeOff, 
			/// <summary>
			/// Departure stage.
			/// </summary>
			Departure, 
			/// <summary>
			/// Route stage.
			/// </summary>
			Route, 
			/// <summary>
			/// Mission stage.
			/// </summary>
			Mission, 
			/// <summary>
			/// Arrival stage.
			/// </summary>
			Arrival, 
			/// <summary>
			/// Approach stage.
			/// </summary>
			Approach, 
			/// <summary>
			/// Land stage.
			/// </summary>
			Land
		};

        #endregion
        #region - Static -

        // Define a bunch of static classes to hold code specific to each type 

		/// <summary>
		/// Code specific to Taxi stage.
		/// </summary>
		static class Taxi
		{
			/// <summary>
			/// Check taxi stage.
			/// </summary>
			/// <param name="s">Taxi <see cref="Stage"/>.</param>
			public static void Check(Stage s)
			{
			}
		}
		
		/// <summary>
		/// Code specific to TakeOff stage.
		/// </summary>
		static class TakeOff
		{
			/// <summary>
			/// Check that take off initial legs are associated to a runway fix. 
			/// </summary>
			public static void Check(Stage s)
			{			
				// TODO: Add whatever checking is required.
				
				// Right now, tests below are more annoying than useful. Skip them.
				return;
				
				foreach (Leg l in s.initialLegs) 
                {
					if (l is IFLeg) 
                    {
						if (!l.destination.isNamed || !(l.destination.fix is RwyFix))
							throw new Exception("Initial fix not bound to a runway fix.");
					}
					else
						throw new Exception("Initial leg of TakeOff stage should be IF.");
						
				}
			}
		}        
        
		/// <summary>
		/// Converts a string stage type to its enumerated counterpart. 
		/// </summary>
		/// <param name="stype">A <see cref="System.String"/> indicating the type.</param>
        /// <returns>A <see cref="Type"/></returns>		
		public static Type StringToType(string stype)
		{
			     if (stype == "Taxi")	   return Type.Taxi;
			else if (stype == "TakeOff")   return Type.TakeOff;
			else if (stype == "Departure") return Type.Departure;
			else if (stype == "Route")	   return Type.Route;
			else if (stype == "Mission")   return Type.Mission;
			else if (stype == "Arrival")   return Type.Arrival;
			else if (stype == "Approach")  return Type.Approach;
			else if (stype == "Land")      return Type.Land;
			else throw new System.Exception(
			                    String.Format("Invalid stage type {0}.", stype));
		}
		
		/// <summary>
		/// Converts a stage type to a string representation. 
		/// </summary>
		/// <param name="t">A <see cref="Type"/></param>
		/// <returns>A <see cref="System.String"/></returns>		
		public static string TypeToString(Type t)
		{
			switch (t) 
            {
			    case Type.Taxi:      return "Taxi";
				case Type.TakeOff:   return "TakeOff";
				case Type.Departure: return "Departure";
				case Type.Route:     return "Route";
				case Type.Mission:   return "Mission";
				case Type.Arrival:   return "Arrival";
				case Type.Approach:  return "Approach";
				case Type.Land:      return "Land";
				default:
					throw new System.Exception("Unknown type.");
			}
		}

        #endregion

        #region - Attributes & Properties -

        /// <summary>
		/// id of the stage.
		/// </summary>
		public string id;
        public string ID
        {
            get { return id;  }
            set { id = value; }
        }

        /// <summary>
		/// Name of the stage.
		/// </summary>
		public string name;
        public string Name
        {
            get { return name; }
            set { name = value; }
        }

		/// <summary>
		/// Description of the stage.
		/// </summary>
		public string description;
        public string Description
        {
            get { return description; }
            set { description = value; }
        }

		/// <summary>
		/// Type of the stage, i.e.:Taxi, TakeOff, Departure, Route, etc.
		/// </summary>
		public Type type;

		/// <summary>
		/// Should this stage be flown only in manual mode?
		/// </summary>
		public bool manualOnly;
        public bool ManualOnly
        {
            get { return manualOnly; }
            set { manualOnly = value; }
        }

        /// <summary>
        /// The ignore flag.
        /// </summary>
        private bool _ignore;
        public  bool Ignore
        {
            get { return _ignore; }
            set { _ignore = value; }
        }


        private String _xmlComment;
        public String XmlComment
        {
            get { return _xmlComment; }
            set { _xmlComment = value; }
        }
		
		private FlightPlan _fp = null;
		/// <summary>
		/// Flight plan the stage belongs to. 
		/// </summary>
		public FlightPlan flightPlan 
        {
			get { return _fp; }
		}
		
		// When reading the flight plan, first all legs will be put into the
		// keyed collection. In a second pass all legs will be revisited to
		// initialize references to prev and nexts legs.
		
		private LegsCollection _legsCol;
		/// <summary>
		/// Collection of legs that belong to this stage.
		/// </summary>
		public LegsCollection legsCol
        {
			get { return _legsCol; }
		}
		
		/// <summary>
		/// Legs that can be the first one of this stage. 
		/// </summary>
		public List <Leg> initialLegs;
		/// <summary>
		/// Legs that cab be the last one of this stage.
		/// </summary>
		public List <Leg> finalLegs;

        String not_found_msg = "Flight plan contains a reference ({0}) to a non-existing leg, check spelling.";

        #endregion
        #region - Constructor -

		/// <summary>
		/// Creates a stage object. 
		/// </summary>
		/// <param name="fp"><see cref="FlightPlan"/> the stage belongs to.</param>		
		public Stage(FlightPlan fp)
		{
			_fp         = fp;
			_legsCol    = new LegsCollection();
			initialLegs = new List <Leg>();
			finalLegs   = new List <Leg>();
            _ignore     = false;
		}

        #endregion
        #region - Public Functions -

        /// <summary>
		/// Check stage for consistency.
		/// </summary>
		public void Check()
		{
			// TODO: complete
			switch (type) 
            {
			    case Type.Taxi:    Stage.Taxi.Check(this);	  break;
			    case Type.TakeOff: Stage.TakeOff.Check(this); break;
			}
		}
		
		/// <summary>
		/// Adds leg to the collection. 
		/// </summary>
		/// <param name="l">A <see cref="Leg"/></param>
		public void Add(Leg l)
		{
			_legsCol.Add(l);
		}
		
		/// <summary>
		/// Clears the list of initial legs. 
		/// </summary>		
		public void ClearInitials()
		{
			initialLegs.Clear();
		}
		
		/// <summary>
		/// Adds an initial leg to the stage. 
		/// </summary>
		/// <param name="legId">A <see cref="System.String"/></param>		
		public void AddInitial(string legId)
		{
            try
            {
                initialLegs.Add(legsCol[legId]);
            }
            catch (KeyNotFoundException e)
            {
                throw new KeyNotFoundException(String.Format(not_found_msg, legId), e);
            }
		}
		
		/// <summary>
		/// Clears the list of final legs. 
		/// </summary>		
		public void ClearFinals()
		{
			finalLegs.Clear();
		}		
		
		/// <summary>
		/// Adds a final leg to the stage. 
		/// </summary>
		/// <param name="legId">A <see cref="System.String"/></param>
		public void AddFinal(string legId)
		{
            try
            {
                finalLegs.Add(legsCol[legId]);
            }
            catch (KeyNotFoundException e)
            {
                throw new KeyNotFoundException(String.Format(not_found_msg, legId), e);
            }
		}

        /// <summary>
        /// Update the following references: parents of all legs; first, last
        /// and body of all iterative ones; nextList and prevList of intersection
        /// legs.
        /// </summary>
        private void UpdateControlLegsReferences()
        {
            log.DebugFormat("Updating parents, first, last and body");

            foreach (Leg l in legsCol)
            {
                log.DebugFormat("\t Leg is {0}", l.id);
                if (l is IterativeLeg)
                {
                    IterativeLeg myIter = l as IterativeLeg;

                    // update parent of all children and add objects to the loop body
                    foreach (string child in myIter.bodyRefs)
                    {
                        try
                        {
                            legsCol[child].parent = myIter;
                            myIter.body.Add(legsCol[child]);
                        }
                        catch (KeyNotFoundException e)
                        {
                            throw new KeyNotFoundException(String.Format(not_found_msg, child), e);
                        }
                    }
                    //  update initial and final legs
                    try
                    {
                        myIter.first = legsCol[myIter.firstId];
                    }
                    catch (KeyNotFoundException e)
                    {
                        throw new KeyNotFoundException(String.Format(not_found_msg, myIter.firstId), e);
                    }
                    try
                    {
                        myIter.last = legsCol[myIter.lastId];
                    }
                    catch (KeyNotFoundException e)
                    {
                        throw new KeyNotFoundException(String.Format(not_found_msg, myIter.lastId), e);
                    }
                }

                if (l is TOLeg)
                {
                    TOLeg myTO = l as TOLeg;

                    // add objects to the next list
                    if (myTO.nextRefs != null)
                    {
                        foreach (string child in myTO.nextRefs)
                            try
                            {
                                if (child != "null")
                                    myTO.nextList.Add(legsCol[child]);
                            }
                            catch (KeyNotFoundException e)
                            {
                                throw new KeyNotFoundException(String.Format(not_found_msg, child), e);
                            }
                    }
                }

                if (l is LALeg)
                {
                }


                if (l is IntersectionLeg)
                {
                    IntersectionLeg myInter = l as IntersectionLeg;

                        // add objects to the next list
                        if (myInter.nextRefs != null)
                            foreach (string child in myInter.nextRefs)
                                try
                                {
                                    myInter.nextList.Add(legsCol[child]);
                                }
                                catch (KeyNotFoundException e)
                                {
                                    throw new KeyNotFoundException(String.Format(not_found_msg, child), e);
                                }

                        // add objects to the prev list
                        if (myInter.prevRefs != null)
                            foreach (string child in myInter.prevRefs)
                                try
                                {
                                    myInter.prevList.Add(legsCol[child]);
                                }
                                catch (KeyNotFoundException e)
                                {
                                    throw new KeyNotFoundException(String.Format(not_found_msg, child), e);
                                }
                }
            }
        }

        /// <summary>
        /// Called from UpdateNextPrevReferences to connect next leg with its previous
        /// </summary>
        private void UpdatePrevReference(Leg current, Leg next)
        {
            // Intersection leg special case
            if (next is IntersectionLeg)
            {
                IntersectionLeg myInter = next as IntersectionLeg;
                if (!myInter.prevList.Contains(current))
                    myInter.prevList.Add(current);

                if (myInter.prevList.Count == 1)
                    myInter.prev = current;
                else
                    myInter.prev = null;
            }
            else
                next.prev = current;
        }

        /// <summary>
        /// Update next and prev references of all legs.
        /// </summary>
        private void UpdateNextPrevReferences()
        {
            log.DebugFormat("Updating next and prev");

            foreach (Leg l in legsCol)
            {
                // next
                if (l is TOLeg)
                {
                    TOLeg tol = l as TOLeg;
                    String s;

                    s = tol.nextRefs[tol.endOP.GetResult()] as string;
                    if (s != "null")
                        tol.nextid = s;
                    else
                        tol.nextid = null;
                   
                }
                if (l.nextid != null)
                {
                    if (finalLegs.Contains(l))
                    {
                        //
                        // Standard leg with next condition assigned or interesction leg will fail being final legs.
                        //
                        string msg = String.Format("Next leg of final leg {0} cannot be explicitly assigned.", l.id);
                        throw new System.Exception(msg);
                    }

                    try
                    {
                        l.next = legsCol[l.nextid];
                        if (l is IntersectionLeg)
                        {
                            LegsCollection nextList = (l as IntersectionLeg).NextList;
                            if (nextList.Count>0 && !nextList.Contains(l.next))
                            {
                                string msg = String.Format("Default leg {0} of {1} not found in next legs list.", l.nextid, l.ID);
                                throw new System.Exception(msg);
                            }                                    
                            foreach (Leg br in (l as IntersectionLeg).NextList)
                                UpdatePrevReference(l, br);
                        }
                        UpdatePrevReference(l, l.next);
                    }
                    catch (KeyNotFoundException e)
                    {
                        throw new KeyNotFoundException(String.Format(not_found_msg, l.nextid), e);
                    }
                }
                else
                {
                    // next is mandatory unless leg is final or intersection
                    if (l is IntersectionLeg)
                    {
                        IntersectionLeg myInter = l as IntersectionLeg;
                        if (myInter.NextList.Count == 1)
                            l.next = myInter.NextList[0];
                    }
                    else if (!finalLegs.Contains(l))
                    {
                        // if leg is last leg of a loop it's ok
                        if (l.parent is IterativeLeg && (l.parent as IterativeLeg).last == l)
                            l.next = (l.parent as IterativeLeg).first;
                        else
                        {
                            string msg = String.Format("Next leg of {0} not specified.", l.id);
                            throw new System.Exception(msg);
                        }
                    }
                }
            }
        }

		/// <summary>
		/// Builds a graph updating leg references (parent, next, prev, etc.).
        /// Intersection legs are the only ones allowed to have more than one
        /// next leg.
		/// The connection between the final legs of the stage and the initial
		/// legs of the next stage is not explicit. This connection is determined
		/// by their position in the finalLegs and initialLegs elements. 
		/// </summary>
		public void UpdateReferences()
		{
			log.DebugFormat("Running UpdateGraph (stage {0})", id);

            UpdateControlLegsReferences();
            UpdateNextPrevReferences();
		}

		//---------------------------------------------------------------------
		// TODO: Check flight plan consistency.
		// * No cycles in non iterative legs.		
		// * Only the last legs of a stage or an iterative leg lack a next leg.
		// * Legs belonging to an iterative leg are not referenced outside the
		//   iterative leg.
		// * None of the initial/final legs has a predecessor/successor.
		//---------------------------------------------------------------------
				
        /// <summary>
        /// Returns next leg of a final leg
        /// </summary>
        /// <param name="l"></param>
        /// <returns></returns>
        public Leg GetNextToFinal(Leg l)
        {
            // If we get here and no next l must be final leg.
            if (finalLegs.IndexOf(l) == -1)
                throw new Exception(
                    String.Format("Leg {0} should be final.", l.ID));

            return flightPlan.GetNextToFinal(this, l);
        }

        /// <summary>
        /// Returns leg previous to l
        /// </summary>
        /// <param name="l"></param>
        /// <returns></returns>
        public Leg GetPrev(Leg l)
        {
            if (l.prev == null)
            {
                // If we get here and no previous l must be initial leg.
                if (initialLegs.IndexOf(l) == -1)
                    throw new Exception(String.Format("Leg {0} should be initial.", l.ID));

                return flightPlan.GetPrevToInitial(this, l);
            }
            else
                return l.prev;
        }

		/// <summary> Returns the indicated initial leg of the stage. </summary>
		/// <param name="il"> Index for choosing the initial leg. </param>
		/// <returns> A Leg. </returns>
		public Leg GetInitialLeg(int il)
		{
			return initialLegs[il];
		}

        /// <summary> Returns the index of the initial leg of the stage. </summary>
        /// <param name="l"> A Leg. </param>
        /// <returns> An int value. </returns>
        public int GetIndexOfInitial(Leg l)
        {
            return initialLegs.IndexOf(l);
        }

        /// <summary> Returns the indicated final of the stage. </summary>
        /// <param name="fl"> Index for choosing the final leg. </param>
        /// <returns> A Leg. </returns>
        public Leg GetFinalLeg(int fl)
        {
            return finalLegs[fl];
        }
		
		/// <summary> Returns the index of the final leg of the stage. </summary>
		/// <param name="l"> A Leg. </param>
		/// <returns> An int value. </returns>
		public int GetIndexOfFinal(Leg l)
		{
			return finalLegs.IndexOf(l);
		}
		
		/// <summary>
		/// Returns true if the stage contains the given leg. 
		/// </summary>
		/// <param name="l">A <see cref="Leg"/></param>
		/// <returns>A <see cref="System.Boolean"/></returns>		
		public bool Contains(Leg l)
		{
			foreach(Leg cl in _legsCol)
				if (cl == l) return true;
			return false;
		}		
		
		/// <summary>
		/// Returns the list of legs in the stage which depend on the given
		/// condition. 
		/// </summary>
		/// <param name="condId">A <see cref="System.String"/> with the id of the condition.</param>
		/// <returns>A <see cref="ArrayList"/> with the found legs.</returns>		
		public ArrayList FindCondition(string condId)
		{
			ArrayList list = new ArrayList();
			
			foreach(Leg l in _legsCol)
				if (l.DependsOnCondition(condId))
					list.Add(l);

			return list;
		}
		
        /// <summary>
        /// TBD
        /// </summary>
        /// <returns></returns>
		public override string ToString()
		{
			string result = "";
			
			result = result + "stage id: " + id + "\n";
			result = result + "name: " + name + "\n";
			result = result + "description: " + description + "\n";

			int i=0;
			while (i<legsCol.Count-1)
			{
				result = result + legsCol[i].ToString() + "\n";
				i++;
			}
			if (legsCol.Count > 0)
				result = result + legsCol[i].ToString();
			
			return result;
		}

        /// <summary>
        /// Returns a deep copy of a stage object.
        /// </summary>
        /// <returns></returns>
        public Stage DeepCopy(FlightPlan fp)
        {
            Stage s = new Stage(fp);
            s.description = String.Copy(this.description);
            s.id = String.Copy(this.id);
            s.manualOnly = this.manualOnly;
            s.name = String.Copy(this.name);
            s.type = this.type;
            foreach (Leg l in this._legsCol)
                s.Add(l.DeepCopy(s));
            foreach (Leg l in this.initialLegs)
                s.AddInitial(l.id);
            foreach (Leg l in this.finalLegs)
                s.AddFinal(l.id);

            s.UpdateReferences();

            return s;
        }

        #endregion
        #region - Xml -

        /// <summary>
        /// Gets the inner xml text for the specified list of Legs.
        /// I.e., a space separated list of leg ids.
        /// </summary>
        /// <param name="legs">The list with the legs.</param>
        /// <returns>The inner xml text for the Legs List.</returns>
        private String GetInnerXmlLegsIds(List<Leg> legs)
        {
            if (legs == null || legs.Count == 0)
            {
                return "";
            }

            String s = "";
            int n = legs.Count;

            for (int i = 0; i < n; i++)
            {
                s += legs[i].ID;
                if (i < n - 1)
                    s += " ";
            }

            return s;
        }

        /// <summary>
        /// Gets the inner xml text for the Initial Legs Ids node.
        /// I.e., a space separated list of leg ids.
        /// </summary>
        /// <returns>The inner xml text for the Initial Legs Ids node.</returns>
        private String GetInnerXmlInitialLegsIds()
        {
            return GetInnerXmlLegsIds(this.initialLegs);
        }

        /// <summary>
        /// Gets the inner xml text for the Final Legs Ids node.
        /// I.e., a space separated list of leg ids.
        /// </summary>
        /// <returns>The inner xml text for the Final Legs Ids node.</returns>
        private String GetInnerXmlFinalLegsIds()
        {
            return GetInnerXmlLegsIds(this.finalLegs);
        }

        /// <summary>
        /// Appends this Object to a xml FixesCollection representation.
        /// </summary>
        /// <param name="stagesElement">The StagesCollection XmlElement representation where to append this Object.</param>
        public virtual void AppendXml(XmlElement stagesElement)
        {
            // Test for errors
            if (stagesElement == null) throw new ArgumentNullException("stagesElement");

            // Creates the element            
            XmlElement xmlStage = stagesElement.OwnerDocument.CreateElement("stage");            

            // id + manualOnly + type + Ignore /////////////////////////////////////////
            xmlStage.SetAttribute("id", this.ID);
            if(ManualOnly) 
                xmlStage.SetAttribute("manualOnly", "true");
            xmlStage.SetAttribute("type", TypeToString(this.type));
            if (Ignore) 
                xmlStage.SetAttribute("ignore", "true");
            ////////////////////////////////////////////////////////////////////////////

            // Name & Description //////////////////////////////////////////////////////
            String auxName = name != null && name != "" ? this.name : "";
            String auxDesc = description != null && description != "" ? this.description : "";
            XmlHelper.AddElement(xmlStage, "name", auxName);
            XmlHelper.AddElement(xmlStage, "description", auxDesc);
            ////////////////////////////////////////////////////////////////////////////


            // Legs ////////////////////////////////////////////////////////////////////
            if (this.legsCol != null && this.legsCol.Count > 0)
                this.legsCol.AppendXml(xmlStage);
            ////////////////////////////////////////////////////////////////////////////

            // Initial & Final Legs (Space separated list of leg ids)///////////////////
            if(this.initialLegs != null && this.initialLegs.Count > 0)
                XmlHelper.AddElement(xmlStage, "initialLegs", GetInnerXmlInitialLegsIds());

            if(this.finalLegs != null && this.finalLegs.Count > 0)
                XmlHelper.AddElement(xmlStage, "finalLegs",   GetInnerXmlFinalLegsIds());
            ////////////////////////////////////////////////////////////////////////////

            // Append the elements
            stagesElement.AppendChild(xmlStage);
        }

        #endregion
    }
}
