// Updater.cs created with MonoDevelop
// User: esantama at 3:31 PM 4/18/2008

using System;
using System.IO;
using System.Xml;
using System.Xml.XPath;
using System.Reflection;
using System.Collections;
using System.Collections.Generic;

using log4net;

using Utils;

namespace FlightPlan
{
	/// <summary>
	/// Processes an XML string with changes to the flight plan. 
    /// Updates the flight plan correspondingly. 
	/// </summary>
	public class Updater
	{
		private static readonly ILog log = LogManager.GetLogger(MethodBase.GetCurrentMethod().DeclaringType);

        #region - Attributes -

        // Reader of the XML update
		private XmlReader reader;
		
		// Flight plan that is going to be updated
		private FlightPlan flight_plan;
		
		private string xs_instance_ns   = "http://www.w3.org/2001/XMLSchema-instance";
		private string xs_flightplan_ns = "http://icarus.upc.es/schema/FlightPlan/1.1";		
		private string xs_update_ns     = "http://icarus.upc.es/schema/FlightPlanUpdate/1.1";
		
		XMLCommonParser common_parser;

        #endregion
        #region - Constructor -

        /// <summary>
		/// Initializes updater. 
		/// </summary>
		/// <param name="s">A <see cref="String"/> containing the XML description of the updates.</param>
		/// <param name="fp">The <see cref="FlightPlan"/> that is going to be updated.</param>		
		public Updater(String s, FlightPlan fp)
		{
			// Create XmlReader instance
			XmlReaderSettings settings = new XmlReaderSettings();
			settings.ConformanceLevel  = ConformanceLevel.Document;
			settings.IgnoreWhitespace  = true;
			settings.IgnoreComments    = true;
			reader = XmlReader.Create(new StringReader(s), settings);
			
			flight_plan = fp;
		}

        #endregion
        #region - Functions -

        /// <summary>
		/// Process additions to the flight plan included in the update. 
		/// </summary>
		/// <param name="xnav">A <see cref="XPathNavigator"/></param>
		/// <returns></returns>
		List<Leg> ProcessAdditions(XPathNavigator xnav)
		{
			List<Leg> updated_legs = new List<Leg>();
			
			// Additions may relate to fixes, emergency plans or the main plan
			bool at_element = xnav.MoveToFollowing(XPathNodeType.Element);
			while (at_element)
			{
				if (xnav.LocalName == "Fixes")
					common_parser.ReadFixes(xnav.Clone());
				// TODO: uncommment and complete
//				else if (xnav.LocalName == "EmergencyPlans")
//					updated_legs.AddRange(ProcessChanges());
//				else if (xnav.LocalName == "MainFP")
//					updated_legs.AddRange(ProcessDeletions());
				else
					throw new XMLFPException("Unexpected element " + 
					                         xnav.LocalName + ".");
				at_element = xnav.MoveToNext();
			}
			
			return updated_legs;			
		}

		/// <summary>
        /// Reads changes to waypoints and updates them accordingly.
		/// </summary>
		/// <param name="xnav"></param>
		/// <param name="l"></param>
        void ChangeWaypoint(XPathNavigator xnav, Leg l)
		{
			common_parser.CheckLocalName(xnav, "dest");			
						
			bool at_waypoint = xnav.MoveToFirstChild();
			while (at_waypoint)
			{
				Fix       fix     = l.destination.fix;
				GPSCoords lat     = l.destination.latitude;
				GPSCoords lon     = l.destination.longitude;
				double?   alt     = l.destination.altitude;
				double?   speed   = l.destination.speed;
				bool      flyOver = l.destination.flyOver;
				
				switch (xnav.LocalName)
				{					
				    case "fix":			fix = flight_plan.Fixes[xnav.ToString()];				  break;
				    case "coordinates":	common_parser.ReadCoords(xnav.Clone(), out lat, out lon); break;
				    case "fly-over":	flyOver = xnav.ValueAsBoolean;							  break;
				    case "altitude":
                        string auxaltitude = xnav.ToString();
                        if (auxaltitude != string.Empty)
                        {
                            double tmpAlt = Convert.ToDouble(auxaltitude,
                                     flight_plan.LocaleInfo.fprovider);
                            alt = Units.LengthConvert(tmpAlt,
                                                      flight_plan.LocaleInfo.altitudeUnits,
                                                      Units.Length.m);	
                        }
                        else 
                        {
                            alt = null;
                        }
					
					    break;

				    case "speed":
                        string auxspeed = xnav.ToString();
                        if (auxspeed != string.Empty)
                        {
                            double tmpSpeed = Convert.ToDouble(auxspeed, flight_plan.LocaleInfo.fprovider);
                            speed = Units.SpeedConvert(tmpSpeed, flight_plan.LocaleInfo.speedUnits, Units.Speed.ms);
                        }
                        else
                        {
                            speed = null;
                        }
					
					    break;

				    default:
					    throw new XMLFPException("Unexpected element " + xnav.LocalName + ".");
				}
				
				if (fix == null) l.destination = new Waypoint(speed, alt, lat, lon, flyOver, l);
				else			 l.destination = new Waypoint(speed, alt, fix, flyOver, l);
				
				at_waypoint = xnav.MoveToNext();
			}
		}


        /// <summary>
        /// Changes the RF leg according to the information contained in the
        /// XML tree.
        /// </summary>
        /// <param name="xnav"></param>
        /// <param name="l"></param>
        void ChangeRFLeg(XPathNavigator xnav, Leg l)
        {
            if (!(l is RFLeg))
                throw new XMLFPException("Leg types do not match.");

            RFLeg rf = l as RFLeg;

            GPSCoords lat = rf.clatitude;
            GPSCoords lon = rf.clongitude;
            // Check elements specific to this leg
            bool at_element = xnav.MoveToFirstChild();
            while (at_element)
            {
                switch (xnav.LocalName)
                {
                    case "direction":
                        rf.direction = Leg.CharToTurnDir(xnav.ToString()[0]);
                        break;

                    case "center":
                        common_parser.ReadCoords(xnav.Clone(),out lat ,out lon);
                        rf.clongitude = lon;
                        rf.clatitude = lat;
                        break;

                    case "dest":                        
                        break; 

                    default:
                        if (!XMLCommonParser.IsCommonLegField(xnav.LocalName))
                            throw new XMLFPException("Unexpected element " +
                                                     xnav.LocalName + ".");
                        break;
                }
                at_element = xnav.MoveToNext();
            }
        }

        /// <summary>
        /// Changes the RF leg according to the information contained in the
        /// XML tree.
        /// </summary>
        /// <param name="xnav"></param>
        /// <param name="l"></param>
        void ChangeIterativeLeg(XPathNavigator xnav, Leg l)
        {
            if (!(l is IterativeLeg))
                throw new XMLFPException("Leg types do not match.");

            IterativeLeg iterative = l as IterativeLeg;
            
            // Check elements specific to this leg
            bool at_element = xnav.MoveToFirstChild();
            while (at_element)
            {
                switch (xnav.LocalName)
                {
                    case "first":
                        if (xnav.ToString() == String.Empty)
                            iterative.next = null;
                        else
                        {
                            Leg first_leg = flight_plan.FindLeg(xnav.ToString());
                            if (first_leg == null)
                            {
                                string msg = "Next leg (" +
                                    xnav.ToString() + ") not found.";
                                throw new XMLFPException(msg);
                            }
                            iterative.first = first_leg;
                        }
                        break;

                    case "last":
                        if (xnav.ToString() == String.Empty)
                            iterative.last = null;
                        else
                        {
                            Leg last_leg = flight_plan.FindLeg(xnav.ToString());
                            if (last_leg == null)
                            {
                                string msg = "Next leg (" +
                                    xnav.ToString() + ") not found.";
                                throw new XMLFPException(msg);
                            }
                            iterative.last = last_leg;
                        }
                        break;

                    case "upperBound":
                        iterative.upperBound = Convert.ToInt32(xnav.ToString());
                        break;

                    default:
                        if (!XMLCommonParser.IsCommonLegField(xnav.LocalName))
                            throw new XMLFPException("Unexpected element " +
                                                     xnav.LocalName + ".");
                        break;
                }

                at_element = xnav.MoveToNext();
            }
        }
        

		/// <summary>
        /// Changes the HF leg according to the information contained in the
        /// XML tree.
        /// </summary>
		/// <param name="xnav"></param>
		/// <param name="l"></param>
        void ChangeHFLeg(XPathNavigator xnav, Leg l)
		{
			if (!(l is HFLeg))
				throw new XMLFPException("Leg types do not match.");
			
			HFLeg hf = l as HFLeg;
			
			// Check elements specific to this leg
			bool at_element = xnav.MoveToFirstChild();
			while (at_element)
			{
				switch (xnav.LocalName)
				{
				    case "course":    hf.course = Convert.ToDouble(xnav.ToString(), flight_plan.LocaleInfo.fprovider); break;
				    case "direction": hf.direction = Leg.CharToTurnDir(xnav.ToString()[0]);					           break;

				    case "d1":
					    double tmpL = Convert.ToDouble(xnav.ToString(), flight_plan.LocaleInfo.fprovider);
					    hf.d1 = Units.LengthConvert(tmpL, flight_plan.LocaleInfo.distanceUnits, Units.Length.m);
					    break;

				    case "d2":
					    double tmpR = Convert.ToDouble(xnav.ToString(), flight_plan.LocaleInfo.fprovider);
					    hf.d2 = Units.LengthConvert(tmpR, flight_plan.LocaleInfo.distanceUnits, Units.Length.m);
					    break;

				    case "upperBound":
					    hf.upperBound = Convert.ToInt32(xnav.ToString(), flight_plan.LocaleInfo.fprovider);
					    break;

				    case "cond":
					    hf.cond = new Condition(xnav.ToString());
					    // Array of legs that contain this condition
					    ArrayList legs = flight_plan.FindCondition(xnav.ToString());
					    if (legs.Count > 0)
					    {
						    Condition c;
						    Leg cond_holder = legs[0] as Leg;

						         if (cond_holder is HFLeg)			 c = (legs[0] as HFLeg).cond;
						    else if (cond_holder is IntersectionLeg) c = (legs[0] as IntersectionLeg).nextCond;
						    else if (cond_holder is IterativeLeg)	 c = (legs[0] as IterativeLeg).cond;
						    else
							    throw new XMLFPException("Unable to get condition");
						
						    hf.cond.SetResult(c.GetResult());
					    }
					    break;
				    // TODO: implement following elements
				    /*
				    case "altitude":
					    break;
				    case "climbRate":
					    break;
				    */
				    default:
					    if (!XMLCommonParser.IsCommonLegField(xnav.LocalName))
						    throw new XMLFPException("Unexpected element " + xnav.LocalName + ".");
					    break;
				}
				
				at_element = xnav.MoveToNext();
			}
		}		


		/// <summary>
        /// Changes the ScanPoint leg according to the information contained
        /// in the XML tree.
        /// </summary>
		/// <param name="xnav"></param>
		/// <param name="l"></param>
        void ChangeScanPointLeg(XPathNavigator xnav, Leg l)
		{
			if (!(l is ScanPoint))
				throw new XMLFPException("Leg types do not match.");
			
			ScanPoint sp = l as ScanPoint;
			
			// Check elements specific to this leg
			bool at_element = xnav.MoveToFirstChild();
			while (at_element)
			{
				switch (xnav.LocalName)
				{
				    case "course":
					    sp.course = Convert.ToDouble(xnav.ToString(), flight_plan.LocaleInfo.fprovider);	
					    break;

				    case "d1":
					    double tmpL = Convert.ToDouble(xnav.ToString(), flight_plan.LocaleInfo.fprovider);
					    sp.d1 = Units.LengthConvert(tmpL, flight_plan.LocaleInfo.distanceUnits, Units.Length.m);
					    break;

				    case "d2":
					    double tmpR = Convert.ToDouble(xnav.ToString(), flight_plan.LocaleInfo.fprovider);
					    sp.d2 = Units.LengthConvert(tmpR, flight_plan.LocaleInfo.distanceUnits, Units.Length.m);
					    break;

				    default:
					    if (!XMLCommonParser.IsCommonLegField(xnav.LocalName))
						    throw new XMLFPException("Unexpected element " + xnav.LocalName + ".");
					    break;
				}
				
				at_element = xnav.MoveToNext();
			}
		}

		/// <summary>
        /// Changes the BasicScan leg according to the information contained
        /// in the XML tree.
        /// </summary>
		/// <param name="xnav"></param>
		/// <param name="l"></param>
        void ChangeBasicScanLeg(XPathNavigator xnav, Leg l)
		{
			if (!(l is BasicScan))
				throw new XMLFPException("Leg types do not match.");
			
			BasicScan bs = l as BasicScan;
			
			// Check elements specific to this leg
			bool at_element = xnav.MoveToFirstChild();
			while (at_element)
			{
				double tmp;
				
				switch (xnav.LocalName)
				{
				    case "startAt":					
					    if (xnav.ToString() == null || xnav.ToString() == String.Empty)
						    bs.startAt = null;
					    else
					    {
						    bs.startAt = new GPSCoords[2];
						    common_parser.ReadCoords(xnav, out bs.startAt[0], out bs.startAt[1]);
					    }
					    break;

				    case "dim1":
					    tmp = Convert.ToDouble(xnav.ToString(), flight_plan.LocaleInfo.fprovider);
					    bs.dim1 = Units.LengthConvert(tmp, flight_plan.LocaleInfo.distanceUnits, Units.Length.m);					
					    break;

				    case "dim2":
					    tmp = Convert.ToDouble(xnav.ToString(), flight_plan.LocaleInfo.fprovider);
					    bs.dim2 = Units.LengthConvert(tmp, flight_plan.LocaleInfo.distanceUnits, Units.Length.m);
					    break;

				    case "angle":
					    // angles assumed to be azimuth in degrees				
					    tmp = Convert.ToDouble(xnav.ToString(), flight_plan.LocaleInfo.fprovider);
					    bs.angle = tmp;	
					    break;

				    case "separation":
					    tmp = Convert.ToDouble(xnav.ToString(), flight_plan.LocaleInfo.fprovider);
					    bs.separation = Units.LengthConvert(tmp, flight_plan.LocaleInfo.distanceUnits, Units.Length.m);
					    break;

				    case "turndirection":
					    bs.turndirection = Leg.CharToTurnDir(xnav.ToString()[0]);
					    break;

				    case "d1":
					    double tmpL = Convert.ToDouble(xnav.ToString(), flight_plan.LocaleInfo.fprovider);
					    bs.d1 = Units.LengthConvert(tmpL, flight_plan.LocaleInfo.distanceUnits, Units.Length.m);
					    break;

				    case "d2":
					    double tmpR = Convert.ToDouble(xnav.ToString(), flight_plan.LocaleInfo.fprovider);
					    bs.d2 = Units.LengthConvert(tmpR, flight_plan.LocaleInfo.distanceUnits, Units.Length.m);
                        break;

				    default:
					    if (!XMLCommonParser.IsCommonLegField(xnav.LocalName))
						    throw new XMLFPException("Unexpected element " + xnav.LocalName + ".");
					    break;
				}
				
				at_element = xnav.MoveToNext();
			}
		}
		
		/// <summary>
		/// Reads changes to legs and updates them accordingly.
		/// </summary>
		/// <param name="xnav">A <see cref="XPathNavigator"/></param>
		/// <returns></returns>		
		List<Leg> ChangeLegs(XPathNavigator xnav)
		{
			List<Leg> changed_legs = new List<Leg>();
			
			common_parser.CheckLocalName(xnav, "legs");			
						
			bool at_leg = xnav.MoveToFirstChild();
			while (at_leg)
			{
				common_parser.CheckLocalName(xnav, "leg");
				
				string attr_value = xnav.GetAttribute("targetId", String.Empty);
				if (attr_value == String.Empty)
					throw new XMLFPException("targetId not found.");

				Leg l = flight_plan.FindLeg(attr_value);
				if (l == null)
					throw new XMLFPException("Leg " + attr_value + " not found");
				
				string leg_type = xnav.GetAttribute("type", xs_instance_ns);				
				if (leg_type == String.Empty)
					// try finding unqualified attribute
					leg_type = xnav.GetAttribute("type", String.Empty);
				if (leg_type == String.Empty)				
					throw new XMLFPException("Leg type not specified.");
				
				int colon_index = leg_type.IndexOf(':');
				if (colon_index >= 0)
					// Remove prefix
					leg_type = leg_type.Substring(colon_index+1); 
				
				changed_legs.Add(l);
				
				XPathNavigator legnav = xnav.Clone();
				
				// Process common elements
				bool at_element = legnav.MoveToFirstChild();
				while (at_element)
				{
					switch (legnav.LocalName)
					{					
					    case "dest": ChangeWaypoint(legnav.Clone(), l); break;
					    case "next":
						    if (legnav.ToString() == String.Empty)
							    l.next = null;
						    else
						    {							
							    Leg next_leg = flight_plan.FindLeg(legnav.ToString());
							    if (next_leg == null)
							    {
								    string msg = "Next leg (" +
									    legnav.ToString() + ") not found.";
								    throw new XMLFPException(msg);
							    }
							    l.next = next_leg;
						    }
						break;

					    case "prev":
						    if (legnav.ToString() == String.Empty)
							    l.prev = null;
						    else
						    {							
							    Leg prev_leg = flight_plan.FindLeg(legnav.ToString());
							    if (prev_leg == null)
							    {
								    string msg = "Previous leg (" +
									    legnav.ToString() + ") not found.";
								    throw new XMLFPException(msg);
							    }
							    l.prev = prev_leg;
						    }
						break;

					    default:
						    break;
					}
					
					at_element = legnav.MoveToNext();
				}
				
				// Process specific elements
				switch (leg_type)
				{
                    case "IFLeg":           break;// no specific elements
                    case "DFLeg":           break;// no specific elements
                    case "TFLeg":           break;// no specific elements
                    case "IntersectionLeg": break;
                    case "RFLeg":           ChangeRFLeg(xnav.Clone(), l);        break;
                    case "IterativeLeg":    ChangeIterativeLeg(xnav.Clone(), l); break;
				    case "HFLeg":			ChangeHFLeg(xnav.Clone(), l);		 break;
				    case "ScanPointLeg":	ChangeScanPointLeg(xnav.Clone(), l); break;
				    case "BasicScanLeg":	ChangeBasicScanLeg(xnav.Clone(), l); break;
				    default:
					    throw new XMLFPException("Unkown leg type " + leg_type + "."); 
				}

				at_leg = xnav.MoveToNext();
			}
			
			return changed_legs;
		}
		
		
		/// <summary>
		/// Reads changes to stages and updates them accordingly.
		/// </summary>
		/// <param name="xnav">A <see cref="XPathNavigator"/></param>
		/// <returns></returns>		
        List<Leg> ChangeStages(XPathNavigator xnav)
		{
			List<Leg> changed_legs = new List<Leg>();
			
			common_parser.CheckLocalName(xnav, "stages");			
						
			bool at_stage = xnav.MoveToFirstChild();
			while (at_stage)
			{
				common_parser.CheckLocalName(xnav, "stage");
				
				string attr_value = xnav.GetAttribute("targetId", String.Empty);
				if (attr_value == String.Empty)
					throw new XMLFPException("targetId not found.");
				
				Stage s = flight_plan.FindStage(attr_value);
				if (s == null)
					throw new XMLFPException("Stage " + attr_value + " not found");
								
				XPathNavigator stagenav = xnav.Clone();
				bool at_element = stagenav.MoveToFirstChild();
				while (at_element)
				{
					switch (stagenav.LocalName)
					{
					    case "name":		s.name = stagenav.ToString();						 break;
					    case "description":	s.description = stagenav.ToString();				 break;
					    case "legs":		changed_legs.AddRange(ChangeLegs(stagenav.Clone())); break;
					    case "initialLegs":						
                            string[] init_legs = stagenav.ToString().Split(' ');
						    s.ClearInitials();
						    foreach (String str in init_legs)
							    s.AddInitial(str);
						    break;
					    case "finalLegs":
						    string[] final_legs = stagenav.ToString().Split(' ');
						    s.ClearFinals();
						    foreach (String str in final_legs)
							    s.AddFinal(str);
						    break;
					    default:
						    throw new XMLFPException("Unexpected element " + stagenav.LocalName + ".");
					}
					
					at_element = stagenav.MoveToNext();
				}			
				at_stage = xnav.MoveToNext();
			}
			
			return changed_legs;
		}
			
		/// <summary>
        /// Reads flight plan changes and updates flight plan accordingly.
		/// </summary>
		/// <param name="xnav"></param>
		/// <returns></returns>
        List<Leg> ChangeFlightPlan(XPathNavigator xnav)
		{
			// Check attributes
			string attr_value = xnav.GetAttribute("targetId", String.Empty);
			if (attr_value == String.Empty)
				throw new XMLFPException("targetId not found.");
			else if (attr_value != flight_plan.id)
				throw new XMLFPException("targetId does not match flight plan id.");

			attr_value = xnav.GetAttribute("defaultTime", xs_flightplan_ns);
			if (attr_value != String.Empty)
				/* TODO: update default time */ ;
			
			attr_value = xnav.GetAttribute("maxTime", xs_flightplan_ns);
			if (attr_value != String.Empty)
				/* TODO: update max time */ ;

			List<Leg> changed_legs = null;
			
			// Now parse elements			
			bool at_element = xnav.MoveToFollowing(XPathNodeType.Element);
			while (at_element)
			{
				switch (xnav.LocalName)
				{
				    case "name":        flight_plan.name        = xnav.ToString();			  break;
				    case "description":	flight_plan.description = xnav.ToString();			  break;
				    case "stages":		changed_legs            = ChangeStages(xnav.Clone()); break;
				    default:			throw new XMLFPException("Unexpected element " + xnav.LocalName + ".");
				}
				
				at_element = xnav.MoveToNext();
			}
			
			return changed_legs;
		}			
		
		/// <summary>
		/// Process changes to the flight plan included in the update. 
		/// </summary>
		/// <param name="xnav">A <see cref="XPathNavigator"/></param>
		/// <returns></returns>		
		List<Leg> ProcessChanges(XPathNavigator xnav)
		{
			List<Leg> updated_legs = new List<Leg>();
			
			// Changes may relate to emergency plans or the main plan
			bool at_element = xnav.MoveToFollowing(XPathNodeType.Element);
			while (at_element)
			{
				if (xnav.LocalName == "MainFP")
					updated_legs = ChangeFlightPlan(xnav.Clone());
				// TODO: uncommment and complete
//				else if (xnav.LocalName == "EmergencyPlans")
//					updated_legs.AddRange(ProcessChanges());
				else
					throw new XMLFPException("Unexpected element " + xnav.LocalName + ".");
				
				
				at_element = xnav.MoveToNext();
			}
			
			return updated_legs;
        }
		
		
		/// <summary>
		/// Parse update and modify flight plan along the way.
		/// </summary>
		/// <param name="xnav">A <see cref="XPathNavigator"/></param>
		/// <returns>A <see cref="Leg"/></returns>
		private Leg[] ParseUpdate(XPathNavigator xnav)
		{
			// root node must be FlightPlanUpdate
			xnav.MoveToRoot();
			xnav.MoveToFollowing(XPathNodeType.Element);
			common_parser.CheckLocalName(xnav, "FlightPlanUpdate");

			// check schema version (namespace) 
			bool supported = false;
			IDictionary<string, string> namespaces = 
				xnav.GetNamespacesInScope(XmlNamespaceScope.Local);
			foreach (KeyValuePair<string, string> kvp in namespaces)
			{
				if (!supported && (kvp.Value == xs_update_ns))
					supported = true;
			}
			// Ignore previous check. Namespace information is lost when XML
			// is generated from the SCXML engine of the mission manager.
//			if (!supported)
//				throw new XMLFPException("Document not using " +
//				                         xs_update_ns + " namespace.");
			
			List<Leg> updated_legs = new List<Leg>();
			
			// Document may contain adds and/or changes and/or deletes
			bool at_element = xnav.MoveToFollowing(XPathNodeType.Element);
			while (at_element)
			{
				if (xnav.LocalName == "Add")
					updated_legs.AddRange(ProcessAdditions(xnav.Clone()));
				else if (xnav.LocalName == "Change")
					updated_legs.AddRange(ProcessChanges(xnav.Clone()));
				// TODO: uncomment and complete
//				else if (xnav.LocalName == "Delete")
//					ProcessDeletions();
				else
					throw new XMLFPException("Unexpected element.");
				
				at_element = xnav.MoveToNext();
			}
			
			// Remove duplicates			
			List<Leg> no_dups = new List<Leg>();
			foreach (Leg l in updated_legs)
			{
				if (!no_dups.Contains(l))
					no_dups.Add(l);
			}
			
			// Convert to array and return
			Leg [] final_updated = new Leg[no_dups.Count];
			no_dups.CopyTo(final_updated);

			return final_updated;
		}
		
		/// <summary>
		/// Update all legs that have changed. Return array with all updated legs. 
		/// </summary>
		/// <returns>A <see cref="Leg"/> list with all modified legs.</returns>		
		public Leg[] Update()
		{
			XPathDocument xdoc = new XPathDocument(reader);
			XPathNavigator xnav = xdoc.CreateNavigator();
			
			common_parser = new XMLCommonParser(xnav, flight_plan);
			
			return ParseUpdate(xnav);
		}								
		
		/// <summary>
		/// Updates the flight plan setting the given value to all occurrences
		/// of condition with id condid.
		/// </summary>
		/// <remarks>
		/// If new leg types appear that make use of conditions this method must
		/// be updated accordingly.
		/// </remarks>
		/// <param name="condid">A <see cref="System.String"/> identifying the condition whose result is to be set.</param>
		/// <param name="res">A <see cref="System.Int32"/> value with the result of the condition.</param>
		/// <param name="fp">The <see cref="FlightPlan"/> that contains the conditions.</param>
		/// <returns>A <see cref="Leg"/> list with all modified legs.</returns>		
		public static Leg[] UpdateCondition(string condid, int res, FlightPlan fp)
		{
			// Array of possibly modified legs
			ArrayList legs = fp.FindCondition(condid);
			
			// Array of unmodified legs
			Leg[] unmodified = new Leg[legs.Count];
			int i;
			for (i=0; i<unmodified.Length; i++)
				unmodified[i] = null;
			
			i = 0;
			foreach (Leg l in legs)
			{
				Condition c;
				
				// Get condition
				     if (l is IterativeLeg)	   c = (l as IterativeLeg).cond;
				else if (l is IntersectionLeg) c = (l as IntersectionLeg).nextCond;
				else c = null;
				
				if (c != null)
					log.Debug(string.Format("Updating leg {0} because it depends on condition {1}.", l.id, c.CondId));
				
				// Check that its value really changes
				bool changes = true;
				if (c != null && c.HasResult())
					changes = (c.GetResult() != res);

				// Update (or remove from modified list)
				if (changes)
					c.SetResult(res);
				else
				{
					unmodified[i] = l;
					i++;
				}
			}
			
			// Remove unmodified ones
			foreach(Leg l in unmodified)
			{
				if (l != null)
					legs.Remove(l);
				else
					break;
			}

			// Return list of modified legs
			Leg[] changed_legs = new Leg[legs.Count];
			legs.CopyTo(changed_legs);

			return changed_legs;
        }

        #endregion
    }
}
