// /home/esantama/DevProjs/mono/FPManager/BlockingQueue.cs created with MonoDevelop
// User: esantama at 1:48 PM 11/9/2007

using System;
using System.Collections;
using System.Collections.Generic;
using System.Linq;
using System.Threading;

namespace Utils
{

	//-------------------------------------------------------------------------
	// Blocking queue. Taken from:
	// http://blogs.msdn.com/toub/archive/2006/04/12/575103.aspx
	//-------------------------------------------------------------------------
	/// <summary>
	/// Blocking queue.
	/// </summary>
	/// <typeparam name="T"></typeparam>
    public class BlockingQueue<T> : IDisposable, IEnumerable<T>
	{
		// In the future we may want to add some list operations, thats why
		// internally we store information in a list.
		private List<T> _queue = new List<T>();
		private Semaphore _semaphore = new Semaphore(0, int.MaxValue);

        /// <summary>
        /// Number of elements in the queue.
        /// </summary>
        public int Count
        {
			get { return _queue.Count; }
		}

		/// <summary>
		/// Returns the first element of the queue without dequeuing it. 
		/// </summary>
		/// <returns></returns>
        public T Peek()
		{
			T head;
			lock (_queue)
			{
				if (_queue.Count == 0)
					throw new InvalidOperationException("Queue is empty");
				else
					head = _queue[0];
			}
			return head;
		}

		/// <summary>
		/// Returns the element located at position i. 
		/// </summary>
		/// <returns></returns>
		public T Get(int i)
		{
			T elem;
			lock (_queue)
			{
				if (_queue.Count < i+1)
					throw new InvalidOperationException("Index out of range.");
				else
					elem = _queue[i];
			}
			return elem;
		}

		/// <summary>
		/// Acquire lock on the queue. 
		/// </summary>
		public void AcquireLock()
		{
			Monitor.Enter(_queue);
		}

		/// <summary>
		/// Release lock on the queue. 
		/// </summary>
		public void ReleaseLock()
		{
			Monitor.Exit(_queue);
		}

		/// <summary>
		/// Add item to the queue. 
		/// </summary>
		/// <param name="data"></param>
		public void Enqueue(T data)
	    {
	        if (data == null) throw new ArgumentNullException("data");
			lock (_queue) _queue.Add(data);
			_semaphore.Release();
	    }

		/// <summary>
		/// Get item at the head of the queue. 
		/// </summary>
		/// <returns></returns>
		public T Dequeue()
	    {
			T result;
			_semaphore.WaitOne();
			lock (_queue) {
				result = _queue[0];
				_queue.RemoveAt(0);
			}
			return result;
	    }

        /// <summary>
        /// Filters a sequence of values based on a predicate.
        /// </summary>
        /// <param name="predicate"></param>
        /// <returns></returns>
        IEnumerable<T> Where(Func<T, bool> predicate)
        {
            lock (_queue)
                return _queue.Where(predicate);
        }

		/// <summary>
		/// Clears queue.
		/// </summary>
		public void Clear()
		{
			lock(_queue)
				_queue.Clear();
		}
		
		/// <summary>
		/// Insert new element before indicated existing one.
		/// </summary>
		/// <param name="new_elem">
		/// New element.
		/// </param>
		/// <param name="existing">
		/// Existing element.
		/// </param>
		public void InsertBefore(T new_elem, T existing)
		{
	        if (new_elem == null) throw new ArgumentNullException("new_elem");
			lock (_queue) 
				_queue.Insert(_queue.IndexOf(existing), new_elem);
			_semaphore.Release();
		}
		
        /// <summary>
        /// Release allocated unmanaged resources.
        /// </summary>
	    void IDisposable.Dispose()
	    {
			if (_semaphore != null)
	        {
				_semaphore.Close();
				_semaphore = null;
	        }
	    }
		
		/// <summary>
		///  Returns a new enumerator.
		/// </summary>
		/// <returns>A <see cref="IEnumerator"/></returns>
		public virtual IEnumerator GetEnumerator ()
		{
			return new Enumerator (_queue);
		}
		
        /// <summary>
        /// Returns a new IEnumerator
        /// </summary>
        /// <returns></returns>
        IEnumerator <T> IEnumerable<T>.GetEnumerator ()
		{
			return (IEnumerator <T>) GetEnumerator ();
		}

		//---------------------------------------------------------------------
		//---------------------------------------------------------------------
		private class Enumerator : IEnumerator, IEnumerator <T>, IDisposable 
        {
			
			IEnumerator queue_enum;

			internal Enumerator (List <T> q)
			{
				queue_enum = q.GetEnumerator(); 
			}

			public void Dispose ()
			{
			}
			
			public bool MoveNext () 
			{
				return queue_enum.MoveNext();
			}

			public T Current 
            {
				get 
                {
					return (queue_enum as IEnumerator <T>).Current;
				}
			}

			void IEnumerator.Reset ()
			{
				queue_enum.Reset();
			}			

			object IEnumerator.Current
			{
				get {
					return queue_enum.Current;
				}
			}			
			
		}
		
	}
	
}
