using System;
using System.Globalization;

namespace Utils
{	
	/// <summary>
    /// Manages unit conversion. 
	/// </summary>
    public static class Units
	{	
		/// <summary>
        /// Speed units 
		/// </summary>
        public enum Speed 
        {
			/// <summary>
            /// meters/second
			/// </summary>
            ms=0,
			/// <summary>
			/// knots
			/// </summary>
            kt,
			/// <summary>
            /// miles/hour
			/// </summary>
            mph
        }; 

		/// <summary>
        /// Factors for converting m/s to other units
		/// </summary>
        public static double[] speedFactor = 
        {
			1.0,		// m/s
			1.943844,	// kt
			2.236936};  // mph

		/// <summary>
		/// Length units
		/// </summary>
        public enum Length 
        {
            /// <summary>
            /// meters
            /// </summary>
			m=0,
			/// <summary>
            /// nautical miles
			/// </summary>
            nm,
			/// <summary>
            /// feet
			/// </summary>
            ft
        };

        /// <summary>
        /// Length conversion factors
        /// </summary>
		public static double[] lengthFactor = 
        {
			1.0,		// m
			0.00054,	// nm
			3.28084};	// feet

		/// <summary>
        /// Convert string to speed units. 
		/// </summary>
		/// <param name="s"></param>
		/// <returns></returns>
        public static Speed StrToSpeed(string s)
		{
			     if (s == "ms")	 return Speed.ms;
			else if (s == "kt")	 return Speed.kt;
			else if (s == "mph") return Speed.mph;
			else				 throw new System.Exception("Unknown speed units (" + s + ").");
		}

        /// <summary>
        /// Converts a Speed enum value to its string representation.
        /// </summary>
        /// <param name="s">The Speed value.</param>
        /// <returns>The speed string representation.</returns>
        public static String SpeedToStr(Speed s)
        {
            switch (s)
            {
                case Speed.ms:  return "ms";
                case Speed.kt:  return "kt";
                case Speed.mph: return "mph";
                default: throw new ArgumentException("s");
            }
        }
		
        /// <summary>
        /// Convert speed data. 
        /// </summary>
        /// <param name="val"></param>
        /// <param name="src"></param>
        /// <param name="dst"></param>
        /// <returns></returns>
        public static double SpeedConvert(double val, Speed src, Speed dst)
		{
			// from src units to m/s and from m/s to dst units 
			return val / speedFactor[(int)src] * speedFactor[(int)dst];
		}
		
		/// <summary>
        /// Convert string to length units. 
		/// </summary>
		/// <param name="s"></param>
		/// <returns></returns>
        public static Length StrToLength(string s)
		{
			     if (s == "m")	return Length.m;
			else if (s == "nm")	return Length.nm;
			else if (s == "ft")	return Length.ft;
			else				throw new System.Exception("Unknown length units (" + s + ").");
		}

        /// <summary>
        /// Converts a Length enum value to its string representation.
        /// </summary>
        /// <param name="l">The Lenght value.</param>
        /// <returns>The lenght string representation.</returns>
        public static String LengthToStr(Length l)
        {
            switch (l)
            {
                case Length.m:  return "m";
                case Length.nm: return "nm";
                case Length.ft: return "ft";
                default: throw new ArgumentException("l");
            }
        }

		/// <summary>
        /// Convert length data. 
		/// </summary>
		/// <param name="val"></param>
		/// <param name="src"></param>
		/// <param name="dst"></param>
		/// <returns></returns>
        public static double LengthConvert(double val, Length src, Length dst)
		{
			// from src units to m and from m to dst units 
			return val / lengthFactor[(int)src] * lengthFactor[(int)dst];
		}		
	}
}
