using System;
using System.IO;
using System.Xml;
using Utils;

namespace FlightPlan
{
	/// <summary>
	/// Stores a flight plan waypoint.
	/// </summary>	
	public class Waypoint
    {
        #region - Attributes & Properties -

        /// <summary>
		/// speed in m/s.
		/// </summary>
		public double? speed 	= null;		
		/// <summary>
		/// altitude in meters.
		/// </summary>
		public double? altitude = null;				
		/// <summary>
		/// fix id if waypoint is associated to a fix
		/// </summary>
		public  Fix fix = null;					
		private GPSCoords _latitude  = null;	// waypoint latitude
		private GPSCoords _longitude = null;	// waypoint longitude
		private bool _flyOver = false;          // fly-over default value is false

		/// <summary>
		/// Return true if this waypoint corresponds to a fix
		/// </summary>
		public bool isNamed 
        {
		 	get { return fix != null; }
		}
		
		/// <summary>
		/// Returns waypoint latitude.
		/// </summary>
		public GPSCoords latitude 
        {
			get { return (isNamed ? fix.Lat : _latitude); }
		}
		
		/// <summary>
		/// Returns waypoint longitude.
		/// </summary>
		public GPSCoords longitude 
        {
			get { return (isNamed ? fix.Lon : _longitude); }
		}

		/// <summary>
		/// Is this a fly over waypoint?
		/// </summary>
		public bool flyOver
        {
			get { return _flyOver; }
			set { _flyOver = value; }
		}
		
		/// <summary>
		/// Is this a fly by waypoint?
		/// </summary>
		public bool flyBy
        {
			get { return !_flyOver; }
			set { _flyOver = !value; }
		}
		
		/// <summary>
		/// Is there a target speed for this waypoint?
		/// </summary>
		public bool hasTargetSpeed 
        {
			get { return (speed.HasValue); }
		}

		/// <summary>
		/// Is there a target altitude for this waypoint?
		/// </summary>
		public bool hasTargetAlt 
        {
			get { return (altitude.HasValue); }
		}

		/// <summary>
		/// This value should be set before sending the waypoint to the VAS, so
		/// that we are able to identify the waypoint when VAS notifications are
		/// received.
		/// </summary>
		public int vas_id = 0;



        /// <summary>
        /// Returns waypoint's stage 
        /// </summary>
        public Stage stage
        {
            get { return _leg.stage; }
        }

        private Leg _leg;
        /// <summary>
        /// Returns waypoint's parent leg 
        /// </summary>
        public Leg leg
        {
            get { return _leg; }
            set { _leg = value; }
        }

        #endregion
        #region - Constructors -

        //---------------------------------------------------------------------
		/// <summary>
		/// Creates a waypoint object associated to a fix. 
		/// </summary>
		/// <param name="speed">A <see cref="System.Nullable"/></param>
		/// <param name="alt">A <see cref="System.Nullable"/></param>
		/// <param name="fix">A <see cref="Fix"/></param>
		/// <param name="fover">A <see cref="System.Boolean"/></param>
		/// <param name="parent">A <see cref="Leg"/></param>
		//---------------------------------------------------------------------
		public Waypoint(double? speed, double? alt, Fix fix, bool fover, Leg parent)
		{
			this.speed = speed;
			this.altitude = alt;
			this.fix = fix;			
			_flyOver = fover;
			
			_latitude  = null;
			_longitude = null;
			
			_leg = parent;
		}
		
		//---------------------------------------------------------------------
		/// <summary>
		/// Creates a waypoint specified via its coordinates. 
		/// </summary>
		/// <param name="speed">A <see cref="System.Nullable"/></param>
		/// <param name="alt">A <see cref="System.Nullable"/></param>
		/// <param name="lat">A <see cref="GPSCoords"/></param>
		/// <param name="lon">A <see cref="GPSCoords"/></param>
		/// <param name="fover">A <see cref="System.Boolean"/></param>
		/// <param name="parent">A <see cref="Leg"/></param>
		//---------------------------------------------------------------------
		public Waypoint(double? speed, double? alt, 
		                GPSCoords lat, GPSCoords lon, bool fover, Leg parent)
		{
			this.speed = speed;
			altitude  = alt;
			_latitude  = lat;
			_longitude = lon;
			_flyOver = fover;
					
			fix = null; // unnamed waypoint
			
			_leg = parent;			
		}

		//---------------------------------------------------------------------
		/// <summary>
		/// Copy constructor.
		/// </summary>
		/// <param name="wp">A <see cref="Waypoint"/></param>
		//---------------------------------------------------------------------
		public Waypoint(Waypoint wp)
		{
			speed = wp.speed;
			altitude = wp.altitude;
			fix = wp.fix;			
			_latitude  = wp.latitude;
			_longitude = wp.longitude;
			_flyOver = wp.flyOver;
			_leg = wp.leg;
		}

        #endregion
        #region - Public Functions -

        //---------------------------------------------------------------------
        /// <summary>
        /// Sets the parent of the waypoint.
        /// </summary>
        /// <remarks>
        /// When reading flight plan it's not possible to set the parent at
        /// creation time and it is done later.
        /// </remarks>
        /// <param name="parent">A <see cref="Leg"/></param>
        //---------------------------------------------------------------------
        public void SetParent(Leg parent)
        {
            _leg = parent;
        }

        //---------------------------------------------------------------------
		/// <summary>
		/// Returns distance beetween two waypoints. 
		/// </summary>
		/// <param name="wp">A <see cref="Waypoint"/></param>
		/// <returns>A <see cref="System.Double"/> with the distance in meters.</returns>
		//---------------------------------------------------------------------
		public double Distance(Waypoint wp)
		{			
			double course, dist;
			this.Distance(wp, out course, out dist);
			return dist;
		}
		
		//---------------------------------------------------------------------
		/// <summary>
		/// Returns angle and distance from this waypoint to wp. 
		/// </summary>
		/// <param name="wp">A <see cref="Waypoint"/></param>
		/// <param name="course">A <see cref="System.Double"/> with the azimuth angle in degrees.</param>
		/// <param name="dist">A <see cref="System.Double"/> with the distance in meters.</param>
		//---------------------------------------------------------------------
		public void Distance(Waypoint wp, out double course, out double dist)
		{			
			double lat1 = this.latitude.AsDouble();
			double lon1 = this.longitude.AsDouble();
			double lat2 = wp.latitude.AsDouble();
			double lon2 = wp.longitude.AsDouble();
			
			double reverse;
				
			Geodesy.geo_inverse_wgs_84(0, lat1, lon1, lat2, lon2, 
			                           out course, out reverse, out dist);
		}

		//---------------------------------------------------------------------
		/// <summary>
		/// Adds offset along a given course.
		/// </summary>
		/// <param name="course">A <see cref="System.Double"/> with an azimuth angle in degrees.</param>
		/// <param name="dist">A <see cref="System.Double"/> with the offset in meters.</param>
		//---------------------------------------------------------------------
		public void Move( double course, double dist)
		{
			// if waypoint is moved it can no longer be associated to a fix
			if (fix != null) {
				_latitude = fix.Lat;
				_longitude = fix.Lon;				
				fix = null;
			}
			
			double tmp;
			double alt = 0;
			double lat = _latitude.AsDouble();
			double lon = _longitude.AsDouble();

			Geodesy.geo_direct_wgs_84(alt, lat, lon, course, dist,
			                          out lat, out lon, out tmp);
			
			_latitude  = new GPSCoords(lat, true);
			_longitude = new GPSCoords(lon, false);
		}
			
		//---------------------------------------------------------------------
		/// <summary>
		/// Rotate this waypoint a given number of degrees with respect to center. 
		/// </summary>
		/// <param name="center">A <see cref="Waypoint"/></param>
		/// <param name="angle">A <see cref="System.Double"/></param>
		//---------------------------------------------------------------------
		public void Rotate(Waypoint center, double angle)
		{
			double dist, course;
			
			center.Distance(this, out course, out dist);			
			// add angle to obtained course
			course = course + angle;
			// compute rotated point
			this._latitude  = center.latitude;
			this._longitude = center.longitude;
			Move(course, dist);
		}
		
        /// <summary>
        /// Gets the string representation for this WayPoint.
        /// </summary>
        /// <returns>The string representation for this WayPoint.</returns>
		public override string ToString ()
		{
			return string.Format("({0}, {1})", _latitude.AsDouble(), _longitude.AsDouble());			                     
		}

        /// <summary>
        /// Returns a deep copy of a waypoint object.
        /// </summary>
        /// <returns></returns>
        public Waypoint DeepCopy(FixesCollection fixes)
        {
            Waypoint w = this.MemberwiseClone() as Waypoint;
            if (this.fix != null)
                w.fix = fixes[this.fix.Id];
            w._latitude = this.latitude.DeepCopy();
            w._longitude = this.longitude.DeepCopy();

            return w;
        }

        /// <summary>
        /// Creates a String representation for this fix coordinates (latitude longitude). 
        /// </summary>
        /// <returns>A String representation for this fix coordinates (latitude longitude).</returns>
        public String CoordinatesToStr()
        {
            if (latitude  == null) throw new Exception("The latitude can't be null");
            if (longitude == null) throw new Exception("The longitude can't be null");

            return String.Format("{0} {1}", latitude.AsDouble(), longitude.AsDouble());
        }

        #endregion
        #region - Xml -

        /// <summary>
        /// Appends this Object to a xml element representation.
        /// </summary>
        /// <param name="parentElement">The WayPoint XmlElement representation where to append this Object.</param>
        public void AppendXml(XmlElement parentElement)
        {
            // Test for errors
            if (parentElement == null) throw new ArgumentNullException("parentElement");

            // Coordinates //////////////////////////////////////////////////////////////////
            if (isNamed) // fix != null
                XmlHelper.AddElement(parentElement, "fix", this.fix.Id);
            else // fix == null
                XmlHelper.AddElement(parentElement, "coordinates", this.CoordinatesToStr());
            /////////////////////////////////////////////////////////////////////////////////


            // Altitude /////////////////////////////////////////////////////////////////////
            if (altitude !=null && altitude.HasValue)
                XmlHelper.AddElement(parentElement, "altitude", this.altitude.ToString());
            /////////////////////////////////////////////////////////////////////////////////

            // Speed ////////////////////////////////////////////////////////////////////////
            if(speed != null && speed.HasValue)
                XmlHelper.AddElement(parentElement, "speed", this.speed.ToString());
            /////////////////////////////////////////////////////////////////////////////////

            // Speed ////////////////////////////////////////////////////////////////////////
            if(this.flyOver)
                XmlHelper.AddElement(parentElement, "fly-over", "true");
            /////////////////////////////////////////////////////////////////////////////////
        }

        #endregion
    }
}
