// XMLCommonParser.cs created with MonoDevelop
// User: esantama at 11:40 AM 11/24/2009

using System;
using System.Xml;
using System.Xml.XPath;
using System.Collections;
using System.Collections.Generic;

using Utils;

namespace FlightPlan
{	
	class XMLCommonParser
    {
        #region - Static -

        /// <summary>
        /// Returns true if the element name represents a field which is common
        /// to all leg types. I.e. it is defined in the base leg type.
        /// </summary>
        /// <param name="elem_name">A <see cref="System.String"/> with the name of the XML element.</param>
        /// <returns>A <see cref="System.Boolean"/></returns>
        public static bool IsCommonLegField(string elem_name)
        {
            String[] common_elems = { "dest", "next", "prev" };
            return Array.IndexOf(common_elems, elem_name) >= 0;
        }

        #endregion

        #region - Attributes -

        private IDictionary<string, string> namespaces;
		
		// Known flight plan related namespaces
		private string xs_instance_ns   = "http://www.w3.org/2001/XMLSchema-instance";
		private string xs_flightplan_ns = "http://icarus.upc.es/schema/FlightPlan/1.1";		
		private string xs_update_ns     = "http://icarus.upc.es/schema/FlightPlanUpdate/1.1";

        public FlightPlan fp;

        /// <summary>
        /// Emergency files as a list of references.
        /// </summary>
        public ArrayList emergencyFileRefs = null;
        public ArrayList EmergencyFileRefs
        {
            get { return emergencyFileRefs; }
            set { emergencyFileRefs = value; }
        }

        #endregion
        #region - Constructor -

        public XMLCommonParser(XPathNavigator xnav, FlightPlan fp)
		{
			xnav.MoveToRoot();
			xnav.MoveToFollowing(XPathNodeType.Element);
			namespaces = xnav.GetNamespacesInScope(XmlNamespaceScope.Local);
			this.fp = fp;
		}

        #endregion
        #region - Public -

        public void SetXMLCommonParser(XPathNavigator xnav, FlightPlan fp)
        {
            xnav.MoveToRoot();
            xnav.MoveToFollowing(XPathNodeType.Element);
            namespaces = xnav.GetNamespacesInScope(XmlNamespaceScope.Local);
            this.fp = fp;
        }
	    
		/// <summary>
		/// Checks that node name corresponds to what is expected. 
		/// </summary>
		/// <param name="xnav">A <see cref="XPathNavigator"/> object.</param>
		/// <param name="name">A <see cref="System.String"/> with the expected name.</param>		
		public void CheckLocalName(XPathNavigator xnav, string name)
		{
			if (xnav.LocalName != name)
			{
				string msg = name + " node expected. Found " + xnav.Name + " instead."; 
				throw new XMLFPException(msg);
			}
		}

        /// <summary>
        /// Reads a coordinates element. Creates new GPSCoords elements and returns them.
        /// Latitud always precedes longitude and the are separated by a space.
        /// Coordinates can be written in two forms:
        ///   - ddºmm'ss"D ddºmm'ss"D
        ///   - or xx.yyyyyyy xx.yyyyyyy
        /// </summary>
        /// <param name="cnav"></param>
        /// <param name="lat"></param>
        /// <param name="lon"></param>
		public void ReadCoords(XPathNavigator cnav, out GPSCoords lat, out GPSCoords lon)
		{
			string coords = cnav.ToString();
			string[] values = coords.Trim().Split();
			
			if (values.Length > 2)
                throw new XMLFPException("Only one space is allowed in coordinates element.");
			
			// which form?
			char[] dirchars = {'N', 'S', 'E', 'W'};
			if (coords.IndexOfAny(dirchars) > -1)
			{
			    char[] separators = {'º', '\'', '"'};
				string[] parts = values[0].Split(separators);
				
				if (parts.Length != 4) 
                {
					string msg = string.Format("Malformed coordinates: {0}.", values[0]);
					throw new XMLFPException(msg);
				}
				
				int degs = Convert.ToInt32(parts[0], fp.LocaleInfo.fprovider);
				int mins = Convert.ToInt32(parts[1], fp.LocaleInfo.fprovider);
				double secs = Convert.ToDouble(parts[2], fp.LocaleInfo.fprovider);
				GPSCoords.Direction dir = GPSCoords.CharToDirection(parts[3][0], true);
				
				lat = new GPSCoords(degs, mins, secs, dir);
				
				parts = values[1].Split(separators);
				if (parts.Length != 4) 
                {
					string msg = string.Format("Malformed coordinates: {0}.", values[1]);
					throw new XMLFPException(msg);
				}
				
				degs = Convert.ToInt32(parts[0], fp.LocaleInfo.fprovider);
				mins = Convert.ToInt32(parts[1], fp.LocaleInfo.fprovider);
				secs = Convert.ToDouble(parts[2], fp.LocaleInfo.fprovider);
				dir  = GPSCoords.CharToDirection(parts[3][0], false);
				
				lon = new GPSCoords(degs, mins, secs, dir);
			}
			else
			{
				lat = new GPSCoords(Convert.ToDouble(values[0], fp.LocaleInfo.fprovider), true);
				lon = new GPSCoords(Convert.ToDouble(values[1], fp.LocaleInfo.fprovider), false);
			}
			
		}
		
        /// <summary>
        /// Read fixes list 
        /// </summary>
        /// <param name="fnav"></param>
		public void ReadFixes(XPathNavigator fnav)
		{			
			bool exists = fnav.MoveToFirstChild();

			while (exists)
			{
				if (fnav.NodeType == XPathNodeType.Element)
				{
					// Read fix and add to the list of fixes				
					CheckLocalName(fnav, "Fix");
					
					// Read fix id
					string id = fnav.GetAttribute("id", String.Empty);
					if (id == String.Empty)
						throw new XMLFPException("Fix id required.");						
											
					// Check whether it's of special type
					string type = fnav.GetAttribute("type", xs_instance_ns);
					if (type != String.Empty && !IsValidFixType(type))
						throw new XMLFPException(type + " is not a valid fix type.");
                    // Remove prefix
                    type = type.Substring(type.IndexOf(':') + 1);
					
					// Read fix's children
					XPathNavigator childnav = fnav.Clone();
									
					childnav.MoveToFollowing(XPathNodeType.Element);
					CheckLocalName(childnav, "name");
					string name = childnav.ToString();

					childnav.MoveToFollowing(XPathNodeType.Element);
					CheckLocalName(childnav, "description");
					string desc = childnav.ToString();
								
					childnav.MoveToFollowing(XPathNodeType.Element);
					CheckLocalName(childnav, "coordinates");
					GPSCoords lat, lon;
					ReadCoords(childnav, out lat, out lon);

					Fix f = null;

					if (type != String.Empty && type == "RwyFixType") 
                    {
						double altitude, heading, distance;
						
						// Read additional elements
						
						childnav.MoveToFollowing(XPathNodeType.Element);
						CheckLocalName(childnav, "altitude");
						altitude = Convert.ToDouble(childnav.ToString(), fp.LocaleInfo.fprovider);
						altitude = Units.LengthConvert(altitude, fp.LocaleInfo.altitudeUnits, Units.Length.m);
						
						childnav.MoveToFollowing(XPathNodeType.Element);
						CheckLocalName(childnav, "heading");
						heading = Convert.ToDouble(childnav.ToString(), fp.LocaleInfo.fprovider);
						
						childnav.MoveToFollowing(XPathNodeType.Element);
						CheckLocalName(childnav, "length");
						distance = Convert.ToDouble(childnav.ToString(), fp.LocaleInfo.fprovider);
						distance = Units.LengthConvert(distance, fp.LocaleInfo.distanceUnits, Units.Length.m);

						f = new RwyFix(id, name, desc, lat, lon, 
						               altitude, heading, distance);
					}
					else
						f = new Fix(id, name, desc, lat, lon);
					
					fp.Fixes.Add(f);
				}
				
				exists = fnav.MoveToNext();
			}		
		}

        /// <summary>
        /// Read emergency reference list
        /// </summary>
        /// <param name="fnav"></param>
		public void ReadEmergencyReferences(XPathNavigator fnav)
		{			
			bool exists = fnav.MoveToFirstChild();
            

            while (exists)
			{
				if (fnav.NodeType == XPathNodeType.Element)
				{
                    //Emergency e = null;

                    // Read Emergency References
                    CheckLocalName(fnav, "emergency-fp");
                    string name = fnav.ToString();

                    this.EmergencyFileRefs.Add(name);
                }

                exists = fnav.MoveToNext();
            }
        }

        #endregion
        #region - Private -

        /// <summary>
        /// Returns true if type corresponds to a valid fix type. Returns false otherwise.
        /// </summary>
        /// <param name="type"></param>
        /// <returns></returns>
        private bool IsValidFixType(string type)
        {
            string[] ValidTypes = { "FixType", "RwyFixType" };

            // get prefix
            string prefix = type.Substring(0, type.IndexOf(':'));
            string unq_type = type.Substring(type.IndexOf(':') + 1);

            // check prefix
            if (namespaces[prefix] != xs_flightplan_ns)
                return false;

            // check unq_type
            return (Array.IndexOf(ValidTypes, unq_type) >= 0);
        }

        #endregion
    }
}
