// /home/esantama/DevProjs/mono/FPManager/XMLFPParser.cs created with MonoDevelop
// User: esantama at 6:04 PM 2/5/2008

using System;
using System.IO;
using System.Xml;
using System.Xml.XPath;
using System.Xml.Schema;
using System.Collections;
using System.Collections.Generic;
using System.Reflection;

using log4net;
using Utils;

namespace FlightPlan
{	
	/// <summary>
	/// Exception raised when problems are encountered during parsing of the flight plan.
	/// </summary>
	public class XMLFPException : Exception
    {
        #region - Constructor -

        /// <summary>
		/// Creates XMLFPException object.
		/// </summary>
		/// <param name="s">A <see cref="System.String"/> with the message associated to the exception.</param>
		public XMLFPException (string s) : base (s) { }

        #endregion
    }	


	/// <summary>
	/// Implements a parser of flight plans given in XML format.
	/// </summary>
	/// <remarks>
	/// An XMLFPParser reads the flight plan file and initializes two data
	/// structures: a collection of fixes and the flight plan.
	/// </remarks>
	public class XMLFPParser
    {
        #region - Static -

        public static string xs_instance_ns   = "http://www.w3.org/2001/XMLSchema-instance";
        public static string xs_flightplan_ns = "http://icarus.upc.es/schema/FlightPlan/1.1";

		/// <summary>
		/// Validation event handler. 
		/// </summary>
		/// <param name="sender">A <see cref="System.Object"/></param>
		/// <param name="e">A <see cref="ValidationEventArgs"/></param>	    
		static void fpValidationEventHandler(object sender, ValidationEventArgs e)
	    {
	        if (e.Severity == XmlSeverityType.Warning)
	        {
	            Console.Write("WARNING: ");
	            Console.WriteLine(e.Message);
	        }
	        else if (e.Severity == XmlSeverityType.Error)
	        {
	            Console.Write("ERROR: ");
	            Console.WriteLine(e.Message);
	        }
	    }

        #endregion

        #region - Attributes -

        private static readonly ILog log = LogManager.GetLogger(MethodBase.GetCurrentMethod().DeclaringType);
		
		private XmlReader reader;
        private XmlReaderSettings settings;
		private FlightPlan fp;
        private string flightPlan;
		private IDictionary<string, string> namespaces; 		
		
		XMLCommonParser common_parser;

        #endregion
        #region - Constructors -

		/// <summary>
		/// Creates a new instance of XMLFPParser.
		/// </summary>
		public XMLFPParser()
		{
		}

        #endregion
        #region - Public -

        /// <summary>
		/// Initialize parser passing a string containing the flight plan. 
		/// </summary>
		/// <param name="content">A <see cref="System.String"/> containing the flight plan data in XML format.</param>		
        public void SetSettings()
        {
            // Create XmlReader instance
            settings = new XmlReaderSettings();
            settings.ConformanceLevel = ConformanceLevel.Document;
            settings.IgnoreWhitespace = true;
            settings.IgnoreComments = true;
            // Set settings to enforce schema validation
            settings.ValidationType = ValidationType.Schema;
            // TODO: Enable validation. Right now validation is disabled due to the reader reporting
            //   incorrect errors.
            // settings.Schemas.Add(null, "/home/eduard/DevProjs/NetBeans/FlighPlan/src/FlightPlan.xsd");
            settings.ValidationEventHandler += new ValidationEventHandler(fpValidationEventHandler);
        }

        public void SetContent(string content)
		{			
			reader = XmlReader.Create(new StringReader(content), this.settings);
		}

        /// <summary>
        /// Reads a leg element from the document.
        /// </summary><param name="legnav">A <see cref="XPathNavigator"/> object.</param>
        /// <param name="lh">A <see cref="LegHelper"/> which is going to be updated with the read data.</param>
        public void ReadLegElement(XPathNavigator legnav, LegHelper lh)
        {
            log.Debug(string.Format("Reading element {0} of leg {1}.", legnav.Name, lh.id));

            if (legnav.Name == "dest") lh.destination = ReadWaypoint(legnav.Clone());	// Read destination waypoint
            else if (legnav.Name == "next") lh.next = legnav.ToString();		        // Read next leg id
            else if (legnav.Name == "prev") lh.prev = legnav.ToString();		        // Read prev leg id}
            else if (legnav.Name == "contingency")
            {
                lh.contingency = ReadContingency(legnav.Clone());	// Read emergency plan id

                if (lh.emergency == null)
                    lh.emergency = new ContingencyList();
                lh.emergency.Add(lh.contingency);
            }

            // RFLeg attributes ///////////////////////////////////////////////////
            else if (legnav.Name == "center") common_parser.ReadCoords(legnav, out lh.centerLat, out lh.centerLon); 	// Read RF leg arc center coordinates
            else if (legnav.Name == "direction") lh.direction = Leg.CharToTurnDir(legnav.ToString()[0]);                // Read turn direction


            // HFLeg attributes ////////////////////////////////////////////////////
            else if (legnav.Name == "course")
            {
                // Read RF leg course
                double tmpC = Convert.ToDouble(legnav.ToString(), common_parser.fp.LocaleInfo.fprovider);
                // angles assumed to be azimuth in degrees
                lh.course = tmpC;
            }
            else if (legnav.Name == "d1")
            {
                // Read holding pattern distance from center to center
                double tmpL = Convert.ToDouble(legnav.ToString(), common_parser.fp.LocaleInfo.fprovider);
                lh.d1 = Units.LengthConvert(tmpL, common_parser.fp.LocaleInfo.distanceUnits, Units.Length.m);
            }
            else if (legnav.Name == "d2")
            {
                // Read diameter
                double tmpR = Convert.ToDouble(legnav.ToString(), common_parser.fp.LocaleInfo.fprovider);
                lh.d2 = Units.LengthConvert(tmpR, common_parser.fp.LocaleInfo.distanceUnits, Units.Length.m);
            }


            // IRFLeg attributes /////////////////////////////////////////////////////
            else if (legnav.Name == "dimN")
            {
                // Read dim north
                double tmpL = Convert.ToDouble(legnav.ToString(), common_parser.fp.LocaleInfo.fprovider);
                lh.dimN = Units.LengthConvert(tmpL, common_parser.fp.LocaleInfo.distanceUnits, Units.Length.m);
            }
            else if (legnav.Name == "dimS")
            {
                // Read dim south
                double tmpR = Convert.ToDouble(legnav.ToString(), common_parser.fp.LocaleInfo.fprovider);
                lh.dimS = Units.LengthConvert(tmpR, common_parser.fp.LocaleInfo.distanceUnits, Units.Length.m);
            }
            else if (legnav.Name == "dimE")
            {
                // Read dim north
                double tmpL = Convert.ToDouble(legnav.ToString(), common_parser.fp.LocaleInfo.fprovider);
                lh.dimE = Units.LengthConvert(tmpL, common_parser.fp.LocaleInfo.distanceUnits, Units.Length.m);
            }
            else if (legnav.Name == "dimW")
            {
                // Read dim south
                double tmpR = Convert.ToDouble(legnav.ToString(), common_parser.fp.LocaleInfo.fprovider);
                lh.dimW = Units.LengthConvert(tmpR, common_parser.fp.LocaleInfo.distanceUnits, Units.Length.m);
            }

            // LALeg attributes //////////////////////////////////////////////////////////////
            else if (legnav.Name == "length")
            {
                // Read runway lenght
                double tmpL = Convert.ToDouble(legnav.ToString(), common_parser.fp.LocaleInfo.fprovider);
                lh.length = Units.LengthConvert(tmpL, common_parser.fp.LocaleInfo.distanceUnits, Units.Length.m);
            }
            else if (legnav.Name == "rwyhead") lh.rwyhead = ReadWaypoint(legnav.Clone());// Read hwp1 waypoint
            else if (legnav.Name == "hwp1") lh.hwp1 = ReadWaypoint(legnav.Clone());	// Read hwp1 waypoint
            else if (legnav.Name == "hwp2") lh.hwp2 = ReadWaypoint(legnav.Clone());	// Read hwp2 waypoint
            else if (legnav.Name == "hwp3") lh.hwp3 = ReadWaypoint(legnav.Clone());	// Read hwp3 waypoint
            else if (legnav.Name == "hwp4") lh.hwp4 = ReadWaypoint(legnav.Clone());	// Read hwp4 waypoint
            else if (legnav.Name == "iwp") lh.iwp = ReadWaypoint(legnav.Clone());	// Read iwp waypoint
            else if (legnav.Name == "idwp") lh.idwp = ReadWaypoint(legnav.Clone());	// Read idwp waypoint
            else if (legnav.Name == "ibwp") lh.ibwp = ReadWaypoint(legnav.Clone());	// Read ibwp waypoint
            else if (legnav.Name == "ifwp") lh.ifwp = ReadWaypoint(legnav.Clone());	// Read ifwp waypoint
            else if (legnav.Name == "tdwp") lh.tdwp = ReadWaypoint(legnav.Clone());	// Read tdwp waypoint
            else if (legnav.Name == "bpwp") lh.bpwp = ReadWaypoint(legnav.Clone());	// Read bpwp waypoint
            else if (legnav.Name == "lawp") lh.lawp = ReadWaypoint(legnav.Clone());	// Read lawp waypoint
            else if (legnav.Name == "eawp") lh.eawp = ReadWaypoint(legnav.Clone());	// Read eawp waypoint
            else if (legnav.Name == "nextFrst") lh.nextAbortId = legnav.ToString();
            else if (legnav.Name == "dvrt-frust") lh.dvrtFrust = legnav.ToString();
            else if (legnav.Name == "start-op") lh.startOP = legnav.ToString();

            // TOLeg attributes //////////////////////////////////////////////////////////////
            else if (legnav.Name == "der") lh.der = ReadWaypoint(legnav.Clone()); // Read der waypoint
            else if (legnav.Name == "edwp-a") lh.edwp_a = ReadWaypoint(legnav.Clone()); // Read hwp2 waypoint
            else if (legnav.Name == "edwp-a-min") lh.edwp_a_min = ReadWaypoint(legnav.Clone()); // Read hwp3 waypoint
            else if (legnav.Name == "edwp-bl") lh.edwp_bl = ReadWaypoint(legnav.Clone()); // Read hwp4 waypoint
            else if (legnav.Name == "edwp-bl-min") lh.edwp_bl_min = ReadWaypoint(legnav.Clone()); // Read iwp waypoint
            else if (legnav.Name == "edwp-br") lh.edwp_br = ReadWaypoint(legnav.Clone()); // Read idwp waypoint
            else if (legnav.Name == "edwp-br-min") lh.edwp_br_min = ReadWaypoint(legnav.Clone()); // Read ibwp waypoint
            else if (legnav.Name == "edwp-cl") lh.edwp_cl = ReadWaypoint(legnav.Clone()); // Read ifwp waypoint
            else if (legnav.Name == "edwp-cr") lh.edwp_cr = ReadWaypoint(legnav.Clone()); // Read tdwp waypoint
            else if (legnav.Name == "rda-w")
            {
                double tmp = Convert.ToDouble(legnav.ToString(), common_parser.fp.LocaleInfo.fprovider);
                lh.rda_w = Units.LengthConvert(tmp, common_parser.fp.LocaleInfo.distanceUnits, Units.Length.m);
            }
            else if (legnav.Name == "rda-l")
            {
                double tmp = Convert.ToDouble(legnav.ToString(), common_parser.fp.LocaleInfo.fprovider);
                lh.rda_l = Units.LengthConvert(tmp, common_parser.fp.LocaleInfo.distanceUnits, Units.Length.m);
            }
            else if (legnav.Name == "end-op")
            {
                lh.endOP = legnav.ToString();
            }

            // IterativeLeg attributes ///////////////////////////////////////////////////
            else if (legnav.Name == "body") lh.body = ReadLegRefSequence(legnav.Clone(), "body");
            else if (legnav.Name == "upperBound") lh.upperBound = Convert.ToInt32(legnav.ToString(), fp.LocaleInfo.fprovider);
            else if (legnav.Name == "first") lh.first = legnav.ToString();
            else if (legnav.Name == "last") lh.last = legnav.ToString();
            else if (legnav.Name == "cond") lh.nextCond = legnav.ToString();


            // IntersectionLeg attributes ///////////////////////////////////////////////
            else if (legnav.Name == "nextCond") lh.nextCond = legnav.ToString();
            else if (legnav.Name == "nextList") lh.nextList = ReadLegRefSequence(legnav.Clone(), "nextList");
            else if (legnav.Name == "prevCond") lh.prevCond = legnav.ToString();
            else if (legnav.Name == "prevList") lh.prevList = ReadLegRefSequence(legnav.Clone(), "prevList");


            // BasicScanLeg attributes ///////////////////////////////////////////////////            
            else if (legnav.Name == "dim1")
            {
                double tmp = Convert.ToDouble(legnav.ToString(), fp.LocaleInfo.fprovider);
                lh.dim1 = Units.LengthConvert(tmp, fp.LocaleInfo.distanceUnits, Units.Length.m);
            }
            else if (legnav.Name == "dim2")
            {
                double tmp = Convert.ToDouble(legnav.ToString(), fp.LocaleInfo.fprovider);
                lh.dim2 = Units.LengthConvert(tmp, fp.LocaleInfo.distanceUnits, Units.Length.m);
            }
            else if (legnav.Name == "angle")
            {
                // angles assumed to be azimuth in degrees				
                double tmp = Convert.ToDouble(legnav.ToString(), fp.LocaleInfo.fprovider);
                lh.angle = tmp;
            }
            else if (legnav.Name == "separation")
            {
                double tmp = Convert.ToDouble(legnav.ToString(), fp.LocaleInfo.fprovider);
                lh.separation = Units.LengthConvert(tmp, fp.LocaleInfo.distanceUnits, Units.Length.m);
            }
            else if (legnav.Name == "turndirection")
            {
                lh.direction = Leg.CharToTurnDir(legnav.ToString()[0]);
            }
            else if (legnav.Name == "d1")
            {
                double tmp = Convert.ToDouble(legnav.ToString(), fp.LocaleInfo.fprovider);
                lh.d1 = Units.LengthConvert(tmp, fp.LocaleInfo.distanceUnits, Units.Length.m);
            }
            else if (legnav.Name == "d2")
            {
                double tmp = Convert.ToDouble(legnav.ToString(), fp.LocaleInfo.fprovider);
                lh.d2 = Units.LengthConvert(tmp, fp.LocaleInfo.distanceUnits, Units.Length.m);
            }
            else if (legnav.Name == "startAt")
            {
                lh.startAt = new GPSCoords[2];
                common_parser.ReadCoords(legnav, out lh.startAt[0], out lh.startAt[1]);
            }

            // Anything else ///////////////////////////////////////////////////////////////
            else if (legnav.NodeType != XPathNodeType.Comment)
            {
                string msg = String.Format("Leg contains unexpected element: {0}.", legnav.Name);
                throw new XMLFPException(msg);
            }
        }

        /// <summary>
        /// Checks the correctness of the initial legs distribution.
        /// </summary>
        public void CheckInitialLegs()
        {
            foreach (Leg l in common_parser.fp.stages[0].initialLegs)
            {
                if ((l is IFLeg) || (l is IRFLeg) || (l is TOLeg))
                {
                }
                else
                {
                    throw new Exception(String.Format("Leg {0} cannot be an initial flightplan leg.", l.ID));
                }
            }
        }

        /// <summary>
        /// Parse XML file and initialize flight plan. 
        /// </summary>
        public FlightPlan Parse()
        {
            fp = new FlightPlan();

            XPathDocument xdoc = new XPathDocument(reader);
            XPathNavigator xnav = xdoc.CreateNavigator();

            common_parser = new XMLCommonParser(xnav, fp);

            ReadFlightPlan(xnav);

            foreach (String efr in common_parser.EmergencyFileRefs)
            {
                FlightPlan efp = new FlightPlan();
                String flightPlan = File.ReadAllText(efr);
                this.SetSettings();
                this.SetContent(flightPlan);
                xdoc = new XPathDocument(this.reader);
                xnav = xdoc.CreateNavigator();
                common_parser.SetXMLCommonParser(xnav, efp);
                common_parser.fp = efp;
                ReadEmergencyFlightPlan(xnav);
            }
            return fp;
        }
		
        #endregion
        #region - Private -

        /// <summary>
        /// Read locale settings and initialize fprovider.
        /// </summary>
		private void ReadLocale(XPathNavigator xnav)			
		{			
			xnav.MoveToFollowing(XPathNodeType.Element);
			common_parser.CheckLocalName(xnav, "speedUnits");
            common_parser.fp.LocaleInfo.speedUnits = Units.StrToSpeed(xnav.ToString());
			
			xnav.MoveToFollowing(XPathNodeType.Element);
			common_parser.CheckLocalName(xnav, "altitudeUnits");
            common_parser.fp.LocaleInfo.altitudeUnits = Units.StrToLength(xnav.ToString());
			
			xnav.MoveToFollowing(XPathNodeType.Element);
			common_parser.CheckLocalName(xnav, "distanceUnits");
            common_parser.fp.LocaleInfo.distanceUnits = Units.StrToLength(xnav.ToString());
			
			xnav.MoveToFollowing(XPathNodeType.Element);
			common_parser.CheckLocalName(xnav, "decimalSeparator");
			string dsep = xnav.ToString();
            common_parser.fp.LocaleInfo.fprovider.NumberDecimalSeparator = dsep;

			xnav.MoveToFollowing(XPathNodeType.Element);
			common_parser.CheckLocalName(xnav, "groupSeparator");
			string gsep = xnav.ToString();
			if (gsep == null)
                common_parser.fp.LocaleInfo.fprovider.NumberGroupSeparator = "";
			else
                common_parser.fp.LocaleInfo.fprovider.NumberGroupSeparator = gsep;
		}

        /// <summary>
        /// Returns true if type corresponds to a valid leg type. Returns false otherwise.
        /// </summary>
		private bool IsValidLegType(string type)
		{   
            if (type == "")
                throw new ArgumentNullException("The Leg type can't be null. Check that the attribute name is 'xsi:type', not 'type'");

			string [] ValidTypes = {
				"IFLeg", "IRFLeg", "TFLeg", "DFLeg", "HFLeg", "RFLeg", "LALeg", "TOLeg",
				"IterativeLeg", "IntersectionLeg", "BasicScanLeg",
				"ScanPointLeg" 
			};

			// get prefix
			string prefix = type.Substring(0,type.IndexOf(':'));
			string unq_type = type.Substring(type.IndexOf(':')+1); 			
			
			// check prefix
			if (namespaces[prefix] != xs_flightplan_ns)
				return false;
			
			// check unq_type
			return (Array.IndexOf(ValidTypes, unq_type) >= 0);
		}
		
		// TODO: Emergency plans and the main flight plan have the same structure and
		//       should be read in the same fashion.

        /// <summary>
        /// Read flight plan (used both for main and emergency plans)
        /// </summary>
		private void ReadSinglePlan()
		{
		}
		
        /// <summary>
        /// Read contingency plans
        /// </summary>
		private Contingency ReadContingency(XPathNavigator xnav)
		{
			bool exists = xnav.MoveToFirstChild();
            string fpid = null;
            string legid = null;
            string type = null;
            string entrytype = null;
			
			while (exists){

                if (xnav.NodeType == XPathNodeType.Element){
                        if (xnav.Name == "fp-name")
                        {
                            // Name of the contingency flight plan
                            fpid = xnav.ToString();
                        }
                        else if (xnav.Name == "leg-name")
                        {
                            // Name of the entry leg in the contingency flight plan
                            legid = xnav.ToString();
                        }
                        else if (xnav.Name == "type")
                        {
                            // Name of the entry leg in the contingency flight plan
                            type = xnav.ToString();
                        }
                        else if (xnav.Name == "entry-type")
                        {
                            // Name of the entry leg in the contingency flight plan
                            entrytype = xnav.ToString();
                        }
                        else
                        {
                            string msg = String.Format("Waypoint contains unexpected element: {0}.",
                                                xnav.Name);
                            throw new XMLFPException(msg);
                        }
                }

                exists = xnav.MoveToNext();
            }

            return new Contingency(fpid, legid, type, entrytype);
		}				

        /// <summary>
        /// Reads a waypoint. Following conditions are checked:
		///   - Speed and altitude changes are optional.
		///   - No lat and lon should be specified if a fix is present.
		///   - No fix should be specified if lat,lon are present.
        /// </summary>
		private Waypoint ReadWaypoint(XPathNavigator wpnav)
		{
			Fix fix = null;
			GPSCoords lat = null, lon = null;
			double? alt = null;
			double? speed = null;
			bool? flyOver = null;
			string msg = "Waypoint must have fix or coordinates, not both.";

			bool exists = wpnav.MoveToFirstChild();			
			while (exists)
			{
				if (wpnav.NodeType == XPathNodeType.Element)
				{
					// Read fix or lat/lon pair
					if (wpnav.Name == "fix")
                    {
						if (lat != null || lon != null)
							throw new XMLFPException(msg);
						string fixid = wpnav.ToString();
                        if (common_parser.fp.Fixes.Contains(fixid))
						    fix = fp.Fixes[fixid];
                        else
                            throw new XMLFPException("Cannot find fix named '"+fixid+"'.");
					}				
					else if (wpnav.Name == "coordinates")
                    {
						if (fix != null)
							throw new XMLFPException(msg);
						common_parser.ReadCoords(wpnav, out lat, out lon);				
					}
					else if (wpnav.Name == "fly-over") 
                    {
						// Optional fly-over present
						flyOver = wpnav.ValueAsBoolean; 
					}
					else if (wpnav.Name == "altitude")
                    {
						// Optional altitude present
						double tmpAlt = Convert.ToDouble(wpnav.ToString(), fp.LocaleInfo.fprovider);
						alt = Units.LengthConvert(tmpAlt, fp.LocaleInfo.altitudeUnits, Units.Length.m);
					}
					else if (wpnav.Name == "speed")
                    {
						// Optional speed present
						double tmpSpeed = Convert.ToDouble(wpnav.ToString(), fp.LocaleInfo.fprovider);
						speed = Units.SpeedConvert(tmpSpeed, fp.LocaleInfo.speedUnits, Units.Speed.ms);
					}
					else 
                    {
						msg = String.Format("Waypoint contains unexpected element: {0}.",
						                    wpnav.Name);		
						throw new XMLFPException(msg);
					}
				}
				
				exists = wpnav.MoveToNext();
			}

			bool flyOverVal = (flyOver.HasValue? flyOver.Value : false);
			if (fix != null)
				return new Waypoint(speed, alt, fix, flyOverVal, null);
			else
				return new Waypoint(speed, alt, lat, lon, flyOverVal, null);		
		}
		
	    /// <summary>
	    /// Reads a sequence of legRef's putting them into a list of strings.
		/// ElemTag is the name of the xml element that contains this list.
	    /// </summary>
	    /// <param name="seqnav"></param>
	    /// <param name="elemTag"></param>
	    /// <returns></returns>
		private ArrayList ReadLegRefSequence(XPathNavigator seqnav, string elemTag)
		{
			ArrayList blist = new ArrayList();
			
			common_parser.CheckLocalName(seqnav, elemTag);
			
			// Read body legs.
			string[] legs = seqnav.ToString().Split(' ');
			foreach (string l in legs)
				if (l.Trim() != "")
					blist.Add(l.Trim());
			
			return blist;
		}
		
        /// <summary>
        /// Reads current stage legs. For each leg, first all leg elements are
		/// read and then a final check is performed to ensure that the given
		/// information is in accordance with what is expected for the specific
		/// type of leg being read.
        /// </summary>
		private void ReadStageLegs(XPathNavigator xnav, Stage curStage)
		{
			// Read all legs.
			bool exists = xnav.MoveToFirstChild();
			while (exists)
			{
				if (xnav.NodeType == XPathNodeType.Element) 
                {
					common_parser.CheckLocalName(xnav, "leg");
					
					// Read leg id
					string legId = xnav.GetAttribute("id", String.Empty);
					if (legId == String.Empty)
						throw new XMLFPException("Leg id required.");		
					
					// Read leg type
					string type = xnav.GetAttribute("type", xs_instance_ns);
					if (!IsValidLegType(type))
						throw new XMLFPException(type + " is not a valid leg type.");
				
					// Read leg elem by elem
					LegHelper lh = new LegHelper(legId);
					XPathNavigator legnav = xnav.Clone();
					bool done = !legnav.MoveToFirstChild();
					while (!done)
					{
						if (legnav.NodeType == XPathNodeType.Element)
							ReadLegElement(legnav, lh);
						done = !legnav.MoveToNext();
					}

					// Create and initialize leg
					curStage.Add(lh.CreateLeg(curStage, type));
				}
				
				exists = xnav.MoveToNext();
			}
		}

        /// <summary>
        /// Checks that the first stage is of certain type.
        /// </summary>
        private void CheckStageInit(Stage.Type req_type, Stage.Type? prev_type, Stage.Type cur_type)
        {
            string msg = null;

            if (prev_type == null && req_type != cur_type)
            {
                msg = "FP starts with an incorrect type.";
                throw new XMLFPException(msg);
            }
        }		

        /// <summary>
        /// Checks order of stages. pre_type can be null if first stage.
        /// </summary>
		private void CheckStageOrder(Stage.Type? pre_type, Stage.Type cur_type)
		{
			string msg = null;
			
			switch (cur_type) 
            {
			    case Stage.Type.Taxi:
				    if (pre_type != null && pre_type != Stage.Type.Land && common_parser.fp.type == "STD")
					    msg = "Taxi must begin or end flight plan.";					
				    break;

			    case Stage.Type.TakeOff:
                    if (pre_type != Stage.Type.Taxi && common_parser.fp.type == "STD")
					    msg = "TakeOff must follow Taxi stage."; 				
				    break;

			    case Stage.Type.Departure:
                    if (pre_type != Stage.Type.TakeOff && common_parser.fp.type == "STD")
					    msg = "Departure must follow TakeOff stage.";					
				    break;

			    case Stage.Type.Route:
				    if (pre_type != Stage.Type.Departure &&
                        pre_type != Stage.Type.Mission   && common_parser.fp.type == "STD")
					    msg = "Route must follow Departure or Mission stages."; 				
				    break;

			    case Stage.Type.Mission:
                    if (pre_type != Stage.Type.Route && common_parser.fp.type == "STD")
					    msg = "Mission must follow Route stage."; 				
				    break;

			    case Stage.Type.Arrival:
                    if (pre_type != Stage.Type.Route && common_parser.fp.type == "STD")
					    msg = "Arrival must follow Route stage."; 				
				    break;

			    case Stage.Type.Approach:
                    if (pre_type != Stage.Type.Arrival && common_parser.fp.type == "STD")
					    msg = "Approach must follow Arrival stage."; 				
				    break;

			    case Stage.Type.Land:
				    if (pre_type != Stage.Type.Approach)
					    msg = "Land must follow Approach stage."; 				
				    break;
			}
			
			if (msg != null)
				throw new XMLFPException(msg);
		}		

        /// <summary>
        /// Read stage elements
        /// </summary>
		private void ReadStageElements(Stage s, XPathNavigator xnav)
		{
			XPathNavigator childnav = xnav.Clone();
							
			childnav.MoveToFollowing(XPathNodeType.Element);
			common_parser.CheckLocalName(childnav, "name");
			s.name = childnav.ToString();

			bool moveon = true;
			childnav.MoveToFollowing(XPathNodeType.Element);
			if (childnav.Name == "description") 
				s.description = childnav.ToString();
			else
				moveon = false;
						
			// Read legs belonging to this stage
			if (moveon)
				childnav.MoveToFollowing(XPathNodeType.Element);
			common_parser.CheckLocalName(childnav, "legs");
			ReadStageLegs(childnav.Clone(), s);

			// Read initial legs
			childnav.MoveToNext(XPathNodeType.Element);
			common_parser.CheckLocalName(childnav, "initialLegs");
			string[] legs = childnav.ToString().Split(' ');
			foreach (string l in legs)
				if (l.Trim() != "")
					s.AddInitial(l.Trim());

			// Read final legs
			childnav.MoveToFollowing(XPathNodeType.Element);
			common_parser.CheckLocalName(childnav, "finalLegs");
			legs = childnav.ToString().Split(' ');
			foreach (string l in legs)
				if (l.Trim() != "")
					s.AddFinal(l.Trim());
			
			// TODO: read emergency plan if present
			
			// TODO: anything with groups?			
		}        

        /// <summary>
        /// Read Stages node.
        /// </summary>
		private void ReadStages(XPathNavigator xnav)
		{			
			int stage_count = 0;
			Stage.Type? preceding_stage_type = null;
			Stage.Type? first_stage_type = null;
			
			Stage s = null;

			bool exists = xnav.MoveToFirstChild();
			
			// Read all stages.
			while (exists)
			{				
				if (xnav.NodeType == XPathNodeType.Element)
				{
					common_parser.CheckLocalName(xnav, "stage");
					stage_count++;

					s = new Stage(common_parser.fp);

					// Read stage id
					s.id = xnav.GetAttribute("id", String.Empty);
					if (s.id == String.Empty)
						throw new XMLFPException("Stage id required.");					
					
					// Read stage type
					s.type = Stage.StringToType(xnav.GetAttribute("type", String.Empty));

                    if (common_parser.fp.Type == "STD")
                    {
                        CheckStageOrder(preceding_stage_type, s.type);
                    } 
                    else if (common_parser.fp.Type == "CONT")
                    {
                        CheckStageInit(Stage.Type.Approach, preceding_stage_type, s.type);
                        CheckStageOrder(preceding_stage_type, s.type);
                    } 
                    else if (common_parser.fp.Type == "REACT")
                    {
                            CheckStageInit(Stage.Type.Route, preceding_stage_type, s.type);
                    }
                   
					if (first_stage_type == null)
						first_stage_type = s.type;
					preceding_stage_type = s.type;
					
					// Read manualOnly attribute
					string mo = xnav.GetAttribute("manualOnly", String.Empty);
					if (mo != String.Empty)
						s.manualOnly = Convert.ToBoolean(mo);
					else
						s.manualOnly = false;
					
					// Read igonre attribute
					string str_ignore = xnav.GetAttribute("ignore", String.Empty);
					bool ignore = false;					
					if (str_ignore != String.Empty)
						ignore = Convert.ToBoolean(str_ignore);

					if (!ignore)
					{
						ReadStageElements(s, xnav);

						// Add stage to flight plan
						common_parser.fp.stages.Add(s);
					}
				}
				
				exists = xnav.MoveToNext();
			}
			
			// Once all stages have been read, initialize references.
			for (int i=0; i<common_parser.fp.stages.Count; i++)
                common_parser.fp.stages[i].UpdateReferences();

            CheckInitialLegs();
		}

        /// <summary>
        /// Read main plan.
        /// </summary>
		private void ReadMainPlan(XPathNavigator xnav)
		{				
			// TODO: Put common code used for reading any flight plan into ReadSinglePlan

			// Read flight plan id
			fp.id = xnav.GetAttribute("id", String.Empty);
			if (fp.id == String.Empty)
				throw new XMLFPException("Flight plan id required.");						

			xnav.MoveToFollowing(XPathNodeType.Element);
			common_parser.CheckLocalName(xnav, "name");
			fp.name = xnav.ToString();
			
			xnav.MoveToFollowing(XPathNodeType.Element);
			common_parser.CheckLocalName(xnav, "description");
			fp.description = xnav.ToString();
			
			// read stages of the flight plan
			xnav.MoveToFollowing(XPathNodeType.Element);
			common_parser.CheckLocalName(xnav, "stages");
			ReadStages(xnav.Clone());
		}
		
        /// <summary>
        /// Read FlightPlan document:
		///  1. Locale
		///  2. Fixes
		///  3. Emergency plans
		///  4. Main flight plan
        /// </summary>
		private void ReadFlightPlan(XPathNavigator xnav)
		{
			// root node must be FlightPlan
			xnav.MoveToRoot();
			xnav.MoveToFollowing(XPathNodeType.Element);
			common_parser.CheckLocalName(xnav, "FlightPlan");

			// check schema version (namespace) 
			bool supported = false;
			namespaces = xnav.GetNamespacesInScope(XmlNamespaceScope.Local);
			foreach (KeyValuePair<string, string> kvp in namespaces)
			{
				if (!supported && (kvp.Value == xs_flightplan_ns))
					supported = true;
			}
			if (!supported)
				throw new XMLFPException("Document not using " + xs_flightplan_ns + "namespace.");
				
			xnav.MoveToFollowing(XPathNodeType.Element);
			common_parser.CheckLocalName(xnav, "Locale");
			ReadLocale(xnav);

			xnav.MoveToFollowing(XPathNodeType.Element);
			common_parser.CheckLocalName(xnav, "Fixes");
			common_parser.ReadFixes(xnav.Clone());

            common_parser.EmergencyFileRefs = new ArrayList();
			xnav.MoveToNext(XPathNodeType.Element);
			common_parser.CheckLocalName(xnav, "EmergencyPlans");
            common_parser.ReadEmergencyReferences(xnav.Clone());
			//ReadEmergencyPlans(xnav.Clone());
			
			xnav.MoveToNext(XPathNodeType.Element);
			common_parser.CheckLocalName(xnav, "MainFP");
			ReadMainPlan(xnav);
		}		

        /// <summary>
        /// Read emergency main plan.
        /// </summary>
        private void ReadEmergencyMainPlan(XPathNavigator xnav)
        {
            // TODO: Put common code used for reading any flight plan into ReadSinglePlan

            // Read flight plan id
            common_parser.fp.id = xnav.GetAttribute("id", String.Empty);
            if (fp.id == String.Empty)
                throw new XMLFPException("Flight plan id required.");

            xnav.MoveToFollowing(XPathNodeType.Element);
            common_parser.CheckLocalName(xnav, "name");
            common_parser.fp.name = xnav.ToString();
            common_parser.fp.type = "CONT";

            xnav.MoveToFollowing(XPathNodeType.Element);
            common_parser.CheckLocalName(xnav, "description");
            common_parser.fp.description = xnav.ToString();

            // read stages of the flight plan
            xnav.MoveToFollowing(XPathNodeType.Element);
            common_parser.CheckLocalName(xnav, "stages");
            ReadStages(xnav.Clone());
        }
        
        /// <summary>
        /// Read FlightPlan document:
        ///  1. Locale
        ///  2. Fixes
        ///  3. Emergency plans
        ///  4. Main flight plan
        /// </summary>
        private void ReadEmergencyFlightPlan(XPathNavigator xnav)
        {
            // root node must be FlightPlan
            xnav.MoveToRoot();
            xnav.MoveToFollowing(XPathNodeType.Element);
            common_parser.CheckLocalName(xnav, "EmergencyFlightPlan");

            // check schema version (namespace) 
            bool supported = false;
            namespaces = xnav.GetNamespacesInScope(XmlNamespaceScope.Local);
            foreach (KeyValuePair<string, string> kvp in namespaces)
            {
                if (!supported && (kvp.Value == xs_flightplan_ns))
                    supported = true;
            }
            if (!supported)
                throw new XMLFPException("Document not using " + xs_flightplan_ns + "namespace.");

            xnav.MoveToFollowing(XPathNodeType.Element);
            common_parser.CheckLocalName(xnav, "Locale");
            ReadLocale(xnav);

            xnav.MoveToFollowing(XPathNodeType.Element);
            common_parser.CheckLocalName(xnav, "Fixes");
            common_parser.ReadFixes(xnav.Clone());

            xnav.MoveToNext(XPathNodeType.Element);
            common_parser.CheckLocalName(xnav, "EmergencyPlans");
            common_parser.ReadEmergencyReferences(xnav.Clone());

            xnav.MoveToNext(XPathNodeType.Element);
            common_parser.CheckLocalName(xnav, "MainFP");
            ReadEmergencyMainPlan(xnav);
        }

        #endregion
    }
}
