﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.IO;

namespace Tsp.BB
{
    /// <summary>
    /// Utility class to perform benchmark operations with BranchAndBound.
    /// </summary>
    public class BB_Benchmark
    {
        #region - Attributes -

        /// <summary>
        /// The folder name where are located the testset files.
        /// </summary>
        public String folderName;

        /// <summary>
        /// The filename where to append the results.
        /// </summary>
        public String resultsFilename;

        #endregion
        #region - Constructors -

        /// <summary>
        /// Creates a new instance of BB_Benchmark with the specified parameters.
        /// </summary>
        /// <param name="folderName">The name of the folder where are located the testset files.</param>
        /// <param name="resultsFilename">The file name where to append the results.</param>
        public BB_Benchmark(string folderName, string resultsFilename)
        {
            this.folderName      = folderName;
            this.resultsFilename = resultsFilename;
        }

        #endregion
        #region - Public -

        /// <summary>
        /// Run this benchmark for all the files at the forlderName directory and save 
        /// the results at the resultsFilename file.
        /// </summary>
        public void Run()
        {
            List<String> files = new List<String>(Directory.GetFiles(this.folderName));
            files.Sort();
            foreach (String s in files)
            {
                Run(s);
            }
        }

        #endregion
        #region - Private -

        /// <summary>
        /// Run all the tests for the specified filename.
        /// </summary>
        /// <param name="filename"></param>
        private void Run(String filename)
        {
            for (int i = 1; i <= 12; i++)
            {
                BranchAndBound bb = GetBB(i);

                bb.Run(filename);
                AppendResult(bb, filename);

                GC.Collect();
            }
        }

        /// <summary>
        /// Append the result of the specified B&B algorithm to the results file.
        /// </summary>
        /// <param name="bb">The Branch and Bound algorithm after the execution.</param>
        private void AppendResult(BranchAndBound bb, String graphFileName)
        {
            if (!File.Exists(resultsFilename))
            {
                String header = "File\tInitType\tBranchMethod\tkClosestEdges\tN\tSeconds\tBestValue\tNbIterations\tStates_Inserted\tStates_Pruned\tMax_ASL_Length\n";
                File.WriteAllText(resultsFilename, header, UTF8Encoding.Default);
            }

            String s = String.Format(
                "{0}\t{1}\t{2}\t{3}\t{4}\t{5}\t{6}\t{7}\t{8}\t{9}\t{10}\n",
                graphFileName,
                bb.InitializationTypeToStr(),
                bb.BranchingMethodToString(),
                bb.nbKClosestEdgesToTake,
                State.N,
                bb.Interval.TotalSeconds,
                State.BestValue,
                bb.iter,
                bb.num_inserted,
                State.NbPruned,
                bb.max_asl);


            File.AppendAllText(resultsFilename, s, UTF8Encoding.Default);
        }

        /// <summary>
        /// Gets a BranchAndBound configuration.
        /// </summary>
        /// <param name="id">Must be a number in the [1-12] interval.</param>
        private BranchAndBound GetBB(int id)
        {
            BranchAndBound bb = new BranchAndBound(false);

            // polynomial initialization
            if (id == 1)
            {
                bb.usePolynomialInitialization = true;
                bb.branchingMethod = BranchAndBound.BranchingMethod.State1;
                bb.nbKClosestEdgesToTake = 0;                
            }
            else if (id == 2)
            {
                bb.usePolynomialInitialization = true;
                bb.branchingMethod = BranchAndBound.BranchingMethod.State1;
                bb.nbKClosestEdgesToTake = 10;
            }
            else if (id == 3)
            {
                bb.usePolynomialInitialization = true;
                bb.branchingMethod = BranchAndBound.BranchingMethod.State1;
                bb.nbKClosestEdgesToTake = 5;
            }
            else if (id == 4)
            {
                bb.usePolynomialInitialization = true;
                bb.branchingMethod = BranchAndBound.BranchingMethod.State2;
                bb.nbKClosestEdgesToTake = 0;
            }
            else if (id == 5)
            {
                bb.usePolynomialInitialization = true;
                bb.branchingMethod = BranchAndBound.BranchingMethod.State2;
                bb.nbKClosestEdgesToTake = 10;
            }
            else if (id == 6)
            {
                bb.usePolynomialInitialization = true;
                bb.branchingMethod = BranchAndBound.BranchingMethod.State2;
                bb.nbKClosestEdgesToTake = 5;
            }



            // genetic algorithm initialization
            else if (id == 7)
            {
                bb.usePolynomialInitialization = false;
                bb.branchingMethod = BranchAndBound.BranchingMethod.State1;
                bb.nbKClosestEdgesToTake = 0;
            }
            else if (id == 8)
            {
                bb.usePolynomialInitialization = false;
                bb.branchingMethod = BranchAndBound.BranchingMethod.State1;
                bb.nbKClosestEdgesToTake = 10;
            }
            else if (id == 9)
            {
                bb.usePolynomialInitialization = false;
                bb.branchingMethod = BranchAndBound.BranchingMethod.State1;
                bb.nbKClosestEdgesToTake = 5;
            }
            else if (id == 10)
            {
                bb.usePolynomialInitialization = false;
                bb.branchingMethod = BranchAndBound.BranchingMethod.State2;
                bb.nbKClosestEdgesToTake = 0;
            }
            else if (id == 11)
            {
                bb.usePolynomialInitialization = false;
                bb.branchingMethod = BranchAndBound.BranchingMethod.State2;
                bb.nbKClosestEdgesToTake = 10;
            }
            else if (id == 12)
            {
                bb.usePolynomialInitialization = false;
                bb.branchingMethod = BranchAndBound.BranchingMethod.State2;
                bb.nbKClosestEdgesToTake = 5;
            }

            return bb;
        }

        #endregion
    }
}
