﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

using Tsp.GA;
using LSCollections;

namespace Tsp.BB
{
    /// <summary>
    /// Class that implements the Branch and Bound algoritm.
    /// </summary>
    /// <author>Santiago Jiménez Serrano [sanjiser@gmail.com]</author>
    public class BranchAndBound
    {
        #region - Enums -

        /// <summary>
        /// Enums the possible branching methods available.
        /// Right now, the State1 has the best performance.
        /// </summary>
        public enum BranchingMethod
        {

            State1,

            State2
        }

        #endregion
        #region - Attributes -

        /// <summary>
        /// Indicates if the algorithm must be verbose or not.
        /// </summary>
        private bool verbose;

        /// <summary>
        /// Number of iterations done.
        /// </summary>
        public int iter = 0;

        /// <summary>
        /// The number of states inserted at the Active states structure.
        /// </summary>
        public int num_inserted = 1;

        /// <summary>
        /// The maximun number active states that had been in the PriorityQueue
        /// </summary>
        public int max_asl = int.MinValue;

        /// <summary>
        /// Start date of the algorithm.
        /// </summary>
        public DateTime StartTime;

        /// <summary>
        /// End date of the algorithm.
        /// </summary>
        public DateTime EndTime;

        /// <summary>
        /// Time interval that the algorithm has take.
        /// </summary>
        public TimeSpan Interval;

        /// <summary>
        /// Number of closest edges to take for each node.
        /// 0 means all the edges.
        /// </summary>
        public int nbKClosestEdgesToTake = 0;

        /// <summary>
        /// The branching method to use.
        /// </summary>
        public BranchingMethod branchingMethod = BranchingMethod.State1;

        /// <summary>
        /// Gets a value indicating if use polynomial initialization.
        /// If false, the Genetic Algorithm initialization is used.
        /// </summary>
        public bool usePolynomialInitialization = true;

        #endregion
        #region - Constructors -

        /// <summary>
        /// Creates a new instance of BranchAndBound.
        /// </summary>
        /// <param name="verbose">A value indicating if it must be verbose or not.</param>
        public BranchAndBound(bool verbose)
        {
            this.verbose = verbose;
        }

        #endregion
        #region - Public -

        /// <summary>
        /// Run the BranchAndBound algorithm with the graph at the specified path.
        /// </summary>
        /// <param name="path">The path containing the graph data.</param>
        public void Run(String path)
        {
            Graph g = new Graph();
            g.Load(path);
            Run(g);            
        }

        /// <summary>
        /// Run the BranchAndBound algorithm with the specified graph.
        /// </summary>
        /// <param name="g">The graph.</param>
        public void Run(Graph g)
        {
            if (g == null)
                throw new ArgumentNullException("g");

            // Init Static Variables
            State.TheGraph     = g;
            State.N            = g.nbNodes;
            State.NbPruned     = 0;
            State.BestSolution = null;
            State.BestValue    = int.MaxValue;

            if(nbKClosestEdgesToTake > 2)
                g.TakeOnlyKClosestEdges(nbKClosestEdgesToTake);

            if (this.usePolynomialInitialization)
            {
                // initialize the Best State to one searched polynomialy

                if (this.branchingMethod == BranchingMethod.State1)
                    this.InitBestState1();
                else
                    this.InitBestState2();
            }
            else
            {
                // initialize the Best State using the Genetic Algorithm
                FoundOptStateByGa(g); // by ga
            }

            // Create the initial state
            State s;
            if(branchingMethod == BranchingMethod.State1)
                s = new State();
            else
                s = new State2();

            // Add the initial state to A and run the main loop
            PriorityQueue A = new PriorityQueue();
            A.Enqueue(s);
            Solve(A);
        }

        /// <summary>
        /// Gets a string representing the type of initialization performed over the Initial State.BestValue.
        /// Useful for benchmarking purposes.
        /// </summary>
        /// <returns>The string representing the type of initialization over the Initial State.BestValue.</returns>
        public String InitializationTypeToStr()
        {
            if (this.usePolynomialInitialization)
            {
                if (this.branchingMethod == BranchingMethod.State1)
                    return "Polynomial1";
                else
                    return "Polynomial2";
            }
            else
            {
                // initialize the Best State using the Genetic Algorithm
                return "GA"; // by ga
            }
        }

        /// <summary>
        /// Gets a string representing the branching method used.
        /// Useful for benchmarking purposes.
        /// </summary>
        /// <returns>A string representing the branching method used.</returns>
        public String BranchingMethodToString()
        {
            if (this.branchingMethod == BranchingMethod.State1)
                return "Method1";
            else
                return "Method2";
        }

        #endregion
        #region - Private -

        /// <summary>
        /// Basic Branch and Bound Algorithm.
        /// </summary>
        /// <param name="A">The PriorityQueue with the initial state enqueued to start the algorithm.</param>
        private void Solve(PriorityQueue A)
	    {		
		    iter         = 0;
		    num_inserted = 1;
		    max_asl      = int.MinValue;		    
            StartTime    = DateTime.Now;
            int current_asl;
		
		    if(verbose)
			    Console.WriteLine("iter   |ASL|    ins      pod     estado_seleccionado      mejor_solucion_final");
		
		    while(A.Count > 0)
		    {
			    // Update statistics
			    iter++;
			    current_asl = A.Count;
			    if(current_asl > max_asl)
				    max_asl = A.Count;
			
			    // Selection
			    State s = (State)A.Dequeue();
			
			    // If the best value is worst than the best
			    // solution, it can't reach a better solution
			    if(s.value > State.BestValue)
				    break;
			
			    // Branch
			    // Solutions are not saved in A
			    // At the State.Branch method are detected
			    // and saved only the Best
			    // 
			    // if(s.isSolution())
			    // {
			    //     ...
			    // }
			    // else
			    // {	
			
			    List<State> branchs = s.Branch();
			    bool insert = branchs != null && branchs.Count > 0;
			    if(insert)
			    {
				    //A.addAll(branchs);
                    foreach (State ss in branchs)
                    {
                        A.Enqueue(ss);
                    }
				    num_inserted += branchs.Count;
			    }
			
			    // } // end else
			
			    if(verbose)
			    {			
				    Console.Write(iter + " \t");
				    Console.Write(current_asl + " \t");
				    Console.Write(insert ? branchs.Count.ToString() : "0" + " \t");
				    Console.Write(State.NbPruned + " \t");
				    Console.Write(s.ToString() + " \t");
				    if(State.BestSolution == null)
					    Console.WriteLine("---empty---");
				    else
					    State.BestSolution.Print();
			    }
		    } // end while

            EndTime  = DateTime.Now;
            Interval = EndTime - StartTime;
		
		    Console.WriteLine();
		    Console.WriteLine("Iteraciones      = " + iter);
		    Console.WriteLine("Max |ASL|        = " + max_asl);
		    Console.WriteLine("Total insertados = " + num_inserted);
		    Console.WriteLine("Total podados    = " + State.NbPruned);
		    Console.WriteLine("Restantes        = " + A.Count);		
		    if(State.BestSolution == null)
			    Console.WriteLine("Solution not found");
		    else
		    {
			    Console.WriteLine("Solucion         = " + State.BestSolution);
		    }
	    }

        /// <summary>
        /// Inits the State.BestState and State.BestValue variables 
        /// searching a relative good solution in a polynomial time.
        /// </summary>
        private void InitBestState1()
	    {
            Console.WriteLine("Starting InitBestState()");

            State s       = new State();
            List<State> A = new List<State>();
		    A.Add(s);

		    while(A.Count > 0 && State.BestSolution == null)
		    {
			    // Selection // last one
			    s = A[A.Count-1];
                A.RemoveAt(A.Count-1);	
			
			    List<State> branchs = s.Branch();
			    bool insert = branchs != null && branchs.Count > 0;
			    if(insert)
			    {
                    // insert as last the best value --> first the worst value in the childs
                    while (branchs.Count > 0)
                    {
                        int worstIndex = -1;
                        int worstValue = int.MinValue;
                        for (int k = 0; k < branchs.Count; k++)
                        {
                            if (branchs[k].value >= worstValue)
                            {
                                worstIndex = k;
                                worstValue = branchs[k].value;
                            }
                        }

                        A.Add(branchs[worstIndex]);
                        branchs.RemoveAt(worstIndex);
                    }
                         
                    //l.AddRange(branchs);
			    }
			
		    } // end while
		

            if (State.BestSolution == null)
                Console.WriteLine("Initial Solution not found");
            else
                State.BestSolution.Print();

            Console.WriteLine("End InitBestState()");
	    }

        /// <summary>
        /// Inits the State.BestState and State.BestValue variables 
        /// searching a relative good solution in a polynomial time.
        /// </summary>
        private void InitBestState2()
        {
            Console.WriteLine("Starting InitBestState2()");

            State2 s = new State2();
            List<State2> A = new List<State2>();
            A.Add(s);

            while (A.Count > 0 && State.BestSolution == null)
            {
                // Selection // last one
                s = A[A.Count - 1];
                A.RemoveAt(A.Count - 1);

                List<State> branchs = s.Branch();
                State2 s2;
                bool insert = branchs != null && branchs.Count > 0;
                if (insert)
                {
                    // insert as last the best value --> first the worst value in the childs
                    while (branchs.Count > 0)
                    {
                        int worstIndex = -1;
                        int worstValue = int.MinValue;                        
                        for (int k = 0; k < branchs.Count; k++)
                        {
                            s2 = (State2)branchs[k];
                            if (s2.pathValue >= worstValue)
                            {
                                worstIndex = k;
                                worstValue = s2.pathValue;
                            }
                        }

                        A.Add((State2)branchs[worstIndex]);
                        branchs.RemoveAt(worstIndex);
                    }

                    //l.AddRange(branchs);
                }

            } // end while


            if (State.BestSolution == null)
                Console.WriteLine("Initial Solution not found");
            else
                State.BestSolution.Print();

            Console.WriteLine("End InitBestState2()");
        }
        
        /// <summary>
        /// Inits the State.BestState and State.BestValue variables using the Genetic Algorithm.
        /// Founds an Optimun State using the Genetic Algorithm with the default parameters.
        /// </summary>
        /// <param name="g">The graph to apply the TspGa algorithm.</param>
        /// <returns>The state found by the algorithm.</returns>
        private State FoundOptStateByGa(Graph g)
        {
            Cities cities      = g.AsTspGaCities();
            TspGARunner runner = new TspGARunner();
            if (g.nbNodes > 15)
                runner.maxGenerations = 1000000; // increment the number of generations
            runner.BeginTsp(cities);

            State s = runner.lastResult.ToState();

            State.BestSolution = s;
            State.BestValue    = s.value;

            if (State.BestSolution == null)
                Console.WriteLine("Initial Solution by GA not found");
            else
                Console.WriteLine("\nInitial Solution by GA = " + State.BestSolution);

            return s;
        }


        #endregion
    }
}
