﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace Tsp.BB
{
    /// <summary>
    /// Represents a State in the Branch and Bound algorithm.
    /// </summary>
    /// <author>Santiago Jiménez Serrano [sanjiser@gmail.com]</author>
    public class State : IComparable<State>, IComparable
    {
        #region - Static -

        /// <summary>
        /// Common instance of the Graph been used by all the States.
        /// </summary>
        public static Graph TheGraph;

        /// <summary>
        /// Number of Nodes in the Graph.
        /// </summary>
        public static int N;

        /// <summary>
        /// Current number of pruned States.
        /// </summary>
        public static int NbPruned;

        /// <summary>
        /// State representing the best solution reached
        /// </summary>
        public static State BestSolution;

        /// <summary>
        /// The best value reached
        /// </summary>
        public static int BestValue;

        #endregion
        #region - Attributes -

        /// <summary>
        /// Size of the path
        /// </summary>
        public int k;

        /// <summary>
        /// The path for this state
        /// </summary>
        public int[] path;

        /// <summary>
        /// Function value for this State
        /// </summary>
        public int value;

        #endregion
        #region - Constructors -

        /// <summary>
        /// Constructor for the initial state.
        /// </summary>
        /// <param name="initialState">Node identifier for the initial state.</param>
        public State(int initialState)
        {
            this.k       = 1;
            this.path    = new int[k];
            this.path[0] = TheGraph.GetNode(initialState).id;
            this.value   = TheGraph.GetSumMinWeight();

            NbPruned = 0;
        }

        /// <summary>
        /// Constructor for the initial state, starting at the node 1.
        /// </summary>
        public State()
            :this(1)
        {
        }

        /// <summary>
        /// Constructor used at the Branch method.
        /// </summary>
        /// <param name="parent">The parent State.</param>
        /// <param name="idNewNode">The new node to add to the path.</param>
        /// <param name="newValue">The new value for this State.</param>
        protected State(State parent, int idNewNode, int newValue)
        {
            this.k    = parent.k + 1;
            this.path = new int[this.k];
            for (int i = 0; i < parent.k; i++)
                this.path[i] = parent.path[i];
            this.path[parent.k] = idNewNode;
            this.value = newValue;
        }

        #endregion
        #region - Public -
        
        /// <summary>
        /// Branchs this node.
        /// Also sets the State.BestValue and State.BestSolution variables if arrive to such situation in the branch call.
        /// And also do a pasive Prune, i.e; it doesn't add childs worst than the State.BestSolution state.
        /// </summary>
        /// <returns>The list of resulting branched nodes. null if it's not possible branch this node.</returns>
        public virtual List<State> Branch()
        {
            int currentNode = path[k - 1];
            Node n = TheGraph.GetNode(currentNode);
            Link l;

            if (this.k == N) // Is Solution?
            {
                l = n.GetLinkWith(this.path[0]);
                if (l != null)
                {
                    this.value = fx(l.weight);
                    if (this.value < State.BestValue) // if we have found a best solution, saved it
                    {
                        BestSolution = this;
                        BestValue    = this.value;
                    }
                }
                return null;
            }
            else
            {
                List<State> list = null;
                int id2;
                int aux;

                for (n.ResetIterator(); !n.EndIterator();)
                {
                    l   = n.NextLink();
                    id2 = l.node2.id;
                    if (!this.PathContainsNode(id2))
                    {
                        aux = fx(l.weight);
                        if (aux < State.BestValue)
                        {
                            State s = new State(this, id2, aux);
                            if (list == null)
                                list = new List<State>();
                            list.Add(s);
                        }
                        else
                        {
                            State.NbPruned++;
                        }
                    }
                }
                return list;
            }
        }

        
        /// <summary>
        /// Compare this State with the specified.
        /// Used at PriorityQueue.
        /// </summary>
        /// <param name="otherState">The state to be compared with.</param>
        public int CompareTo(State otherState)
        {
            if (this.value == otherState.value) return 0;
            if (this.value  < otherState.value) return -1;
            return 1;
        }

        /// <summary>
        /// Compare this State with the specified.
        /// </summary>
        /// <param name="obj">An object of type State to be compared with this one.</param>
        public int CompareTo(object obj)
        {
            return CompareTo((State)obj);
        }

        /// <summary>
        /// Gets the String representation for this State.
        /// </summary>
        /// <returns>The String representation for this State.</returns>
        public override String ToString()
        {
            StringBuilder sb = new StringBuilder();

            sb.Append("[");
            for (int i = 0; i < this.k; i++)
            {
                sb.Append(this.path[i]);
                if (i < this.k - 1)
                    sb.Append(",");
            }
            sb.Append("] "); sb.Append(this.value);

            return sb.ToString();
        }

        /// <summary>
        /// Print at the console the string representation for this State.
        /// </summary>
        public void Print()
	    {
	    	Console.WriteLine(this.ToString());
	    }

        /// <summary>
        /// Gets the total weight of the current path for this state. (Real value of fitness function).
        /// Use only this functions in strange situations, not branching.
        /// </summary>
        /// <returns>The total weight of the current path for this state. (Real value of fitness function).</returns>
        public int GetPathValue()
        {
            int v = 0;
            for (int i = 0; i < k - 1; i++)
            {
                v += TheGraph.GetWeight(path[i], path[i + 1]);
            }
            if(k == N)
                v += TheGraph.GetWeight(path[k - 1], path[0]);
            return v;
        }

        #endregion
        #region - Protected -

        /// <summary>
        /// Gets a value indicating if in the current path exists a node with the specified id.
        /// </summary>
        /// <param name="idNode">The node id to search.</param>
        /// <returns>True if a node with the specified id exists in the path. False otherwise.</returns>
        protected bool PathContainsNode(int idNode)
        {
            for (int i = 0; i < this.k; i++)
                if (this.path[i] == idNode)
                    return true;
            return false;
        }

        #endregion
        #region - Private -

        /// <summary>
        /// Cost function for a possible child node.
        /// It takes into account the possible next weight in the path.
        /// </summary>
        /// <param name="linkWeight">The real cost of arrive to a node with the specified weight.</param>
        public int fx(int linkWeight)
        {
            // fx =  coste de llegar a este estado 
            //     + coste del link para llegar desde este estado a otro nodo
            //     - minimo peso de la mejor arista que salia desde el ultimo nodo de este camino
            return this.value + linkWeight - TheGraph.GetNode(this.path[this.k - 1]).minLinkWeight;
        }

        #endregion
    }
}
