//////////////////////////////////////////////////////////////////////////////////////////////////
// File Name: City.cs
//      Date: 06/01/2006
// Copyright (c) 2006 Michael LaLena. All rights reserved.  (www.lalena.com)
// Permission to use, copy, modify, and distribute this Program and its documentation,
//  if any, for any purpose and without fee is hereby granted, provided that:
//   (i) you not charge any fee for the Program, and the Program not be incorporated
//       by you in any software or code for which compensation is expected or received;
//   (ii) the copyright notice listed above appears in all copies;
//   (iii) both the copyright notice and this Agreement appear in all supporting documentation; and
//   (iv) the name of Michael LaLena or lalena.com not be used in advertising or publicity
//          pertaining to distribution of the Program without specific, written prior permission. 
///////////////////////////////////////////////////////////////////////////////////////////////////
using System;
using System.Collections.Generic;

namespace Tsp.GA
{
    /// <summary>
    /// An individual City in our tour.
    /// </summary>
    public class City
    {
        #region - Attributes -

        /// <summary>
        /// The city identifier.
        /// </summary>
        private int id;

        /// <summary>
        /// Private copy of the distance from this city to every other city.
        /// The index in this array is the number of the city linked to.
        /// </summary>
        private List<Link> links = new List<Link>();

        /// <summary>
        /// Private copy of the list of the cities that are closest to this one.
        /// </summary>
        private List<int> closeCities = new List<int>();

        #endregion
        #region - Constructor -

        /// <summary>
        /// Constructor for City.
        /// </summary>
        public City(int id)
        {
            this.id = id;
        }

        /// <summary>
        /// Empty constructor for City.
        /// </summary>
        public City()
            : this(-1)
        {
        }

        #endregion
        #region - Properties -

        /// <summary>
        /// The identifier of this city.
        /// </summary>
        public int Id
        {
            get { return id; }
            set { id = value; }
        }

        /// <summary>
        /// The distance from this city to every other city.
        /// </summary>
        public List<Link> Links
        {
            get { return links; }
            set { links = value; }
        }

        /// <summary>
        /// A list of the cities that are closest to this one.
        /// </summary>
        public List<int> CloseCities
        {
            get { return closeCities; }
        }

        #endregion
        #region - Public -

        /// <summary>
        /// Find the cities that are closest to this one.
        /// </summary>
        /// <param name="numberOfCloseCities">When creating the initial population of tours, this is a greater chance
        /// that a nearby city will be chosen for a link. This is the number of nearby cities that will be considered close.</param>
        public void FindClosestCities( int numberOfCloseCities )
        {
            double shortestDistance;
            int shortestCity = 0;
            Link   aux;
            Link[] auxList = new Link[this.links.Count];
            this.links.CopyTo(auxList);

            if (numberOfCloseCities > Links.Count - 1)
            {
                numberOfCloseCities = Links.Count - 1;
            }

            this.closeCities.Clear();

            for (int i = 0; i < numberOfCloseCities; i++)
            {
                shortestDistance = Double.MaxValue;
                shortestCity     = -1;
                aux              = null;

                for (int cityNum = 0; cityNum < Links.Count; cityNum++)
                {
                    if (auxList[cityNum] != null && this.links[cityNum].Weight < shortestDistance && this.links[cityNum].Connection2 != this.id)
                    {
                        aux              = this.links[cityNum];
                        shortestDistance = this.links[cityNum].Weight;
                        shortestCity     = cityNum;                        
                    }
                }

                this.closeCities.Add(shortestCity);
                auxList[shortestCity] = null;
            }
        }

        #endregion
    }
}
