﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Text;

namespace Tsp.GA
{
    /// <summary>
    /// Utility class to perform benchmark operations with the tsp Genetic Algorithm.
    /// </summary>
    public class GA_Benchmark
    {
        #region - Attributes -

        /// <summary>
        /// The folder name where are located the testset files.
        /// </summary>
        public String folderName;

        /// <summary>
        /// The filename where to append the results.
        /// </summary>
        public String resultsFilename;

        #endregion
        #region - Constructors -

        /// <summary>
        /// Creates a new instance of GA_Benchmark with the specified parameters.
        /// </summary>
        /// <param name="folderName">The name of the folder where are located the testset files.</param>
        /// <param name="resultsFilename">The file name where to append the results.</param>
        public GA_Benchmark(string folderName, string resultsFilename)
        {
            this.folderName      = folderName;
            this.resultsFilename = resultsFilename;
        }

        #endregion
        #region - Public -

        /// <summary>
        /// Run this benchmark for all the files at the forlderName directory and save 
        /// the results at the resultsFilename file.
        /// </summary>
        public void Run()
        {
            List<String> files = new List<String>(Directory.GetFiles(this.folderName));
            files.Sort();
            foreach (String s in files)
            {
                Run(s);
            }
        }

        #endregion
        #region - Private -

        /// <summary>
        /// Run all the tests for the specified filename.
        /// </summary>
        /// <param name="filename"></param>
        private void Run(String filename)
        {
            for (int i = 1; i <= 8; i++)
            {
                TspGARunner runner = GetTSP(i);

                Graph g = new Graph();
                g.Load(filename);
                // Init Static Variables
                BB.State.TheGraph = g;
                BB.State.N = g.nbNodes;
                BB.State.NbPruned = 0;
                BB.State.BestSolution = null;
                BB.State.BestValue = int.MaxValue;

                Cities cities = g.AsTspGaCities();
            
                DateTime t1 = DateTime.Now;
                runner.BeginTsp(cities); // run the algorithm
                DateTime t2 = DateTime.Now;
                double sec = (t2 - t1).TotalSeconds;


                int bestValue = runner.lastResult.ToState().value;
                AppendResult(runner, g.nbNodes, sec, bestValue, filename);

                GC.Collect();
            }
        }

        /// <summary>
        /// Append the result of the specified B&B algorithm to the results file.
        /// </summary>
        /// <param name="tsp">The TSP algorithm runner.</param>
        private void AppendResult(TspGARunner tsp, int nbNodes, double totalSeconds, int bestValue, String graphFileName)
        {
            if (!File.Exists(resultsFilename))
            {
                String header = "File\tN\tSeconds\tPopSize\tNbGenerations\tBestValue\n";
                File.WriteAllText(resultsFilename, header, UTF8Encoding.Default);
            }

            String s = String.Format(
                "{0}\t{1}\t{2}\t{3}\t{4}\t{5}\n",
                graphFileName,
                nbNodes,
                totalSeconds,
                tsp.populationSize,
                tsp.maxGenerations,
                bestValue);

            File.AppendAllText(resultsFilename, s, UTF8Encoding.Default);
        }

        /// <summary>
        /// Gets a TSP configuration.
        /// </summary>
        /// <param name="id">Must be a number in the [1-8] interval.</param>
        private TspGARunner GetTSP(int id)
        {
            TspGARunner tsp = new TspGARunner();

            // polynomial initialization
            if (id == 1)
            {
                tsp.maxGenerations = 300000;               
            }
            else if (id == 2)
            {
                tsp.maxGenerations = 400000;
            }
            else if (id == 3)
            {
                tsp.maxGenerations = 500000;
            }
            else if (id == 4)
            {
                tsp.maxGenerations = 600000;
            }
            else if (id == 5)
            {
                tsp.maxGenerations = 700000;
            }
            else if (id == 6)
            {
                tsp.maxGenerations = 800000;
            }
            else if (id == 7)
            {
                tsp.maxGenerations = 900000;
            }
            else if (id == 8)
            {
                tsp.maxGenerations = 1000000;
            }

            tsp.seed = Math.Abs((int)DateTime.Now.Ticks);
            return tsp;
        }

        #endregion
    }
}
