//////////////////////////////////////////////////////////////////////////////////////////////////
// File Name: TspEventArgs.cs
//      Date: 06/01/2006
// Copyright (c) 2006 Michael LaLena. All rights reserved.  (www.lalena.com)
// Permission to use, copy, modify, and distribute this Program and its documentation,
//  if any, for any purpose and without fee is hereby granted, provided that:
//   (i) you not charge any fee for the Program, and the Program not be incorporated
//       by you in any software or code for which compensation is expected or received;
//   (ii) the copyright notice listed above appears in all copies;
//   (iii) both the copyright notice and this Agreement appear in all supporting documentation; and
//   (iv) the name of Michael LaLena or lalena.com not be used in advertising or publicity
//          pertaining to distribution of the Program without specific, written prior permission. 
///////////////////////////////////////////////////////////////////////////////////////////////////
using System;
using System.Collections.Generic;

namespace Tsp.GA
{
    /// <summary>
    /// Event arguments when the TspGA class wants the GUI to draw a tour.
    /// </summary>
    public class TspGAEventArgs : EventArgs
    {
        #region - Attributes -

        /// <summary>Private copy of the list of cities.</summary>
        private Cities cityList;

        /// <summary>Private copy of the tour of the cities.</summary>
        private Tour bestTour;

        /// <summary>Private copy for which generation this tour came from.</summary>
        private int generation;
        
        /// <summary>Private copy indicating if the TSP algorithm is complete.</summary>
        private bool complete;

        #endregion
        #region - Constructors -

        /// <summary>
        /// Default Constructor.
        /// </summary>
        public TspGAEventArgs()
        {
            this.complete = false;
        }

        /// <summary>
        /// Constructor that sets all the properties.
        /// </summary>
        /// <param name="cityList">The list of cities to draw.</param>
        /// <param name="bestTour">The tour that connects all the cities.</param>
        /// <param name="generation">Which generation is this.</param>
        /// <param name="complete">Is this the last update before we are done.</param>
        public TspGAEventArgs(Cities cityList, Tour bestTour, int generation, bool complete)
        {
            this.cityList   = cityList;
            this.bestTour   = bestTour;
            this.generation = generation;
            this.complete   = complete;
        }

        #endregion
        #region - Properties -

        /// <summary>Public property for list of cities.</summary>
        public Cities CityList
        {
            get { return cityList; }
        }

        /// <summary>Public property for the tour of the cities.</summary>
        public Tour BestTour
        {
            get { return bestTour; }
        }

        /// <summary>Public property for which generation this tour came from.</summary>
        public int Generation
        {
            get { return generation; }
            set { generation = value; }
        }

        /// <summary>Public property indicating if the TSP algorithm is complete.</summary>
        public bool Complete
        {
            get { return complete; }
            set { complete = value; }
        }

        #endregion
        #region - Public -

        /// <summary>
        /// Gets the path of this tour that begins at the specified node.
        /// </summary>
        /// <param name="idNodeStart">The identifier of the node where starts the path.</param>
        /// <returns>The path of this tour that begins at the specified node.</returns>
        public int[] GetPathStartingAt(int idNodeStart)
        {
            int lastCity = 0;
            int nextCity = BestTour[0].Connection1;

            List<int> path = new List<int>();

            foreach (City c in CityList)
            {
                // figure out if the next city in the list is [0] or [1]
                if (lastCity != BestTour[nextCity].Connection1)
                {
                    lastCity = nextCity;
                    nextCity = BestTour[nextCity].Connection1;
                }
                else
                {
                    lastCity = nextCity;
                    nextCity = BestTour[nextCity].Connection2;
                }

                path.Add(lastCity + 1);   // Add 1 to the identifier!!! In our graphs the idenfiers go from 1 to N, including N
                //Console.Write("{0} -> ", lastCity);
            }

            // Gets the part of the array that starts at the idNodeStart and put it at the beggining.
            int index = path.IndexOf(idNodeStart);
            List<int> subset = path.GetRange(index, path.Count - index);
            path.RemoveRange(index, path.Count - index);
            path.InsertRange(0, subset);
            return path.ToArray();
        }

        /// <summary>
        /// Gets the tour and cities information and creates a new State representation according to it.
        /// </summary>
        /// <returns>A State with the path and value information about the tour.</returns>
        public Tsp.BB.State ToState()
        {
            Tsp.BB.State s = new Tsp.BB.State();
            s.path  = this.GetPathStartingAt(1);
            s.k     = s.path.Length;            
            s.value = s.GetPathValue();
            return s;
        }

        #endregion
    }
}