//////////////////////////////////////////////////////////////////////////////////////////////////
// File Name: TspForm.cs
//      Date: 06/01/2006
// Copyright (c) 2006 Michael LaLena. All rights reserved.  (www.lalena.com)
// Permission to use, copy, modify, and distribute this Program and its documentation,
//  if any, for any purpose and without fee is hereby granted, provided that:
//   (i) you not charge any fee for the Program, and the Program not be incorporated
//       by you in any software or code for which compensation is expected or received;
//   (ii) the copyright notice listed above appears in all copies;
//   (iii) both the copyright notice and this Agreement appear in all supporting documentation; and
//   (iv) the name of Michael LaLena or lalena.com not be used in advertising or publicity
//          pertaining to distribution of the Program without specific, written prior permission. 
///////////////////////////////////////////////////////////////////////////////////////////////////
using System;
using System.Collections.Generic;

namespace Tsp.GA
{
    /// <summary>
    /// Utility class to run the TspGa algorithm.
    /// </summary>
    /// <author>Santiago Jimnez Serrano [sanjiser@gmail.com]</author>
    public partial class TspGARunner
    {
        #region - Attributes -

        /// <summary>
        /// The nodes of the graph at the algorithm.
        /// </summary>
        public Cities cityList = null;

        // Default parameter values
        public int populationSize      = 10000;
        public int maxGenerations      = 300000;
        public int mutation            = 3;
        public int groupSize           = 5;
        public int seed                = 0;
        public int numberOfCloseCities = 5;
        public int chanceUseCloseCity  = 90;

        /// <summary>
        /// A value indicating if output or not information at each NewBestTourFound event.
        /// </summary>
        public bool verbose = true;

        /// <summary>
        /// The last result found by the algorithm.
        /// </summary>
        public TspGAEventArgs lastResult = null;

        #endregion
        #region - Public -

        /// <summary>
        /// Begins the TspGa algorithm with the cityList attribute.
        /// </summary>
        public void BeginTsp()
        {
            if (this.cityList == null)
                throw new Exception("A cityList must be set before start the algorithm.");

            this.BeginTsp(this.cityList);
        }

        /// <summary>
        /// Begins the TspGa algorithm with the specified cities.
        /// </summary>
        /// <param name="cityList">The cities representing the graph.</param>
        public void BeginTsp(Cities cityList)
        {
            if (cityList == null)
                throw new ArgumentNullException("cityList");

            this.cityList = cityList;
            cityList.ResetClosestCitiesDistances(numberOfCloseCities);

            TestParameters();

            TspGA tsp = new TspGA();
            tsp.NewBestTourFound += new TspGA.NewBestTourEventHandler(tsp_NewBestTourFound);
            tsp.Begin(populationSize, maxGenerations, groupSize, mutation, seed, chanceUseCloseCity, cityList);
            tsp.NewBestTourFound -= tsp_NewBestTourFound;
            tsp = null;
        }

        #endregion
        #region - Private -

        /// <summary>
        /// Handler for the tsp.NewBestTourFound event.
        /// </summary>
        private void tsp_NewBestTourFound(object sender, TspGAEventArgs e)
        {
            this.lastResult = e;

            if (!verbose)
                return;

            Console.WriteLine("Best Tour Found");
            Console.WriteLine("Fitness    = {0}", e.BestTour.Fitness);
            Console.WriteLine("Generation = {0}", e.Generation);


            if (e.Complete)
            {
                Console.WriteLine("Complete = {0}", e.Complete);
                int lastCity = 0;
                int nextCity = e.BestTour[0].Connection1;

                foreach (City c in e.CityList)
                {
                    // figure out if the next city in the list is [0] or [1]
                    if (lastCity != e.BestTour[nextCity].Connection1)
                    {
                        lastCity = nextCity;
                        nextCity = e.BestTour[nextCity].Connection1;
                    }
                    else
                    {
                        lastCity = nextCity;
                        nextCity = e.BestTour[nextCity].Connection2;
                    }

                    Console.Write("{0} -> ", lastCity + 1);
                }
            }
            Console.WriteLine();
        }

        /// <summary>
        /// Test this attributes before run the algorithm.
        /// If someone is incorrect, throws the corresponding exception.
        /// </summary>
        private void TestParameters()
        {
            if (populationSize <= 0)
            {
                throw new Exception("You must specify a Population Size");
            }
            if (maxGenerations <= 0)
            {
                throw new Exception("You must specify a Maximum Number of Generations");
            }
            if ((mutation < 0) || (mutation > 100))
            {
                throw new Exception("Mutation must be between 0 and 100.");
            }
            if ((groupSize < 2) || (groupSize > populationSize))
            {
                throw new Exception("You must specify a Group (Neighborhood) Size between 2 and the population size.");
            }
            if ((numberOfCloseCities < 3) || (numberOfCloseCities >= cityList.Count))
            {
                throw new Exception("The number of nearby cities to evaluate for the greedy initial populations must be more than 3 and less than the total number of cities.");
            }
            if ((chanceUseCloseCity < 0) || (chanceUseCloseCity > 95))
            {
                throw new Exception("The odds of using a nearby city when creating the initial population must be between 0% - 95%.");
            }
            if (seed < 0)
            {
                throw new Exception("You must specify a Seed for the Random Number Generator");
            }
            if (cityList.Count < 5)
            {
                throw new Exception("You must either load a City List file, or click the map to place at least 5 cities. ");
            }

        }

        #endregion
    }
}