﻿using System;
using System.Collections.Generic;
using System.Text;
using System.IO;

using Tsp.GA;

namespace Tsp
{
    /// <summary>
    /// Represents a Graph.
    /// </summary>
    /// <author>Santiago Jiménez Serrano [sanjiser@gmail.com]</author>
    public class Graph
    {
        #region - Static -

        private static Random r = new Random((int)DateTime.Now.Ticks);

        /// <summary>
        /// Creates a random graph with the specified parameters.
        /// </summary>
        /// <param name="nbNodes">The number of nodes that the graph will contain.</param>
        /// <param name="max_weight">The maximun weight for links.</param>
        /// <returns>The resulting graph.</returns>
        public static Graph Random(int nbNodes, int max_weight)
        {
            Graph g = new Graph();
            g.nbNodes = nbNodes;

            g.CreateNodes(nbNodes);

            for (int i = 1; i <= nbNodes; i++)
            {
                for (int j = i+1; j <= nbNodes; j++)
                {
                    g.AddLink(i, j, r.Next(max_weight));
                }
            }

            g.ResetAdyM();
            return g;
        }

        /// <summary>
        /// Creates a random graph with the specified parameters.
        /// </summary>
        /// <param name="nbNodes">The number of nodes that the graph will contain.</param>
        /// <returns>The resulting graph.</returns>
        public static Graph Random(int nbNodes)
        {
            return Random(nbNodes, 100000);
        }

        /// <summary>
        /// Creates a test set of random graphs at the specified folder.
        /// </summary>
        /// <param name="folder">The folder where to save the files.</param>
        /// <param name="minNbNodes">The minimun number of nodes.</param>
        /// <param name="maxNbNodes">The maximun number of nodes.</param>
        /// <param name="repetitions">The number of graphs to create for each iteration between [min-max]</param>
        public static void CreateRandomTestSet(String folder, int minNbNodes, int maxNbNodes, int repetitions)
        {
            for (int i = minNbNodes; i <= maxNbNodes; i++)
            {
                for (int j = 0; j < repetitions; j++)
                {
                    Graph g = Random(i);
                    String path = folder + "graph_" + i + "_" + j + ".txt";
                    g.Save(path);
                }
            }
        }

        #endregion
        #region - Attributes -

        /// <summary>
        /// List of nodes. Also contains the node 0
        /// </summary>
        private List<Node> nodes;
 
        /// <summary>
        /// List of links.
        /// </summary>
        private List<Link> links;

        /// <summary>
        /// Number of nodes at this Graph.
        /// </summary>
        public int nbNodes;

        /// <summary>
        /// Adyacency matrix for this Graph.
        /// </summary>
        public  int[,]     adyM;

        /// <summary>
        /// Floyd-Warshall matrix for this Graph.
        /// </summary>
        public  int[,]   floydM;

        #endregion
        #region - Constructors -

        /// <summary>
        /// Creates a new instance of Graph.
        /// </summary>
        public Graph()
        {
            this.nodes   = new List<Node>();
            this.links   = new List<Link>();
            this.nbNodes = 0;
        }

        #endregion
        #region - Public -

        /// <summary>
        /// Get the node with the specified identifier.
        /// </summary>
        /// <param name="id">The identifier.</param>
        /// <returns>The node with the specified identifier.</returns>
        public Node GetNode(int id)
        {
            return this.nodes[id];
        }

        /// <summary>
        /// Gets the sum of minLinkWeight for all the nodes in this graph.
        /// </summary>
        /// <returns>The sum of minLinkWeight for all the nodes in this graph.</returns>
        public int GetSumMinWeight()
        {
            int sum = 0;
            for (int i = 1; i <= nbNodes; i++)
                sum += nodes[i].minLinkWeight;
            //Console.WriteLine("Sum(MinLinkW) = {0}", sum);
            return sum;
        }

        /// <summary>
        /// Gets a value indicating if exist link between two nodes with the specified identifier.
        /// </summary>
        /// <param name="id1">The first identifier.</param>
        /// <param name="id2">The second identifier.</param>
        /// <returns>True if exist link. False otherwise.</returns>
        public bool ExistLink(int id1, int id2)
        {
            return this.adyM[id1, id2] != int.MaxValue;
        }

        /// <summary>
        /// Get the weight of the link between two nodes.
        /// </summary>
        /// <param name="id1">The first node identifier.</param>
        /// <param name="id2">The second node identifier.</param>
        /// <returns>The weight of the link between two nodes with the specified identifiers.</returns>
        public int GetWeight(int id1, int id2)
        {
            return this.adyM[id1, id2];
        }

        /// <summary>
        /// Add a new node to this graph.
        /// </summary>
        /// <param name="id">The node identifier.</param>
        /// <param name="numNodes"></param>
        public void AddNode(int id, int numNodes)
        {
            Node n = new Node(id, numNodes);
            this.nodes.Add(n);
        }

        /// <summary>
        /// Add a new link to this graph.
        /// Really adds two links in both directions (n1->n2 and n2->n1).
        /// </summary>
        /// <param name="id_node1">The node1 identifier.</param>
        /// <param name="id_node2">The node2 identifier.</param>
        /// <param name="weight">The weight for the link.</param>
        public void AddLink(int id_node1, int id_node2, int weight)
        {
            Node n1 = GetNode(id_node1);
            Node n2 = GetNode(id_node2);
            if (n1 != null && n2 != null)
            {
                Link l1 = new Link(n1, n2, weight);
                Link l2 = new Link(n2, n1, weight);

                n1.AddLink(l1);
                n2.AddLink(l2);

                this.links.Add(l1);
                this.links.Add(l2);
            }
        }

        /// <summary>
        /// Resets this Adyacency and Floyd Matrix.
        /// </summary>
        public void ResetAdyM()
        {
            this.adyM   = CreateAdyMatrix();
            this.floydM = Utils.GetFloydWarshallM(this);
        }

        /// <summary>
        /// Creates a new Adyacency Matrix corresponding to this Graph.
        /// </summary>
        /// <returns>A new Adyacency Matrix that corresponds to this Graph.</returns>
        public int[,] CreateAdyMatrix()
        {
            int[,] ady = new int[this.nbNodes + 1, this.nbNodes + 1];

            // reset the index 0
            for (int i = 0; i <= this.nbNodes; i++)
            {
                ady[0, i] = int.MaxValue;
                ady[i, 0] = int.MaxValue;
            }

            Node aux;
            Link l;
            for (int i = 1; i <= this.nbNodes; i++)
            {
                aux = this.GetNode(i);

                for (int j = 1; j <= this.nbNodes; j++)
                {
                    if (i == j)
                        ady[i, j] = 0;
                    else
                    {
                        l = aux.links[j];
                        if (l != null)
                        {
                            ady[i, j] = l.weight;
                        }
                        else
                        {
                            ady[i, j] = int.MaxValue;
                        }
                    }
                }
            }

            return ady;
        }

        /// <summary>
        /// Makes that all the nodes in this graph mantains only the k closest edges in its links, removing the other ones.
        /// </summary>
        /// <param name="kClosest">The number of edges to mantain.</param>
        public void TakeOnlyKClosestEdges(int kClosest)
        {
            this.links.Clear();

            foreach (Node n in this.nodes)
            {
                n.TakeOnlyKClosestEdges(kClosest);

                // add the links
                for (n.ResetIterator(); !n.EndIterator();)
                {
                    this.links.Add(n.NextLink());
                }
            }

            this.ResetAdyM();
        }

        /// <summary>
        /// Load this graph from the specified file.
        /// </summary>
        /// <param name="path">The filename containing graph data.</param>
        public void Load(String path)
	    {
		    try
		    {
                StreamReader input = new StreamReader(path);

			    this.nbNodes = int.Parse(input.ReadLine());
			    CreateNodes(this.nbNodes);

			    while(!input.EndOfStream)
			    {
				    AddLink(input.ReadLine());
			    }

                this.ResetAdyM();

			    input.Close();
		    }
		    catch(Exception e)
		    {
			    Console.WriteLine("Error: " + e.ToString());
		    }
	    }

        /// <summary>
        /// Save this graph to the specified file.
        /// </summary>
        /// <param name="path">The filename that will contain the graph data.</param>
        public void Save(String path)
        {
            try
            {
                StreamWriter output = new StreamWriter(path);

                output.WriteLine(this.nbNodes);

                foreach (Link l in this.links)
                {
                    output.WriteLine("{0}\t{1}\t{2}", l.node1.id, l.node2.id, l.weight);
                }

                output.Close();
            }
            catch (Exception e)
            {
                Console.WriteLine("Error: " + e.ToString());
            }
        }

        /// <summary>
        /// Print this graph at the console.
        /// </summary>
        public void Print()
	    {
            Console.WriteLine(this.ToString());
	    }

        /// <summary>
        /// Gets a String representation for this graph.
        /// </summary>
        /// <returns>A String representation for this graph.</returns>
        public override String ToString()
        {
            StringBuilder sb = new StringBuilder();

            sb.Append(nodes.Count);
            sb.Append("\n");

            for (int i = 1; i < nodes.Count; i++)
            {
                Node n = nodes[i];
                sb.Append(n.ToString());
                sb.Append("\n");
            }

            sb.Append("\n");

            for (int i = 0; i < links.Count; i++)
            {
                Link l = links[i];
                sb.Append(l.ToString());
                if (i < links.Count - 1)
                    sb.Append("\n");
            }

            return sb.ToString();
        }

        #region - Tsp.Ga -

        /// <summary>
        /// Converts this graph to a Tsp.Ga.Cities object.
        /// </summary>
        /// <returns>This graph as a Tsp.Ga.Cities object.</returns>
        public Cities AsTspGaCities()
        {
            Cities cities = new Cities();

            // add the cities
            for (int i = 1; i <= nbNodes; i++)
            {
                cities.Add(new City(i));
            }

            // Set the distances
            for (int i = 0; i < cities.Count; i++)
            {
                City city = cities[i];

                for (int j = 0; j < cities.Count; j++)
                {
                    double dist;
                    if (ExistLink(i + 1, j + 1))
                        dist = (double)GetWeight(i + 1, j + 1);
                    else if (i == j)
                        dist = 0;
                    else
                        dist = 10000000; // MAX_VALUE;
                    Tsp.GA.Link l = new Tsp.GA.Link();
                    l.Connection1 = i + 1;
                    l.Connection2 = j + 1;
                    l.Weight = dist;
                    city.Links.Add(l);
                }
            }

            // reset the closest cities
            for (int i = 0; i < cities.Count; i++)
            {
                cities[i].FindClosestCities(5);
            }

            return cities;
        }

        #endregion

        #endregion
        #region - Private -

        /// <summary>
        /// Creates the specified number of nodes at this graph.
        /// </summary>
        /// <param name="n">The number of nodes to be created.</param>
        private void CreateNodes(int n)
        {
            // Add also the node 0
            for (int i = 0; i <= n; i++)
                AddNode(i, n);
        }

        /// <summary>
        /// Add a new link from a data file line.
        /// </summary>
        /// <param name="line">The line of data containing the information.</param>
        private void AddLink(String line)
        {
            String[] tokens = line.Split(new String[] { "\t" }, StringSplitOptions.RemoveEmptyEntries);

            if (tokens.Length != 3)
                throw new ArgumentException("The line must contain 3 tokens: " + line);

            int id1 =   int.Parse(tokens[0]);
            int id2 =   int.Parse(tokens[1]);
            float w = float.Parse(tokens[2]);

            AddLink(id1, id2, (int)w);
        }


        #endregion
    }
}
