﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace Tsp
{
    /// <summary>
    /// Represents a node in a graph.
    /// </summary>
    /// <author>Santiago Jiménez Serrano [sanjiser@gmail.com]</author>
    public class Node
    {
        #region - Attributes -

        /// <summary>
        /// The node indentifier.
        /// </summary>
        public int        id;

        /// <summary>
        /// The weight of the link with minimun weight at this node links.
        /// </summary>
        public int        minLinkWeight;

        /// <summary>
        /// The array of links for this Node.
        /// </summary>
        public Link[]     links;

        /// <summary>
        /// The current number of links for this Node.
        /// </summary>
        private int       count;

        #endregion
        #region - Constructors -

        /// <summary>
        /// Creates a new instance of Node.
        /// </summary>
        /// <param name="id">The node identifier.</param>
        /// <param name="numNodes">The whole number of nodes that the graph will contain.</param>
        public Node(int id, int numNodes)
        {
            this.id            = id;
            this.minLinkWeight = int.MaxValue;
            this.links         = new Link[numNodes+1];// zero is included
            this.count         = 0;
        }

        #endregion
        #region - Properties -

        /// <summary>
        /// Gets the current number of links that this node has.
        /// </summary>
        public int CountLinks
        {
            get { return this.count; }
        }

        #endregion
        #region - Public -

        /// <summary>
        /// Add a link to this node. The link.node1 has to be this node.
        /// </summary>
        /// <param name="l">The link to be added.</param>
        public void AddLink(Link l)
        {
            if(this.links[l.node2.id] == null)
                this.count++; // avoid errors

            links[l.node2.id] = l;
            if (l.weight < this.minLinkWeight)
                this.minLinkWeight = l.weight;
            
        }

        /// <summary>
        /// Gets a value indicating if exist link with a node that has the specied identifier.
        /// </summary>
        /// <param name="idOtherNode">The other node identifier.</param>
        /// <returns>True if exist link. False otherwise.</returns>
        public bool ExistLinkWith(int idOtherNode)
        {
            return this.links[idOtherNode] != null;            
        }

        /// <summary>
        /// Gets the link with the node that has the specified identifier.
        /// </summary>
        /// <param name="idOtherNode">The other node identifier.</param>
        /// <returns>The link with the that has the specified identifier. Null if it doesn't exist.</returns>
        public Link GetLinkWith(int idOtherNode)
        {
            return this.links[idOtherNode];            
        }

        /// <summary>
        /// Makes that this node mantains only the k closest edges in its links, removing the other ones.
        /// </summary>
        /// <param name="kClosest">The number of edges to mantain.</param>
        public void TakeOnlyKClosestEdges(int kClosest)
        {
            if (kClosest >= this.count)
                return;

            Link[] closest = new Link[kClosest];

            // get the k-closest links
            for (int i = 0; i < kClosest; i++)
            {
                int closeValue = int.MaxValue;
                int closeIndex = -1;

                for (int j = 0; j < this.links.Length; j++)
                {
                    if (this.links[j] != null && this.links[j].weight < closeValue)
                    {
                        closeValue = this.links[j].weight;
                        closeIndex = j;
                    }
                }

                if (closeIndex != -1)
                {
                    closest[i] = this.links[closeIndex];
                    this.links[closeIndex] = null;
                }

            }

            // reset the links
            for (int i = 0; i < this.links.Length; i++)
                links[i] = null;

            for (int i = 0; i < kClosest; i++)
            {
                int id2 = closest[i].node2.id;
                links[id2] = closest[i];
                closest[i] = null;
            }
            this.count = kClosest;
        }

        /// <summary>
        /// Gets a string representation for this node.
        /// </summary>
        /// <returns>A string representation for this node.</returns>
        public override String ToString()
        {
            return "Id=" + this.id + " \t nbLinks=" + this.count + " \t minW=" + this.minLinkWeight;
        }

        #region - Iterator -

        private int index        = -1;
        private int currentCount = 0;

        /// <summary>
        /// Reset the Links iterator for this node.
        /// </summary>
        public void ResetIterator()
        {
            this.index = -1;
            this.currentCount = 0;
        }

        /// <summary>
        /// Gets a value indicating if the Links iterator has ended. I.e, there are no more links to iterate on.
        /// </summary>
        /// <returns></returns>
        public bool EndIterator()
        {
            return this.currentCount == this.count;
        }

        /// <summary>
        /// Gets the next link in the iterator and advances one position.
        /// </summary>
        /// <returns>The next link in the iterator.</returns>
        public Link NextLink()
        {
            while(this.links[++index] == null);
            currentCount++;
                return this.links[index];
        }

        #endregion

        #endregion
    }
}
