﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace Tsp
{
    /// <summary>
    /// Class that provides different utility methods to be used in the TSP problem.
    /// </summary>
    /// <author>Santiago Jiménez Serrano [sanjiser@gmail.com]</author>
    public static class Utils
    {
        #region - Static -

        /// <summary>
        /// Apply the Floyd-Warshall algorithm to the specified adyacency matrix.
        /// So, the matrix will contains the shortest possible length between each node.
        /// </summary>
        /// <param name="adyM">The adyacency matrix.</param>
        /// <param name="n">The number of rows and columns in the adyacency matrix.</param>
        /// <see cref="http://es.wikipedia.org/wiki/Algoritmo_de_Floyd-Warshall"/>
        public static void ApplyFloydWarshall(int [,] adyM, int n)
        {
            for(int i = 0; i < n; i++)
                adyM[i, i] = 0;
 
            for(int k = 1; k <= n; k++)
                for(int i = 1; i <= n; i++)
                    for(int j = 1; j <= n; j++)
                    {
                        if (adyM[i, k] != int.MaxValue && adyM[k, j] != int.MaxValue)
                        {
                            int dt = adyM[i, k] + adyM[k, j];
                            if(adyM[i, j] > dt)
                                adyM[i,j] = dt;
                        }
                        
                    }
        }

        /// <summary>
        /// Gets the Floyd-Warshall matrix for the specified graph.
        /// </summary>
        /// <param name="g">The graph to get the adyacency matrix.</param>
        /// <returns>The Floyd-Warshall matrix.</returns>
        /// <see cref="http://es.wikipedia.org/wiki/Algoritmo_de_Floyd-Warshall"/>
        public static int [,] GetFloydWarshallM(Graph g)
        {
            int[,] path = g.CreateAdyMatrix();

            //Console.WriteLine("\nady matrix");
            //Print(path);
 
            ApplyFloydWarshall(path, g.nbNodes);

            //Console.WriteLine("\nfloyd matrix");
            //Print(path);
 
            return path;
        }

        /// <summary>
        /// Gets the Floyd-Warshall matrix for the specified graph avoiding a set of nodes.
        /// </summary>
        /// <param name="g">The graph to get the adyacency matrix.</param>
        /// <param name="avoidNodes">The list of nodes to avoid. Note that the last node in the list will not be avoided.</param>
        /// <returns>The Floyd-Warshall matrix</returns>
        /// <see cref="http://es.wikipedia.org/wiki/Algoritmo_de_Floyd-Warshall"/>
        public static int[,] GetFloydWarshallM(Graph g, int[] avoidNodes)
        {
            int[,] path = g.CreateAdyMatrix();
            int n = g.nbNodes;

            for (int i = 0; i < avoidNodes.Length-1; i++) // don't avoid the last node
            {
                int index = avoidNodes[i];

                for (int j = 0; j <= n; j++)
                {
                    path[j, index] = int.MaxValue;
                    path[index, j] = int.MaxValue;
                }
            }

            ApplyFloydWarshall(path, g.nbNodes);

            //Console.WriteLine("\nfloyd matrix after avoid");
            //Print(path);

            return path;
        }

        /// <summary>
        /// Print to console the specified matrix.
        /// </summary>
        /// <param name="mat">The matrix to be printed.</param>
        public static void Print(int[,] mat)
        {
            int n = mat.GetLength(0);
            int m = mat.GetLength(1);

            for (int i = 0; i < n; i++)
            {
                for (int j = 0; j < m; j++)
                {
                    Console.Write("[{0}, {1}] = {2} ", i, j, mat[i, j]);
                }

                Console.WriteLine();
            }
        }

        /// <summary>
        /// Reduce the specified matrix and returns the reduced value.
        /// </summary>
        /// <param name="mat">The matrix to be reduced.</param>
        /// <returns>The reduced value.</returns>
        public static int Reduce(int[,] mat)
        {
            int n = mat.GetLength(0);
            int m = mat.GetLength(1);

            int sum = 0;
            int min; 

            // filas
            for (int i = 0; i < n; i++)
            {
                min = int.MaxValue;

                // get the minimun
                for (int j = 0; j < m; j++)
                {
                    if(i!=j && mat[i,j] != int.MaxValue && mat[i, j] < min)
                        min = mat[i, j];
                }

                // set the minimun
                for (int j = 0; j < m; j++)
                {
                    if (i != j && mat[i, j] != int.MaxValue)
                        mat[i, j] -= min;
                }

                if (min != int.MaxValue)
                    sum += min;
            }


            // columns
            for (int i = 0; i < m; i++)
            {
                min = int.MaxValue;

                // get the minimun
                for (int j = 0; j < n; j++)
                {
                    if (i != j && mat[j, i] != int.MaxValue && mat[j, i] < min)
                        min = mat[j, i];
                }

                // set the minimun
                for (int j = 0; j < n; j++)
                {
                    if (i != j && mat[j, i] != int.MaxValue)
                        mat[j, i] -= min;
                }

                if (min != int.MaxValue)
                    sum += min;
            }

            return sum;            
        }

        /// <summary>
        /// Creates a new copy of the specified matrix.
        /// </summary>
        /// <param name="from">The matrix to copy.</param>
        /// <returns>The new copied matrix.</returns>
        public static int[,] Copy(int[,] from)
        {
            int n = from.GetLength(0);
            int m = from.GetLength(1);

            int[,] copy = new int[n, m];

            for (int i = 0; i < n; i++)
                for (int j = 0; j < m; j++)
                    copy[i, j] = from[i, j];

            return copy;
        }

        #endregion
    }
}
